@@ -29,6 +29,7 @@
 #include "maprenderer.h"
 
 #include "imagelayer.h"
+#include "mapobject.h"
 #include "tile.h"
 #include "tilelayer.h"
 
@@ -40,6 +41,9 @@
 
 using namespace Tiled;
 
+MapRenderer::~MapRenderer()
+{}
+
 QRectF MapRenderer::boundingRect(const ImageLayer *imageLayer) const
 {
     return QRectF(QPointF(), imageLayer->image().size());
@@ -54,6 +58,62 @@ void MapRenderer::drawImageLayer(QPainter *painter,
     painter->drawPixmap(QPointF(), imageLayer->image());
 }
 
+void MapRenderer::drawPointObject(QPainter *painter, const QColor &color) const
+{
+    const qreal lineWidth = objectLineWidth();
+    const qreal scale = painterScale();
+    const qreal shadowDist = (lineWidth == 0 ? 1 : lineWidth) / scale;
+    const QPointF shadowOffset = QPointF(shadowDist * 0.5,
+                                         shadowDist * 0.5);
+
+    QPen linePen(color, lineWidth, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin);
+    linePen.setCosmetic(true);
+    QPen shadowPen(linePen);
+    shadowPen.setColor(Qt::black);
+
+    QColor brushColor = color;
+    brushColor.setAlpha(50);
+    const QBrush fillBrush(brushColor);
+
+    painter->setPen(Qt::NoPen);
+    painter->setBrush(fillBrush);
+
+    QPainterPath path;
+
+    const qreal radius = 10.0;
+    const qreal sweep = 235.0;
+    const qreal startAngle = 90.0 - sweep / 2;
+    QRectF rectangle(-radius, -radius, radius * 2, radius * 2);
+    path.moveTo(radius * cos(startAngle * M_PI / 180.0), -radius * sin(startAngle * M_PI / 180.0));
+    path.arcTo(rectangle, startAngle, sweep);
+    path.lineTo(0, 2 * radius);
+    path.closeSubpath();
+
+    painter->translate(0, -2 * radius);
+
+    painter->setPen(shadowPen);
+    painter->setBrush(Qt::NoBrush);
+    painter->drawPath(path.translated(shadowOffset));
+
+    painter->setPen(linePen);
+    painter->setBrush(fillBrush);
+    painter->drawPath(path);
+
+    const QBrush opaqueBrush(color);
+    painter->setBrush(opaqueBrush);
+    const qreal smallRadius = radius / 3.0;
+    painter->drawEllipse(-smallRadius, -smallRadius, smallRadius * 2, smallRadius * 2);
+}
+
+QPainterPath MapRenderer::pointShape(const MapObject *object) const
+{
+    Q_ASSERT(object->shape() == MapObject::Point);
+    QPainterPath path;
+    path.addRect(QRect(-10, -30, 20, 30));
+    path.translate(pixelToScreenCoords(object->position()));
+    return path;
+}
+
 void MapRenderer::setFlag(RenderFlag flag, bool enabled)
 {
     if (enabled)
@@ -71,7 +131,7 @@ QPolygonF MapRenderer::lineToPolygon(const QPointF &start, const QPointF &end)
     QPointF direction = QVector2D(end - start).normalized().toPointF();
     QPointF perpendicular(-direction.y(), direction.x());
 
-    const qreal thickness = 5.0f; // 5 pixels on each side
+    const qreal thickness = 5.0; // 5 pixels on each side
     direction *= thickness;
     perpendicular *= thickness;
 
@@ -92,7 +152,7 @@ QPen MapRenderer::makeGridPen(const QPaintDevice *device, QColor color) const
 #endif
 
 #ifdef Q_OS_MAC
-    const qreal dpiScale = 1.0f;
+    const qreal dpiScale = 1.0;
 #else
     const qreal dpiScale = device->logicalDpiX() / 96.0;
 #endif
@@ -164,7 +224,9 @@ void CellRenderer::render(const Cell &cell, const QPointF &pos, const QSizeF &si
         return;
     }
 
-    if (mTile != tile)
+    // The USHRT_MAX limit is rather arbitrary but avoids a crash in
+    // drawPixmapFragments for a large number of fragments.
+    if (mTile != tile || mFragments.size() == USHRT_MAX)
         flush();
 
     const QPixmap &image = tile->image();
@@ -190,7 +252,7 @@ void CellRenderer::render(const Cell &cell, const QPointF &pos, const QSizeF &si
     fragment.scaleY = flippedVertically ? -1 : 1;
     fragment.rotation = 0;
     fragment.opacity = 1;
-    
+
     if (origin == BottomCenter)
         fragment.x -= sizeHalf.x();
 
@@ -205,7 +267,7 @@ void CellRenderer::render(const Cell &cell, const QPointF &pos, const QSizeF &si
     } else if (cell.flippedAntiDiagonally()) {
         Q_ASSERT(mCellType == OrthogonalCells);
         fragment.rotation = 90;
-        
+
         flippedHorizontally = cell.flippedVertically();
         flippedVertically = !cell.flippedHorizontally();
 
@@ -215,7 +277,7 @@ void CellRenderer::render(const Cell &cell, const QPointF &pos, const QSizeF &si
         if (origin != BottomCenter)
             fragment.x += halfDiff;
     }
-    
+
     fragment.scaleX = scale.width() * (flippedHorizontally ? -1 : 1);
     fragment.scaleY = scale.height() * (flippedVertically ? -1 : 1);
 
