@@ -26,11 +26,12 @@
 #include "map.h"
 #include "mapobject.h"
 #include "objectgroup.h"
+#include "objecttemplate.h"
 #include "properties.h"
+#include "terrain.h"
 #include "tile.h"
 #include "tilelayer.h"
 #include "tileset.h"
-#include "terrain.h"
 
 #include <QCoreApplication>
 
@@ -59,6 +60,7 @@ QVariant MapToVariantConverter::toVariant(const Map &map, const QDir &mapDir)
     mapVariant[QLatin1String("height")] = map.height();
     mapVariant[QLatin1String("tilewidth")] = map.tileWidth();
     mapVariant[QLatin1String("tileheight")] = map.tileHeight();
+    mapVariant[QLatin1String("infinite")] = map.infinite();
     mapVariant[QLatin1String("nextobjectid")] = map.nextObjectId();
 
     addProperties(mapVariant, map.properties());
@@ -81,7 +83,7 @@ QVariant MapToVariantConverter::toVariant(const Map &map, const QDir &mapDir)
     unsigned firstGid = 1;
     for (const SharedTileset &tileset : map.tilesets()) {
         tilesetVariants << toVariant(*tileset, firstGid);
-        mGidMapper.insert(firstGid, tileset.data());
+        mGidMapper.insert(firstGid, tileset);
         firstGid += tileset->nextTileId();
     }
     mapVariant[QLatin1String("tilesets")] = tilesetVariants;
@@ -99,6 +101,26 @@ QVariant MapToVariantConverter::toVariant(const Tileset &tileset,
     return toVariant(tileset, 0);
 }
 
+QVariant MapToVariantConverter::toVariant(const ObjectTemplate &objectTemplate,
+                                          const QDir &directory)
+{
+    mMapDir = directory;
+    QVariantMap objectTemplateVariant;
+
+    objectTemplateVariant[QLatin1String("type")] = QLatin1String("template");
+
+    mGidMapper.clear();
+    if (Tileset *tileset = objectTemplate.object()->cell().tileset()) {
+        unsigned firstGid = 1;
+        mGidMapper.insert(firstGid, tileset->sharedPointer());
+        objectTemplateVariant[QLatin1String("tileset")] = toVariant(*tileset, firstGid);
+    }
+
+    objectTemplateVariant[QLatin1String("object")] = toVariant(*objectTemplate.object());
+
+    return objectTemplateVariant;
+}
+
 QVariant MapToVariantConverter::toVariant(const Tileset &tileset,
                                           int firstGid) const
 {
@@ -108,7 +130,7 @@ QVariant MapToVariantConverter::toVariant(const Tileset &tileset,
         tilesetVariant[QLatin1String("firstgid")] = firstGid;
 
     const QString &fileName = tileset.fileName();
-    if (!fileName.isEmpty()) {
+    if (!fileName.isEmpty() && firstGid > 0) {
         QString source = mMapDir.relativeFilePath(fileName);
         tilesetVariant[QLatin1String("source")] = source;
 
@@ -151,9 +173,9 @@ QVariant MapToVariantConverter::toVariant(const Tileset &tileset,
     }
 
     // Write the image element
-    const QString &imageSource = tileset.imageSource();
+    const QUrl &imageSource = tileset.imageSource();
     if (!imageSource.isEmpty()) {
-        const QString rel = mMapDir.relativeFilePath(tileset.imageSource());
+        const QString rel = toFileReference(imageSource, mMapDir);
 
         tilesetVariant[QLatin1String("image")] = rel;
 
@@ -185,10 +207,10 @@ QVariant MapToVariantConverter::toVariant(const Tileset &tileset,
                 terrainIds << QVariant(tile->cornerTerrainId(j));
             tileVariant[QLatin1String("terrain")] = terrainIds;
         }
-        if (tile->probability() != 1.f)
+        if (tile->probability() != 1.0)
             tileVariant[QLatin1String("probability")] = tile->probability();
         if (!tile->imageSource().isEmpty()) {
-            const QString rel = mMapDir.relativeFilePath(tile->imageSource());
+            const QString rel = toFileReference(tile->imageSource(), mMapDir);
             tileVariant[QLatin1String("image")] = rel;
 
             const QSize tileSize = tile->size();
@@ -245,11 +267,7 @@ QVariant MapToVariantConverter::toVariant(const Properties &properties) const
     Properties::const_iterator it = properties.constBegin();
     Properties::const_iterator it_end = properties.constEnd();
     for (; it != it_end; ++it) {
-        QVariant value = toExportValue(it.value());
-
-        if (it.value().userType() == filePathTypeId())
-            value = mMapDir.relativeFilePath(value.toString());
-
+        const QVariant value = toExportValue(it.value(), mMapDir);
         variantMap[it.key()] = value;
     }
 
@@ -298,25 +316,27 @@ QVariant MapToVariantConverter::toVariant(const TileLayer &tileLayer,
     QVariantMap tileLayerVariant;
     tileLayerVariant[QLatin1String("type")] = QLatin1String("tilelayer");
 
-    tileLayerVariant[QLatin1String("width")] = tileLayer.width();
-    tileLayerVariant[QLatin1String("height")] = tileLayer.height();
+    QRect bounds = tileLayer.bounds().translated(-tileLayer.position());
+
+    if (tileLayer.map()->infinite()) {
+        tileLayerVariant[QLatin1String("width")] = bounds.width();
+        tileLayerVariant[QLatin1String("height")] = bounds.height();
+        tileLayerVariant[QLatin1String("startx")] = bounds.left();
+        tileLayerVariant[QLatin1String("starty")] = bounds.top();
+    } else {
+        tileLayerVariant[QLatin1String("width")] = tileLayer.width();
+        tileLayerVariant[QLatin1String("height")] = tileLayer.height();
+    }
 
     addLayerAttributes(tileLayerVariant, tileLayer);
 
     switch (format) {
     case Map::XML:
-    case Map::CSV: {
-        QVariantList tileVariants;
-        for (int y = 0; y < tileLayer.height(); ++y)
-            for (int x = 0; x < tileLayer.width(); ++x)
-                tileVariants << mGidMapper.cellToGid(tileLayer.cellAt(x, y));
-
-        tileLayerVariant[QLatin1String("data")] = tileVariants;
+    case Map::CSV:
         break;
-    }
     case Map::Base64:
     case Map::Base64Zlib:
-    case Map::Base64Gzip: {
+    case Map::Base64Gzip:
         tileLayerVariant[QLatin1String("encoding")] = QLatin1String("base64");
 
         if (format == Map::Base64Zlib)
@@ -324,10 +344,29 @@ QVariant MapToVariantConverter::toVariant(const TileLayer &tileLayer,
         else if (format == Map::Base64Gzip)
             tileLayerVariant[QLatin1String("compression")] = QLatin1String("gzip");
 
-        QByteArray layerData = mGidMapper.encodeLayerData(tileLayer, format);
-        tileLayerVariant[QLatin1String("data")] = layerData;
         break;
     }
+
+    if (tileLayer.map()->infinite()) {
+        QVariantList chunkVariants;
+
+        for (const QRect &rect : tileLayer.sortedChunksToWrite()) {
+            QVariantMap chunkVariant;
+
+            chunkVariant[QLatin1String("x")] = rect.x();
+            chunkVariant[QLatin1String("y")] = rect.y();
+            chunkVariant[QLatin1String("width")] = rect.width();
+            chunkVariant[QLatin1String("height")] = rect.height();
+
+            addTileLayerData(chunkVariant, tileLayer, format, rect);
+
+            chunkVariants.append(chunkVariant);
+        }
+
+        tileLayerVariant[QLatin1String("chunks")] = chunkVariants;
+    } else {
+        addTileLayerData(tileLayerVariant, tileLayer, format,
+                         QRect(0, 0, tileLayer.width(), tileLayer.height()));
     }
 
     return tileLayerVariant;
@@ -345,67 +384,108 @@ QVariant MapToVariantConverter::toVariant(const ObjectGroup &objectGroup) const
 
     addLayerAttributes(objectGroupVariant, objectGroup);
     QVariantList objectVariants;
-    for (const MapObject *object : objectGroup.objects()) {
-        QVariantMap objectVariant;
-        const QString &name = object->name();
-        const QString &type = object->type();
+    for (const MapObject *object : objectGroup.objects())
+        objectVariants << toVariant(*object);
 
-        addProperties(objectVariant, object->properties());
+    objectGroupVariant[QLatin1String("objects")] = objectVariants;
 
-        objectVariant[QLatin1String("id")] = object->id();
+    return objectGroupVariant;
+}
+
+QVariant MapToVariantConverter::toVariant(const MapObject &object) const
+{
+    QVariantMap objectVariant;
+    const QString &name = object.name();
+    const QString &type = object.type();
+
+    addProperties(objectVariant, object.properties());
+
+    if (const ObjectTemplate *objectTemplate = object.objectTemplate()) {
+        QString relativeFileName = mMapDir.relativeFilePath(objectTemplate->fileName());
+        objectVariant[QLatin1String("template")] = relativeFileName;
+    }
+
+    bool notTemplateInstance = !object.isTemplateInstance();
+
+    int id = object.id();
+    if (id != 0)
+        objectVariant[QLatin1String("id")] = id;
+
+    if (notTemplateInstance || object.propertyChanged(MapObject::NameProperty))
         objectVariant[QLatin1String("name")] = name;
+
+    if (notTemplateInstance || object.propertyChanged(MapObject::TypeProperty))
         objectVariant[QLatin1String("type")] = type;
-        if (!object->cell().isEmpty())
-            objectVariant[QLatin1String("gid")] = mGidMapper.cellToGid(object->cell());
 
-        objectVariant[QLatin1String("x")] = object->x();
-        objectVariant[QLatin1String("y")] = object->y();
-        objectVariant[QLatin1String("width")] = object->width();
-        objectVariant[QLatin1String("height")] = object->height();
-        objectVariant[QLatin1String("rotation")] = object->rotation();
 
-        objectVariant[QLatin1String("visible")] = object->isVisible();
+    if (notTemplateInstance || object.propertyChanged(MapObject::CellProperty))
+        if (!object.cell().isEmpty())
+            objectVariant[QLatin1String("gid")] = mGidMapper.cellToGid(object.cell());
 
-        /* Polygons are stored in this format:
-         *
-         *   "polygon/polyline": [
-         *       { "x": 0, "y": 0 },
-         *       { "x": 1, "y": 1 },
-         *       ...
-         *   ]
-         */
-        switch (object->shape()) {
-        case MapObject::Rectangle:
-            break;
-        case MapObject::Polygon:
-        case MapObject::Polyline: {
+    if (id != 0) {
+        objectVariant[QLatin1String("x")] = object.x();
+        objectVariant[QLatin1String("y")] = object.y();
+    }
+
+    if (notTemplateInstance || object.propertyChanged(MapObject::SizeProperty)) {
+        objectVariant[QLatin1String("width")] = object.width();
+        objectVariant[QLatin1String("height")] = object.height();
+    }
+
+    if (notTemplateInstance || object.propertyChanged(MapObject::RotationProperty))
+        objectVariant[QLatin1String("rotation")] = object.rotation();
+
+    if (notTemplateInstance || object.propertyChanged(MapObject::VisibleProperty))
+        objectVariant[QLatin1String("visible")] = object.isVisible();
+
+    /* Polygons are stored in this format:
+     *
+     *   "polygon/polyline": [
+     *       { "x": 0, "y": 0 },
+     *       { "x": 1, "y": 1 },
+     *       ...
+     *   ]
+     */
+    switch (object.shape()) {
+    case MapObject::Rectangle:
+        break;
+    case MapObject::Polygon:
+    case MapObject::Polyline: {
+        if (notTemplateInstance || object.propertyChanged(MapObject::ShapeProperty)) {
             QVariantList pointVariants;
-            for (const QPointF &point : object->polygon()) {
+            for (const QPointF &point : object.polygon()) {
                 QVariantMap pointVariant;
                 pointVariant[QLatin1String("x")] = point.x();
                 pointVariant[QLatin1String("y")] = point.y();
                 pointVariants.append(pointVariant);
             }
 
-            if (object->shape() == MapObject::Polygon)
+            if (object.shape() == MapObject::Polygon)
                 objectVariant[QLatin1String("polygon")] = pointVariants;
             else
                 objectVariant[QLatin1String("polyline")] = pointVariants;
-            break;
         }
-        case MapObject::Ellipse:
+        break;
+    }
+    case MapObject::Ellipse:
+        if (notTemplateInstance || object.propertyChanged(MapObject::ShapeProperty))
             objectVariant[QLatin1String("ellipse")] = true;
-            break;
-        case MapObject::Text:
-            objectVariant[QLatin1String("text")] = toVariant(object->textData());
-            break;
-        }
-
-        objectVariants << objectVariant;
+        break;
+    case MapObject::Text:
+        if (notTemplateInstance || (object.propertyChanged(MapObject::TextProperty) ||
+                                    object.propertyChanged(MapObject::TextFontProperty) ||
+                                    object.propertyChanged(MapObject::TextAlignmentProperty) ||
+                                    object.propertyChanged(MapObject::TextWordWrapProperty) ||
+                                    object.propertyChanged(MapObject::TextColorProperty)))
+            objectVariant[QLatin1String("text")] = toVariant(object.textData());
+        break;
+    case MapObject::Point:
+        if (notTemplateInstance || object.propertyChanged(MapObject::ShapeProperty))
+            objectVariant[QLatin1String("point")] = true;
+        break;
     }
 
-    objectGroupVariant[QLatin1String("objects")] = objectVariants;
-    return objectGroupVariant;
+    return objectVariant;
 }
 
 QVariant MapToVariantConverter::toVariant(const TextData &textData) const
@@ -457,7 +537,7 @@ QVariant MapToVariantConverter::toVariant(const ImageLayer &imageLayer) const
 
     addLayerAttributes(imageLayerVariant, imageLayer);
 
-    const QString rel = mMapDir.relativeFilePath(imageLayer.imageSource());
+    const QString rel = toFileReference(imageLayer.imageSource(), mMapDir);
     imageLayerVariant[QLatin1String("image")] = rel;
 
     const QColor transColor = imageLayer.transparentColor();
@@ -481,6 +561,32 @@ QVariant MapToVariantConverter::toVariant(const GroupLayer &groupLayer,
     return groupLayerVariant;
 }
 
+void MapToVariantConverter::addTileLayerData(QVariantMap &variant,
+                                             const TileLayer &tileLayer,
+                                             Map::LayerDataFormat format,
+                                             const QRect &bounds) const
+{
+    switch (format) {
+    case Map::XML:
+    case Map::CSV: {
+        QVariantList tileVariants;
+        for (int y = bounds.top(); y <= bounds.bottom(); ++y)
+            for (int x = bounds.left(); x <= bounds.right(); ++x)
+                tileVariants << mGidMapper.cellToGid(tileLayer.cellAt(x, y));
+
+        variant[QLatin1String("data")] = tileVariants;
+        break;
+    }
+    case Map::Base64:
+    case Map::Base64Zlib:
+    case Map::Base64Gzip: {
+        QByteArray layerData = mGidMapper.encodeLayerData(tileLayer, format, bounds);
+        variant[QLatin1String("data")] = layerData;
+        break;
+    }
+    }
+}
+
 void MapToVariantConverter::addLayerAttributes(QVariantMap &layerVariant,
                                                const Layer &layer) const
 {
@@ -512,10 +618,7 @@ void MapToVariantConverter::addProperties(QVariantMap &variantMap,
     Properties::const_iterator it_end = properties.constEnd();
     for (; it != it_end; ++it) {
         int type = it.value().userType();
-        QVariant value = toExportValue(it.value());
-
-        if (type == filePathTypeId())
-            value = mMapDir.relativeFilePath(value.toString());
+        const QVariant value = toExportValue(it.value(), mMapDir);
 
         propertiesMap[it.key()] = value;
         propertyTypesMap[it.key()] = typeToName(type);
