@@ -32,13 +32,80 @@
 #include "tile.h"
 #include "hex.h"
 
+#include <algorithm>
+
 using namespace Tiled;
 
+QRegion Chunk::region(std::function<bool (const Cell &)> condition) const
+{
+    QRegion region;
+
+    for (int y = 0; y < CHUNK_SIZE; ++y) {
+        for (int x = 0; x < CHUNK_SIZE; ++x) {
+            if (condition(cellAt(x, y))) {
+                const int rangeStart = x;
+                for (++x; x <= CHUNK_SIZE; ++x) {
+                    if (x == CHUNK_SIZE || !condition(cellAt(x, y))) {
+                        const int rangeEnd = x;
+                        region += QRect(rangeStart, y, rangeEnd - rangeStart, 1);
+                        break;
+                    }
+                }
+            }
+        }
+    }
+
+    return region;
+}
+
+void Chunk::setCell(int x, int y, const Cell &cell)
+{
+    int index = x + y * CHUNK_SIZE;
+
+    mGrid[index] = cell;
+}
+
+bool Chunk::isEmpty() const
+{
+    for (int y = 0; y < CHUNK_SIZE; ++y) {
+        for (int x = 0; x < CHUNK_SIZE; ++x) {
+            if (!cellAt(x, y).isEmpty())
+                return false;
+        }
+    }
+
+    return true;
+}
+
+bool Chunk::hasCell(std::function<bool (const Cell &)> condition) const
+{
+    for (const Cell &cell : mGrid)
+        if (condition(cell))
+            return true;
+
+    return false;
+}
+
+void Chunk::removeReferencesToTileset(Tileset *tileset)
+{
+    for (int i = 0, i_end = mGrid.size(); i < i_end; ++i) {
+        if (mGrid.at(i).tileset() == tileset)
+            mGrid.replace(i, Cell());
+    }
+}
+
+void Chunk::replaceReferencesToTileset(Tileset *oldTileset, Tileset *newTileset)
+{
+    for (Cell &cell : mGrid) {
+        if (cell.tileset() == oldTileset)
+            cell.setTile(newTileset, cell.tileId());
+    }
+}
+
 TileLayer::TileLayer(const QString &name, int x, int y, int width, int height)
     : Layer(TileLayerType, name, x, y)
     , mWidth(width)
     , mHeight(height)
-    , mGrid(width * height)
     , mUsedTilesetsDirty(false)
 {
     Q_ASSERT(width >= 0);
@@ -88,20 +155,11 @@ QRegion TileLayer::region(std::function<bool (const Cell &)> condition) const
 {
     QRegion region;
 
-    for (int y = 0; y < mHeight; ++y) {
-        for (int x = 0; x < mWidth; ++x) {
-            if (condition(cellAt(x, y))) {
-                const int rangeStart = x;
-                for (++x; x <= mWidth; ++x) {
-                    if (x == mWidth || !condition(cellAt(x, y))) {
-                        const int rangeEnd = x;
-                        region += QRect(rangeStart + mX, y + mY,
-                                        rangeEnd - rangeStart, 1);
-                        break;
-                    }
-                }
-            }
-        }
+    QHashIterator<QPoint, Chunk> it(mChunks);
+    while (it.hasNext()) {
+        it.next();
+        region += it.value().region(condition).translated(it.key().x() * CHUNK_SIZE + mX,
+                                                          it.key().y() * CHUNK_SIZE + mY);
     }
 
     return region;
@@ -112,12 +170,21 @@ QRegion TileLayer::region(std::function<bool (const Cell &)> condition) const
  */
 void Tiled::TileLayer::setCell(int x, int y, const Cell &cell)
 {
-    Q_ASSERT(contains(x, y));
+    if (!findChunk(x, y)) {
+        if (cell == mEmptyCell) {
+            return;
+        } else {
+            mBounds = mBounds.united(QRect(x - (x & CHUNK_MASK),
+                                           y - (y & CHUNK_MASK),
+                                           CHUNK_SIZE,
+                                           CHUNK_SIZE));
+        }
+    }
 
-    Cell &existingCell = mGrid[x + y * mWidth];
+    Chunk &_chunk = chunk(x, y);
 
     if (!mUsedTilesetsDirty) {
-        Tileset *oldTileset = existingCell.isEmpty() ? nullptr : existingCell.tileset();
+        Tileset *oldTileset = _chunk.cellAt(x & CHUNK_MASK, y & CHUNK_MASK).tileset();
         Tileset *newTileset = cell.tileset();
         if (oldTileset != newTileset) {
             if (oldTileset)
@@ -127,26 +194,22 @@ void Tiled::TileLayer::setCell(int x, int y, const Cell &cell)
         }
     }
 
-    existingCell = cell;
+    _chunk.setCell(x & CHUNK_MASK, y & CHUNK_MASK, cell);
 }
 
 TileLayer *TileLayer::copy(const QRegion &region) const
 {
-    const QRegion area = region.intersected(QRect(0, 0, width(), height()));
-    const QRect bounds = region.boundingRect();
-    const QRect areaBounds = area.boundingRect();
-    const int offsetX = qMax(0, areaBounds.x() - bounds.x());
-    const int offsetY = qMax(0, areaBounds.y() - bounds.y());
+    const QRect areaBounds = region.boundingRect();
 
     TileLayer *copied = new TileLayer(QString(),
                                       0, 0,
-                                      bounds.width(), bounds.height());
+                                      areaBounds.width(), areaBounds.height());
 
-    for (const QRect &rect : area.rects())
+    for (const QRect &rect : region.rects())
         for (int x = rect.left(); x <= rect.right(); ++x)
             for (int y = rect.top(); y <= rect.bottom(); ++y)
-                copied->setCell(x - areaBounds.x() + offsetX,
-                                y - areaBounds.y() + offsetY,
+                copied->setCell(x - areaBounds.x(),
+                                y - areaBounds.y(),
                                 cellAt(x, y));
 
     return copied;
@@ -171,9 +234,7 @@ void TileLayer::merge(const QPoint &pos, const TileLayer *layer)
 void TileLayer::setCells(int x, int y, TileLayer *layer,
                          const QRegion &mask)
 {
-    // Determine the overlapping area
     QRegion area = QRect(x, y, layer->width(), layer->height());
-    area &= QRect(0, 0, width(), height());
 
     if (!mask.isEmpty())
         area &= mask;
@@ -214,31 +275,40 @@ void TileLayer::erase(const QRegion &area)
 
 void TileLayer::flip(FlipDirection direction)
 {
-    QVector<Cell> newGrid(mWidth * mHeight);
+    QScopedPointer<TileLayer> newLayer(new TileLayer(QString(), 0, 0, mWidth, mHeight));
 
     Q_ASSERT(direction == FlipHorizontally || direction == FlipVertically);
 
-    for (int y = 0; y < mHeight; ++y) {
-        for (int x = 0; x < mWidth; ++x) {
-            Cell &dest = newGrid[x + y * mWidth];
-            if (direction == FlipHorizontally) {
-                const Cell &source = cellAt(mWidth - x - 1, y);
-                dest = source;
-                dest.setFlippedHorizontally(!source.flippedHorizontally());
-            } else if (direction == FlipVertically) {
-                const Cell &source = cellAt(x, mHeight - y - 1);
-                dest = source;
-                dest.setFlippedVertically(!source.flippedVertically());
+    QHashIterator<QPoint, Chunk> it(mChunks);
+    while (it.hasNext()) {
+        it.next();
+        for (int y = 0; y < CHUNK_SIZE; ++y) {
+            for (int x = 0; x < CHUNK_SIZE; ++x) {
+                int _x = it.key().x() * CHUNK_SIZE + x;
+                int _y = it.key().y() * CHUNK_SIZE + y;
+
+                Cell dest(it.value().cellAt(x, y));
+
+                if (dest.isEmpty())
+                    continue;
+
+                if (direction == FlipHorizontally) {
+                    dest.setFlippedHorizontally(!dest.flippedHorizontally());
+                    newLayer->setCell(mWidth - _x - 1, _y, dest);
+                } else if (direction == FlipVertically) {
+                    dest.setFlippedVertically(!dest.flippedVertically());
+                    newLayer->setCell(_x, mHeight - _y - 1, dest);
+                }
             }
         }
     }
 
-    mGrid = newGrid;
+    mChunks = newLayer->mChunks;
 }
 
 void TileLayer::flipHexagonal(FlipDirection direction)
 {
-    QVector<Cell> newGrid(mWidth * mHeight);
+    QScopedPointer<TileLayer> newLayer(new TileLayer(QString(), 0, 0, mWidth, mHeight));
 
     Q_ASSERT(direction == FlipHorizontally || direction == FlipVertically);
 
@@ -248,28 +318,41 @@ void TileLayer::flipHexagonal(FlipDirection direction)
 
     const char (&flipMask)[16] = (direction == FlipHorizontally ? flipMaskH : flipMaskV);
 
-    for (int y = 0; y < mHeight; ++y) {
-        for (int x = 0; x < mWidth; ++x) {
-            const Cell &source = (direction == FlipHorizontally ? cellAt(mWidth - x - 1, y) : cellAt(x, mHeight - y - 1));
-            Cell &dest = newGrid[x + y * mWidth];
-            dest = source;
+    QHashIterator<QPoint, Chunk> it(mChunks);
+    while (it.hasNext()) {
+        it.next();
+        for (int y = 0; y < CHUNK_SIZE; ++y) {
+            for (int x = 0; x < CHUNK_SIZE; ++x) {
+                int _x = it.key().x() * CHUNK_SIZE + x;
+                int _y = it.key().y() * CHUNK_SIZE + y;
 
-            unsigned char mask =
-                    (static_cast<unsigned char>(dest.flippedHorizontally()) << 3) |
-                    (static_cast<unsigned char>(dest.flippedVertically()) << 2) |
-                    (static_cast<unsigned char>(dest.flippedAntiDiagonally()) << 1) |
-                    (static_cast<unsigned char>(dest.rotatedHexagonal120()) << 0);
+                Cell dest(it.value().cellAt(x, y));
 
-            mask = flipMask[mask];
+                if (dest.isEmpty())
+                    continue;
 
-            dest.setFlippedHorizontally((mask & 8) != 0);
-            dest.setFlippedVertically((mask & 4) != 0);
-            dest.setFlippedAntiDiagonally((mask & 2) != 0);
-            dest.setRotatedHexagonal120((mask & 1) != 0);
+                unsigned char mask =
+                        (static_cast<unsigned char>(dest.flippedHorizontally()) << 3) |
+                        (static_cast<unsigned char>(dest.flippedVertically()) << 2) |
+                        (static_cast<unsigned char>(dest.flippedAntiDiagonally()) << 1) |
+                        (static_cast<unsigned char>(dest.rotatedHexagonal120()) << 0);
+
+                mask = flipMask[mask];
+
+                dest.setFlippedHorizontally((mask & 8) != 0);
+                dest.setFlippedVertically((mask & 4) != 0);
+                dest.setFlippedAntiDiagonally((mask & 2) != 0);
+                dest.setRotatedHexagonal120((mask & 1) != 0);
+
+                if (direction == FlipHorizontally)
+                    newLayer->setCell(mWidth - _x - 1, _y, dest);
+                else
+                    newLayer->setCell(_x, mHeight - _y - 1, dest);
+            }
         }
     }
 
-    mGrid = newGrid;
+    mChunks = newLayer->mChunks;
 }
 
 void TileLayer::rotate(RotateDirection direction)
@@ -282,34 +365,43 @@ void TileLayer::rotate(RotateDirection direction)
 
     int newWidth = mHeight;
     int newHeight = mWidth;
-    QVector<Cell> newGrid(newWidth * newHeight);
+    QScopedPointer<TileLayer> newLayer(new TileLayer(QString(), 0, 0, newWidth, newHeight));
 
-    for (int y = 0; y < mHeight; ++y) {
-        for (int x = 0; x < mWidth; ++x) {
-            const Cell &source = cellAt(x, y);
-            Cell dest = source;
+    QHashIterator<QPoint, Chunk> it(mChunks);
+    while (it.hasNext()) {
+        it.next();
+        for (int y = 0; y < CHUNK_SIZE; ++y) {
+            for (int x = 0; x < CHUNK_SIZE; ++x) {
+                int _x = it.key().x() * CHUNK_SIZE + x;
+                int _y = it.key().y() * CHUNK_SIZE + y;
 
-            unsigned char mask =
-                    (dest.flippedHorizontally() << 2) |
-                    (dest.flippedVertically() << 1) |
-                    (dest.flippedAntiDiagonally() << 0);
+                Cell dest(it.value().cellAt(x, y));
 
-            mask = rotateMask[mask];
+                if (dest.isEmpty())
+                    continue;
 
-            dest.setFlippedHorizontally((mask & 4) != 0);
-            dest.setFlippedVertically((mask & 2) != 0);
-            dest.setFlippedAntiDiagonally((mask & 1) != 0);
+                unsigned char mask =
+                        (dest.flippedHorizontally() << 2) |
+                        (dest.flippedVertically() << 1) |
+                        (dest.flippedAntiDiagonally() << 0);
 
-            if (direction == RotateRight)
-                newGrid[x * newWidth + (mHeight - y - 1)] = dest;
-            else
-                newGrid[(mWidth - x - 1) * newWidth + y] = dest;
+                mask = rotateMask[mask];
+
+                dest.setFlippedHorizontally((mask & 4) != 0);
+                dest.setFlippedVertically((mask & 2) != 0);
+                dest.setFlippedAntiDiagonally((mask & 1) != 0);
+
+                if (direction == RotateRight)
+                    newLayer->setCell(mHeight - _y - 1, _x, dest);
+                else
+                    newLayer->setCell(_y, mWidth - _x - 1, dest);
+            }
         }
     }
 
     mWidth = newWidth;
     mHeight = newHeight;
-    mGrid = newGrid;
+    mChunks = newLayer->mChunks;
 }
 
 void TileLayer::rotateHexagonal(RotateDirection direction, Map *map)
@@ -329,7 +421,7 @@ void TileLayer::rotateHexagonal(RotateDirection direction, Map *map)
 
     int newWidth = topRight.toStaggered(staggerIndex, staggerAxis).x() * 2 + 2;
     int newHeight = bottomRight.toStaggered(staggerIndex, staggerAxis).y() * 2 + 2;
-    QVector<Cell> newGrid(newWidth * newHeight);
+    QScopedPointer<TileLayer> newLayer(new TileLayer(QString(), 0, 0, newWidth, newHeight));
 
     Hex newCenter(newWidth / 2, newHeight / 2, staggerIndex, staggerAxis);
 
@@ -357,40 +449,47 @@ void TileLayer::rotateHexagonal(RotateDirection direction, Map *map)
     const char (&rotateMask)[16] =
             (direction == RotateRight) ? rotateRightMask : rotateLeftMask;
 
-    for (int y = 0; y < mHeight; ++y) {
-        for (int x = 0; x < mWidth; ++x) {
-            const Cell &source = cellAt(x, y);
-            Cell dest = source;
+    QHashIterator<QPoint, Chunk> it(mChunks);
+    while (it.hasNext()) {
+        it.next();
+        for (int y = 0; y < CHUNK_SIZE; ++y) {
+            for (int x = 0; x < CHUNK_SIZE; ++x) {
+                int _x = it.key().x() * CHUNK_SIZE + x;
+                int _y = it.key().y() * CHUNK_SIZE + y;
 
-            unsigned char mask =
-                    (static_cast<unsigned char>(dest.flippedHorizontally()) << 3) |
-                    (static_cast<unsigned char>(dest.flippedVertically()) << 2) |
-                    (static_cast<unsigned char>(dest.flippedAntiDiagonally()) << 1) |
-                    (static_cast<unsigned char>(dest.rotatedHexagonal120()) << 0);
+                Cell dest(it.value().cellAt(x, y));
 
-            mask = rotateMask[mask];
+                if (dest.isEmpty())
+                    continue;
 
-            dest.setFlippedHorizontally((mask & 8) != 0);
-            dest.setFlippedVertically((mask & 4) != 0);
-            dest.setFlippedAntiDiagonally((mask & 2) != 0);
-            dest.setRotatedHexagonal120((mask & 1) != 0);
+                unsigned char mask =
+                        (static_cast<unsigned char>(dest.flippedHorizontally()) << 3) |
+                        (static_cast<unsigned char>(dest.flippedVertically()) << 2) |
+                        (static_cast<unsigned char>(dest.flippedAntiDiagonally()) << 1) |
+                        (static_cast<unsigned char>(dest.rotatedHexagonal120()) << 0);
 
-            Hex rotatedHex(x, y, staggerIndex, staggerAxis);
-            rotatedHex -= center;
-            rotatedHex.rotate(direction);
-            rotatedHex += newCenter;
+                mask = rotateMask[mask];
 
-            QPoint rotatedPoint = rotatedHex.toStaggered(staggerIndex, staggerAxis);
+                dest.setFlippedHorizontally((mask & 8) != 0);
+                dest.setFlippedVertically((mask & 4) != 0);
+                dest.setFlippedAntiDiagonally((mask & 2) != 0);
+                dest.setRotatedHexagonal120((mask & 1) != 0);
 
-            int index = rotatedPoint.y() * newWidth + rotatedPoint.x();
+                Hex rotatedHex(_x, _y, staggerIndex, staggerAxis);
+                rotatedHex -= center;
+                rotatedHex.rotate(direction);
+                rotatedHex += newCenter;
 
-            newGrid[index] = dest;
+                QPoint rotatedPoint = rotatedHex.toStaggered(staggerIndex, staggerAxis);
+
+                newLayer->setCell(rotatedPoint.x(), rotatedPoint.y(), dest);
+            }
         }
     }
 
     mWidth = newWidth;
     mHeight = newHeight;
-    mGrid = newGrid;
+    mChunks = newLayer->mChunks;
 
     QRect filledRect = region().boundingRect();
 
@@ -411,9 +510,11 @@ QSet<SharedTileset> TileLayer::usedTilesets() const
     if (mUsedTilesetsDirty) {
         QSet<SharedTileset> tilesets;
 
-        for (const Cell &cell : mGrid)
-            if (const Tile *tile = cell.tile())
-                tilesets.insert(tile->sharedTileset());
+        for (const Chunk &chunk : mChunks) {
+            for (const Cell &cell : chunk)
+                if (const Tile *tile = cell.tile())
+                    tilesets.insert(tile->sharedTileset());
+        }
 
         mUsedTilesets.swap(tilesets);
         mUsedTilesetsDirty = false;
@@ -424,9 +525,10 @@ QSet<SharedTileset> TileLayer::usedTilesets() const
 
 bool TileLayer::hasCell(std::function<bool (const Cell &)> condition) const
 {
-    for (const Cell &cell : mGrid)
-        if (condition(cell))
+    for (const Chunk &chunk : mChunks) {
+        if (chunk.hasCell(condition))
             return true;
+    }
 
     return false;
 }
@@ -438,10 +540,8 @@ bool TileLayer::referencesTileset(const Tileset *tileset) const
 
 void TileLayer::removeReferencesToTileset(Tileset *tileset)
 {
-    for (int i = 0, i_end = mGrid.size(); i < i_end; ++i) {
-        if (mGrid.at(i).tileset() == tileset)
-            mGrid.replace(i, Cell());
-    }
+    for (Chunk &chunk : mChunks)
+        chunk.removeReferencesToTileset(tileset);
 
     mUsedTilesets.remove(tileset->sharedPointer());
 }
@@ -449,10 +549,8 @@ void TileLayer::removeReferencesToTileset(Tileset *tileset)
 void TileLayer::replaceReferencesToTileset(Tileset *oldTileset,
                                            Tileset *newTileset)
 {
-    for (Cell &cell : mGrid) {
-        if (cell.tileset() == oldTileset)
-            cell.setTile(newTileset, cell.tileId());
-    }
+    for (Chunk &chunk : mChunks)
+        chunk.replaceReferencesToTileset(oldTileset, newTileset);
 
     if (mUsedTilesets.remove(oldTileset->sharedPointer()))
         mUsedTilesets.insert(newTileset->sharedPointer());
@@ -463,68 +561,87 @@ void TileLayer::resize(const QSize &size, const QPoint &offset)
     if (this->size() == size && offset.isNull())
         return;
 
-    QVector<Cell> newGrid(size.width() * size.height());
+    QScopedPointer<TileLayer> newLayer(new TileLayer(QString(), 0, 0, size.width(), size.height()));
 
     // Copy over the preserved part
-    const int startX = qMax(0, -offset.x());
-    const int startY = qMax(0, -offset.y());
-    const int endX = qMin(mWidth, size.width() - offset.x());
-    const int endY = qMin(mHeight, size.height() - offset.y());
+    QRect area = mBounds.translated(offset).intersected(newLayer->rect());
+    for (int y = area.top(); y <= area.bottom(); ++y)
+        for (int x = area.left(); x <= area.right(); ++x)
+            newLayer->setCell(x, y, cellAt(x - offset.x(), y - offset.y()));
 
-    for (int y = startY; y < endY; ++y) {
-        for (int x = startX; x < endX; ++x) {
-            const int index = x + offset.x() + (y + offset.y()) * size.width();
-            newGrid[index] = cellAt(x, y);
-        }
-    }
-
-    mGrid = newGrid;
+    mChunks = newLayer->mChunks;
+    mBounds = newLayer->mBounds;
     setSize(size);
 }
 
+static int clampWrap(int value, int min, int max)
+{
+    int v = value - min;
+    int d = max - min;
+    return (v < 0 ? (v + 1) % d + d - 1 : v % d) + min;
+}
+
 void TileLayer::offsetTiles(const QPoint &offset,
                             const QRect &bounds,
                             bool wrapX, bool wrapY)
 {
-    QVector<Cell> newGrid(mWidth * mHeight);
+    if (offset.isNull())
+        return;
 
-    for (int y = 0; y < mHeight; ++y) {
-        for (int x = 0; x < mWidth; ++x) {
-            // Skip out of bounds tiles
-            if (!bounds.contains(x, y)) {
-                newGrid[x + y * mWidth] = cellAt(x, y);
-                continue;
-            }
+    QScopedPointer<TileLayer> newLayer(clone());
 
+    for (int y = bounds.top(); y <= bounds.bottom(); ++y) {
+        for (int x = bounds.left(); x <= bounds.right(); ++x) {
             // Get position to pull tile value from
             int oldX = x - offset.x();
             int oldY = y - offset.y();
 
             // Wrap x value that will be pulled from
-            if (wrapX && bounds.width() > 0) {
-                while (oldX < bounds.left())
-                    oldX += bounds.width();
-                while (oldX > bounds.right())
-                    oldX -= bounds.width();
-            }
+            if (wrapX)
+                oldX = clampWrap(oldX, bounds.left(), bounds.right() + 1);
 
             // Wrap y value that will be pulled from
-            if (wrapY && bounds.height() > 0) {
-                while (oldY < bounds.top())
-                    oldY += bounds.height();
-                while (oldY > bounds.bottom())
-                    oldY -= bounds.height();
-            }
+            if (wrapY)
+                oldY = clampWrap(oldY, bounds.top(), bounds.bottom() + 1);
 
             // Set the new tile
-            if (contains(oldX, oldY) && bounds.contains(oldX, oldY))
-                newGrid[x + y * mWidth] = cellAt(oldX, oldY);
+            if (bounds.contains(oldX, oldY))
+                newLayer->setCell(x, y, cellAt(oldX, oldY));
             else
-                newGrid[x + y * mWidth] = Cell();
+                newLayer->setCell(x, y, Cell());
         }
     }
 
-    mGrid = newGrid;
+    mChunks = newLayer->mChunks;
+    mBounds = newLayer->mBounds;
+}
+
+void TileLayer::offsetTiles(const QPoint &offset)
+{
+    QScopedPointer<TileLayer> newLayer(new TileLayer(QString(), 0, 0, 0, 0));
+
+    // Process only the allocated chunks
+    QHashIterator<QPoint, Chunk> it(mChunks);
+    while (it.hasNext()) {
+        it.next();
+
+        const QPoint p = it.key();
+        const Chunk &chunk = it.value();
+        const QRect r(p.x() * CHUNK_SIZE,
+                      p.y() * CHUNK_SIZE,
+                      CHUNK_SIZE, CHUNK_SIZE);
+
+        for (int y = r.top(); y <= r.bottom(); ++y) {
+            for (int x = r.left(); x <= r.right(); ++x) {
+                int newX = x + offset.x();
+                int newY = y + offset.y();
+                newLayer->setCell(newX, newY, chunk.cellAt(x - r.left(), y - r.top()));
+            }
+        }
+    }
+
+    mChunks = newLayer->mChunks;
+    mBounds = newLayer->mBounds;
 }
 
 bool TileLayer::canMergeWith(Layer *other) const
@@ -537,12 +654,12 @@ Layer *TileLayer::mergedWith(Layer *other) const
     Q_ASSERT(canMergeWith(other));
 
     const TileLayer *o = static_cast<TileLayer*>(other);
-    const QRect unitedBounds = bounds().united(o->bounds());
-    const QPoint offset = position() - unitedBounds.topLeft();
+    const QRect unitedRect = rect().united(o->rect());
+    const QPoint offset = position() - unitedRect.topLeft();
 
     TileLayer *merged = clone();
-    merged->resize(unitedBounds.size(), offset);
-    merged->merge(o->position() - unitedBounds.topLeft(), o);
+    merged->resize(unitedRect.size(), offset);
+    merged->merge(o->position() - unitedRect.topLeft(), o);
     return merged;
 }
 
@@ -552,8 +669,8 @@ QRegion TileLayer::computeDiffRegion(const TileLayer *other) const
 
     const int dx = other->x() - mX;
     const int dy = other->y() - mY;
-    QRect r = QRect(0, 0, width(), height());
-    r &= QRect(dx, dy, other->width(), other->height());
+
+    const QRect r = bounds().united(other->bounds()).translated(-position());
 
     for (int y = r.top(); y <= r.bottom(); ++y) {
         for (int x = r.left(); x <= r.right(); ++x) {
@@ -574,13 +691,48 @@ QRegion TileLayer::computeDiffRegion(const TileLayer *other) const
 
 bool TileLayer::isEmpty() const
 {
-    for (const Cell &cell : mGrid)
-        if (!cell.isEmpty())
+    for (const Chunk &chunk : mChunks)
+        if (!chunk.isEmpty())
             return false;
 
     return true;
 }
 
+static bool compareRectPos(const QRect &a, const QRect &b)
+{
+    if (a.y() != b.y())
+        return a.y() < b.y();
+    return a.x() < b.x();
+}
+
+/**
+ * Returns a list of rectangles that cover all the used area of this layer.
+ * The list is sorted by the top-left of each rectangle.
+ *
+ * This function is used to determine the chunks to write when saving a tile
+ * layer.
+ */
+QVector<QRect> TileLayer::sortedChunksToWrite() const
+{
+    QVector<QRect> chunksToWrite;
+    chunksToWrite.reserve(mChunks.size());
+
+    QHashIterator<QPoint, Chunk> it(mChunks);
+    while (it.hasNext()) {
+        it.next();
+        if (!it.value().isEmpty()) {
+            const QPoint p = it.key();
+            chunksToWrite.append(QRect(p.x() * CHUNK_SIZE,
+                                       p.y() * CHUNK_SIZE,
+                                       CHUNK_SIZE, CHUNK_SIZE));
+        }
+    }
+
+    std::sort(chunksToWrite.begin(), chunksToWrite.end(), compareRectPos);
+
+    return chunksToWrite;
+}
+
 /**
  * Returns a duplicate of this TileLayer.
  *
@@ -594,7 +746,8 @@ TileLayer *TileLayer::clone() const
 TileLayer *TileLayer::initializeClone(TileLayer *clone) const
 {
     Layer::initializeClone(clone);
-    clone->mGrid = mGrid;
+    clone->mChunks = mChunks;
+    clone->mBounds = mBounds;
     clone->mUsedTilesets = mUsedTilesets;
     clone->mUsedTilesetsDirty = mUsedTilesetsDirty;
     return clone;
