@@ -31,10 +31,13 @@
 
 class QEvent;
 class QKeyEvent;
+class QToolBar;
 
 namespace Tiled {
 
 class Layer;
+class Tile;
+class ObjectTemplate;
 
 namespace Internal {
 
@@ -90,6 +93,10 @@ public:
     bool isEnabled() const;
     void setEnabled(bool enabled);
 
+    Tile *tile() const;
+
+    ObjectTemplate *objectTemplate() const;
+
     /**
      * Activates this tool. If the tool plans to add any items to the scene, it
      * probably wants to do it here.
@@ -131,6 +138,14 @@ public:
      */
     virtual void mouseReleased(QGraphicsSceneMouseEvent *event) = 0;
 
+    /**
+     * Called when a mouse button is pressed a second time on the scene, after
+     * a short interval.
+     *
+     * By default, this function calls mousePressed.
+     */
+    virtual void mouseDoubleClicked(QGraphicsSceneMouseEvent *event);
+
     /**
      * Called when the user presses or releases a modifier key resulting
      * in a change of modifier status, and when the tool is enabled with
@@ -143,8 +158,12 @@ public:
      */
     virtual void languageChanged() = 0;
 
+    virtual void populateToolBar(QToolBar*) {}
+
 public slots:
     void setMapDocument(MapDocument *mapDocument);
+    void setTile(Tile *tile);
+    void setObjectTemplate(ObjectTemplate *objectTemplate);
 
 protected:
     /**
@@ -183,6 +202,8 @@ private:
     QString mStatusInfo;
     QCursor mCursor;
     bool mEnabled;
+    Tile *mTile;
+    ObjectTemplate *mObjectTemplate;
 
     MapDocument *mMapDocument;
 };
@@ -233,6 +254,26 @@ inline bool AbstractTool::isEnabled() const
     return mEnabled;
 }
 
+inline Tile *AbstractTool::tile() const
+{
+    return mTile;
+}
+
+inline void AbstractTool::setTile(Tile *tile)
+{
+    mTile = tile;
+}
+
+inline ObjectTemplate *AbstractTool::objectTemplate() const
+{
+    return mObjectTemplate;
+}
+
+inline void AbstractTool::setObjectTemplate(ObjectTemplate *objectTemplate)
+{
+    mObjectTemplate = objectTemplate;
+}
+
 } // namespace Internal
 } // namespace Tiled
 
