@@ -108,7 +108,8 @@ AdjustTileIndexes::AdjustTileIndexes(MapDocument *mapDocument,
 
         case Layer::ObjectGroupType:
             for (MapObject *mapObject : *static_cast<ObjectGroup*>(layer)) {
-                if (isFromTileset(mapObject->cell())) {
+                if ((!mapObject->isTemplateInstance() || mapObject->propertyChanged(MapObject::CellProperty))
+                        && isFromTileset(mapObject->cell())) {
                     MapObjectCell change;
                     change.object = mapObject;
                     change.cell = adjustCell(mapObject->cell());
@@ -151,7 +152,7 @@ AdjustTileMetaData::AdjustTileMetaData(TilesetDocument *tilesetDocument)
 
     // Adjust tile meta data
     QList<Tile*> tilesChangingProbability;
-    QList<float> tileProbabilities;
+    QList<qreal> tileProbabilities;
     ChangeTileTerrain::Changes terrainChanges;
     QSet<Tile*> tilesToReset;
 
@@ -169,7 +170,7 @@ AdjustTileMetaData::AdjustTileMetaData(TilesetDocument *tilesetDocument)
     auto applyMetaData = [&](Tile *toTile,
                              const Properties &properties,
                              unsigned terrain,
-                             float probability,
+                             qreal probability,
                              ObjectGroup *objectGroup,
                              const QVector<Frame> &frames)
     {
@@ -241,7 +242,7 @@ AdjustTileMetaData::AdjustTileMetaData(TilesetDocument *tilesetDocument)
     QSetIterator<Tile*> resetIterator(tilesToReset);
     while (resetIterator.hasNext()) {
         applyMetaData(resetIterator.next(),
-                      Properties(), -1, 1.0f, nullptr, QVector<Frame>());
+                      Properties(), -1, 1.0, nullptr, QVector<Frame>());
     }
 
     if (!tilesChangingProbability.isEmpty()) {
