@@ -35,10 +35,11 @@
 #include "objectgroup.h"
 #include "tile.h"
 #include "tilelayer.h"
-#include "tilesetmanager.h"
 
 #include <QDebug>
 
+#include "qtcompat_p.h"
+
 using namespace Tiled;
 using namespace Tiled::Internal;
 
@@ -64,9 +65,6 @@ AutoMapper::AutoMapper(MapDocument *workingDocument, Map *rules,
 {
     Q_ASSERT(mMapRules);
 
-    TilesetManager *tilesetManager = TilesetManager::instance();
-    tilesetManager->addReferences(rules->tilesets());
-
     if (!setupRuleMapProperties())
         return;
 
@@ -94,41 +92,61 @@ bool AutoMapper::ruleLayerNameUsed(QString ruleLayerName) const
 
 bool AutoMapper::setupRuleMapProperties()
 {
-    QMapIterator<QString,QVariant> it(mMapRules->properties());
+    QMapIterator<QString, QVariant> it(mMapRules->properties());
     while (it.hasNext()) {
         it.next();
 
         const QString &name = it.key();
         const QVariant &value = it.value();
 
-        bool raiseWarning = true;
-
         if (name.compare(QLatin1String("deletetiles"), Qt::CaseInsensitive) == 0) {
             if (value.canConvert(QVariant::Bool)) {
                 mDeleteTiles = value.toBool();
-                raiseWarning = false;
+                continue;
             }
         } else if (name.compare(QLatin1String("automappingradius"), Qt::CaseInsensitive) == 0) {
             if (value.canConvert(QVariant::Int)) {
                 mAutoMappingRadius = value.toInt();
-                raiseWarning = false;
+                continue;
             }
         } else if (name.compare(QLatin1String("nooverlappingrules"), Qt::CaseInsensitive) == 0) {
             if (value.canConvert(QVariant::Bool)) {
                 mNoOverlappingRules = value.toBool();
-                raiseWarning = false;
+                continue;
             }
         }
 
-        if (raiseWarning) {
-            mWarning += tr("'%1': Property '%2' = '%3' does not make sense. "
-                           "Ignoring this property.")
-                    .arg(mRulePath, name, value.toString()) + QLatin1Char('\n');
-        }
+        mWarning += tr("'%1': Property '%2' = '%3' does not make sense. "
+                       "Ignoring this property.")
+                .arg(mRulePath, name, value.toString()) + QLatin1Char('\n');
     }
     return true;
 }
 
+void AutoMapper::setupInputLayerProperties(InputLayer &inputLayer)
+{
+    inputLayer.strictEmpty = false;
+
+    QMapIterator<QString, QVariant> it(inputLayer.tileLayer->properties());
+    while (it.hasNext()) {
+        it.next();
+
+        const QString &name = it.key();
+        const QVariant &value = it.value();
+
+        if (name.compare(QLatin1String("strictempty"), Qt::CaseInsensitive) == 0) {
+            if (value.canConvert(QVariant::Bool)) {
+                inputLayer.strictEmpty = value.toBool();
+                continue;
+            }
+        }
+
+        mWarning += tr("'%1': Property '%2' = '%3' on layer '%4' does not make sense. "
+                       "Ignoring this property.")
+                .arg(mRulePath, name, value.toString(), inputLayer.tileLayer->name()) + QLatin1Char('\n');
+    }
+}
+
 bool AutoMapper::setupRuleMapTileLayers()
 {
     Q_ASSERT(mLayerList.isEmpty());
@@ -214,10 +232,15 @@ bool AutoMapper::setupRuleMapTileLayers()
             if (!mInputRules[index].contains(name))
                 mInputRules[index].insert(name, InputConditions());
 
+            InputLayer inputLayer;
+            inputLayer.tileLayer = tileLayer;
+            setupInputLayerProperties(inputLayer);
+
+            InputConditions &conditions = mInputRules[index][name];
             if (isNotList)
-                mInputRules[index][name].listNo.append(tileLayer);
+                conditions.listNo.append(inputLayer);
             else
-                mInputRules[index][name].listYes.append(tileLayer);
+                conditions.listYes.append(inputLayer);
 
             continue;
         }
@@ -400,8 +423,7 @@ bool AutoMapper::setupTilesets()
 
     mMapDocument->unifyTilesets(mMapRules, mAddedTilesets);
 
-    const auto &addedTilesets = mAddedTilesets;
-    for (const SharedTileset &tileset : addedTilesets)
+    for (const SharedTileset &tileset : qAsConst(mAddedTilesets))
         mMapDocument->undoStack()->push(new AddTileset(mMapDocument, tileset));
 
     return true;
@@ -413,21 +435,24 @@ void AutoMapper::autoMap(QRegion *where)
     // first resize the active area
     if (mAutoMappingRadius) {
         QRegion region;
-        for (const QRect &r : where->rects()) {
+#if QT_VERSION < 0x050800
+        foreach (const QRect &r, where->rects()) {
+#else
+        for (const QRect &r : *where) {
+#endif
             region += r.adjusted(- mAutoMappingRadius,
                                  - mAutoMappingRadius,
                                  + mAutoMappingRadius,
                                  + mAutoMappingRadius);
         }
-        *where += region;
+        where->swap(region);
     }
 
     // delete all the relevant area, if the property "DeleteTiles" is set
     if (mDeleteTiles) {
-        const QRegion setLayersRegion = getSetLayersRegion();
-        for (RuleOutput &translationTable : mLayerList) {
-            foreach (Layer *layer, translationTable.keys()) {
-                const int index = translationTable.value(layer);
+        const QRegion setLayersRegion = computeSetLayersRegion();
+        for (const RuleOutput &translationTable : mLayerList) {
+            for (const int index : translationTable) {
                 Layer *dstLayer = mMapWork->layerAt(index);
                 const QRegion region = setLayersRegion.intersected(*where);
                 TileLayer *dstTileLayer = dstLayer->asTileLayer();
@@ -446,7 +471,11 @@ void AutoMapper::autoMap(QRegion *where)
     // This needs to be done, so you can rely on the order of the rules at all
     // locations
     QRegion ret;
+#if QT_VERSION < 0x050800
     foreach (const QRect &rect, where->rects()) {
+#else
+    for (const QRect &rect : *where) {
+#endif
         for (int i = 0; i < mRulesInput.size(); ++i) {
             // at the moment the parallel execution does not work yet
             // TODO: make multithreading available!
@@ -457,10 +486,10 @@ void AutoMapper::autoMap(QRegion *where)
     *where = where->united(ret);
 }
 
-QRegion AutoMapper::getSetLayersRegion() const
+QRegion AutoMapper::computeSetLayersRegion() const
 {
     QRegion result;
-    for (const QString &name : mInputRules.names) {
+    for (const QString &name : qAsConst(mInputRules.names)) {
         const int index = mMapWork->indexOfLayer(name, Layer::TileLayerType);
         if (index == -1)
             continue;
@@ -470,140 +499,29 @@ QRegion AutoMapper::getSetLayersRegion() const
     return result;
 }
 
-static bool compareLayerTo(const TileLayer *setLayer,
-                           const QVector<TileLayer*> &listYes,
-                           const QVector<TileLayer*> &listNo,
-                           const QRegion &ruleRegion, const QPoint &offset);
-
-QRect AutoMapper::applyRule(const int ruleIndex, const QRect &where)
-{
-    QRect ret;
-
-    if (mLayerList.isEmpty())
-        return ret;
-
-    const QRegion &ruleInputRegion = mRulesInput.at(ruleIndex);
-    const QRegion &ruleOutputRegion = mRulesOutput.at(ruleIndex);
-    QRect rbr = ruleInputRegion.boundingRect();
-
-    // Since the rule itself is translated, we need to adjust the borders of the
-    // loops. Decrease the size at all sides by one: There must be at least one
-    // tile overlap to the rule.
-    const int minX = where.left() - rbr.left() - rbr.width() + 1;
-    const int minY = where.top() - rbr.top() - rbr.height() + 1;
-
-    const int maxX = where.right() - rbr.left() + rbr.width() - 1;
-    const int maxY = where.bottom() - rbr.top() + rbr.height() - 1;
-
-    // In this list of regions it is stored which parts or the map have already
-    // been altered by exactly this rule. We store all the altered parts to
-    // make sure there are no overlaps of the same rule applied to
-    // (neighbouring) places
-    QVector<QRegion> appliedRegions;
-    if (mNoOverlappingRules)
-        appliedRegions.resize(mMapWork->layerCount());
-
-    for (int y = minY; y <= maxY; ++y)
-    for (int x = minX; x <= maxX; ++x) {
-        bool anyMatch = false;
-
-        const auto &inputRules = mInputRules;
-        for (const InputIndex &inputIndex : inputRules) {
-            bool allLayerNamesMatch = true;
-
-            QMapIterator<QString, InputConditions> inputIndexIterator(inputIndex);
-            while (inputIndexIterator.hasNext()) {
-                inputIndexIterator.next();
-
-                const QString &name = inputIndexIterator.key();
-                const InputConditions &conditions = inputIndexIterator.value();
-
-                const int i = mMapWork->indexOfLayer(name, Layer::TileLayerType);
-                if (i == -1) {
-                    allLayerNamesMatch = false;
-                } else {
-                    const TileLayer *setLayer = mMapWork->layerAt(i)->asTileLayer();
-                    allLayerNamesMatch &= compareLayerTo(setLayer,
-                                                         conditions.listYes,
-                                                         conditions.listNo,
-                                                         ruleInputRegion,
-                                                         QPoint(x, y));
-                }
-            }
-            if (allLayerNamesMatch) {
-                anyMatch = true;
-                break;
-            }
-        }
-
-        if (anyMatch) {
-            // choose by chance which group of rule_layers should be used:
-            const int r = qrand() % mLayerList.size();
-            const RuleOutput &translationTable = mLayerList.at(r);
-
-            if (!mNoOverlappingRules) {
-                copyMapRegion(ruleOutputRegion, QPoint(x, y), translationTable);
-                ret = ret.united(rbr.translated(QPoint(x, y)));
-                continue;
-            }
-
-            bool missmatch = false;
-            const QList<Layer*> layers = translationTable.keys();
-
-            // check if there are no overlaps within this rule.
-            QVector<QRegion> ruleRegionInLayer;
-            for (int i = 0; i < layers.size(); ++i) {
-                Layer *layer = layers.at(i);
-
-                QRegion appliedPlace;
-
-                if (TileLayer *tileLayer = layer->asTileLayer())
-                    appliedPlace = tileLayer->region();
-                else if (ObjectGroup *objectGroup = layer->asObjectGroup())
-                    appliedPlace = tileRegionOfObjectGroup(objectGroup);
-                else
-                    continue;
-
-                ruleRegionInLayer.append(appliedPlace.intersected(ruleOutputRegion));
-
-                if (appliedRegions.at(i).intersects(ruleRegionInLayer.at(i).translated(x, y))) {
-                    missmatch = true;
-                    break;
-                }
-            }
-            if (missmatch)
-                continue;
-
-            copyMapRegion(ruleOutputRegion, QPoint(x, y), translationTable);
-            ret = ret.united(rbr.translated(QPoint(x, y)));
-            for (int i = 0; i < translationTable.size(); ++i)
-                appliedRegions[i] += ruleRegionInLayer[i].translated(x, y);
-        }
-    }
-
-    return ret;
-}
-
 /**
- * Returns a list of all cells which can be found within all tile layers
- * within the given region.
+ * Fills \a cells with the list of all cells which can be found within all
+ * tile layers within the given region.
  */
-static QVector<Cell> cellsInRegion(const QVector<TileLayer*> &list,
-                                   const QRegion &r)
+static void collectCellsInRegion(const QVector<InputLayer> &list,
+                                 const QRegion &r,
+                                 QVarLengthArray<Cell, 8> &cells)
 {
-    QVector<Cell> cells;
-    for (const TileLayer *tilelayer : list) {
+    for (const InputLayer &inputLayer : list) {
+#if QT_VERSION < 0x050800
         foreach (const QRect &rect, r.rects()) {
+#else
+        for (const QRect &rect : r) {
+#endif
             for (int x = rect.left(); x <= rect.right(); ++x) {
                 for (int y = rect.top(); y <= rect.bottom(); ++y) {
-                    const Cell &cell = tilelayer->cellAt(x, y);
+                    const Cell &cell = inputLayer.tileLayer->cellAt(x, y);
                     if (!cells.contains(cell))
                         cells.append(cell);
                 }
             }
         }
     }
-    return cells;
 }
 
 /**
@@ -662,108 +580,184 @@ static QVector<Cell> cellsInRegion(const QVector<TileLayer*> &list,
  *      (need of less layers.)
  *      It was not added to the case, when having only listNo layers to
  *      avoid total symmetry between those lists.
+ *      It can be turned off by setting the StrictEmpty property on the input
+ *      layer.
  *
  * If all positions are considered good, return true.
  * return false otherwise.
  *
  * @return bool, if the tile layer matches the given list of layers.
  */
-static bool compareLayerTo(const TileLayer *setLayer,
-                           const QVector<TileLayer*> &listYes,
-                           const QVector<TileLayer*> &listNo,
-                           const QRegion &ruleRegion, const QPoint &offset)
+static bool layerMatchesConditions(const TileLayer &setLayer,
+                                   const InputConditions &conditions,
+                                   const QRegion &ruleRegion,
+                                   const QPoint &offset)
 {
+    const auto &listYes = conditions.listYes;
+    const auto &listNo = conditions.listNo;
     if (listYes.isEmpty() && listNo.isEmpty())
         return false;
 
-    QVector<Cell> cells;
+    QVarLengthArray<Cell, 8> cells;
     if (listNo.isEmpty())
-        cells = cellsInRegion(listYes, ruleRegion);
+        collectCellsInRegion(listYes, ruleRegion, cells);
 
+#if QT_VERSION < 0x050800
     foreach (const QRect &rect, ruleRegion.rects()) {
+#else
+    for (const QRect &rect : ruleRegion) {
+#endif
         for (int x = rect.left(); x <= rect.right(); ++x) {
             for (int y = rect.top(); y <= rect.bottom(); ++y) {
-                // this is only used in the case where only one list has layers
-                // it is needed for the exception mentioned above
+                const Cell &setCell = setLayer.cellAt(x + offset.x(),
+                                                      y + offset.y());
+
+                // First check listNo. If any tile matches there, we can
+                // immediately know there is no match.
+                for (const InputLayer &inputNotLayer : listNo) {
+                    const Cell &noCell = inputNotLayer.tileLayer->cellAt(x, y);
+                    if ((inputNotLayer.strictEmpty || !noCell.isEmpty()) && setCell == noCell)
+                        return false;
+                }
+
+                // ruleDefinedListYes will be set when there is a tile in at
+                // least one layer. If there is a tile in at least one layer,
+                // only the given tiles in the different listYes layers are
+                // valid. Otherwise, consider all tiles not used elsewhere in
+                // the input as valid.
                 bool ruleDefinedListYes = false;
-
                 bool matchListYes = false;
-                bool matchListNo  = false;
 
-
-                if (!setLayer->contains(x + offset.x(), y + offset.y())) {
-                    foreach (const TileLayer *comparedTileLayer, listYes) {
-                        if (!comparedTileLayer->contains(x, y))
-                            return false;
-
-                        const Cell &c2 = comparedTileLayer->cellAt(x, y);
-                        if (!c2.isEmpty())
-                            return false;
-                    }
-                    continue;
-                }
-
-                const Cell &c1 = setLayer->cellAt(x + offset.x(),
-                                                  y + offset.y());
-
-                // ruleDefined will be set when there is a tile in at least
-                // one layer. if there is a tile in at least one layer, only
-                // the given tiles in the different listYes layers are valid.
-                // if there is given no tile at all in the listYes layers,
-                // consider all tiles valid.
-
-                for (const TileLayer *comparedTileLayer : listYes) {
-                    if (!comparedTileLayer->contains(x, y))
-                        return false;
-
-                    const Cell &c2 = comparedTileLayer->cellAt(x, y);
-                    if (!c2.isEmpty())
+                for (const InputLayer &inputLayer : listYes) {
+                    const Cell &yesCell = inputLayer.tileLayer->cellAt(x, y);
+                    if (inputLayer.strictEmpty || !yesCell.isEmpty()) {
                         ruleDefinedListYes = true;
-
-                    if (!c2.isEmpty() && c1 == c2)
-                        matchListYes = true;
+                        if (setCell == yesCell) {
+                            matchListYes = true;
+                            break;
+                        }
+                    }
                 }
-                for (const TileLayer *comparedTileLayer : listNo) {
-                    if (!comparedTileLayer->contains(x, y))
+
+                if (!ruleDefinedListYes) {
+                    // if there were only layers in the listYes, check the exception
+                    if (listNo.isEmpty() && cells.contains(setCell))
                         return false;
-
-                    const Cell &c2 = comparedTileLayer->cellAt(x, y);
-
-                    if (!c2.isEmpty() && c1 == c2)
-                        matchListNo = true;
-                }
-
-                // when there are only layers in the listNo
-                // check only if these layers are unmatched
-                // no need to check explicitly the exception in this case.
-                if (listYes.isEmpty()) {
-                    if (matchListNo)
-                        return false;
-                    else
-                        continue;
-                }
-                // when there are only layers in the listYes
-                // check if these layers are matched, or if the exception works
-                if (listNo.isEmpty()) {
-                    if (matchListYes)
-                        continue;
-                    if (!ruleDefinedListYes && !cells.contains(c1))
-                        continue;
+                } else if (!matchListYes) {
                     return false;
                 }
-
-                // there are layers in both lists:
-                // no need to consider ruleDefinedListXXX
-                if ((matchListYes || !ruleDefinedListYes) && !matchListNo)
-                    continue;
-                else
-                    return false;
             }
         }
     }
+
     return true;
 }
 
+QRect AutoMapper::applyRule(int ruleIndex, const QRect &where)
+{
+    QRect ret;
+
+    if (mLayerList.isEmpty())
+        return ret;
+
+    const QRegion &ruleInputRegion = mRulesInput.at(ruleIndex);
+    const QRegion &ruleOutputRegion = mRulesOutput.at(ruleIndex);
+    const QRect rbr = ruleInputRegion.boundingRect();
+
+    // Since the rule itself is translated, we need to adjust the borders of the
+    // loops. Decrease the size at all sides by one: There must be at least one
+    // tile overlap to the rule.
+    const int minX = where.left() - rbr.left() - rbr.width() + 1;
+    const int minY = where.top() - rbr.top() - rbr.height() + 1;
+
+    const int maxX = where.right() - rbr.left() + rbr.width() - 1;
+    const int maxY = where.bottom() - rbr.top() + rbr.height() - 1;
+
+    // In this list of regions it is stored which parts or the map have already
+    // been altered by exactly this rule. We store all the altered parts to
+    // make sure there are no overlaps of the same rule applied to
+    // (neighbouring) places
+    QVector<QRegion> appliedRegions;
+    if (mNoOverlappingRules)
+        appliedRegions.resize(mMapWork->layerCount());
+
+    const TileLayer dummy(QString(), 0, 0, 0, 0);
+
+    for (int y = minY; y <= maxY; ++y)
+    for (int x = minX; x <= maxX; ++x) {
+        bool anyMatch = false;
+
+        for (const InputIndex &inputIndex : qAsConst(mInputRules)) {
+            bool allLayerNamesMatch = true;
+
+            QMapIterator<QString, InputConditions> inputIndexIterator(inputIndex);
+            while (inputIndexIterator.hasNext()) {
+                inputIndexIterator.next();
+
+                const QString &name = inputIndexIterator.key();
+                const InputConditions &conditions = inputIndexIterator.value();
+
+                const int i = mMapWork->indexOfLayer(name, Layer::TileLayerType);
+                const TileLayer &setLayer = (i >= 0) ? *mMapWork->layerAt(i)->asTileLayer() : dummy;
+
+                if (!layerMatchesConditions(setLayer, conditions, ruleInputRegion, QPoint(x, y))) {
+                    allLayerNamesMatch = false;
+                    break;
+                }
+            }
+
+            if (allLayerNamesMatch) {
+                anyMatch = true;
+                break;
+            }
+        }
+
+        if (anyMatch) {
+            // choose by chance which group of rule_layers should be used:
+            const int r = qrand() % mLayerList.size();
+            const RuleOutput &translationTable = mLayerList.at(r);
+
+            if (mNoOverlappingRules) {
+                bool overlap = false;
+                const QList<Layer*> layers = translationTable.keys();
+
+                // check if there are no overlaps within this rule.
+                QVector<QRegion> ruleRegionInLayer;
+                for (int i = 0; i < layers.size(); ++i) {
+                    Layer *layer = layers.at(i);
+
+                    QRegion appliedPlace;
+
+                    if (TileLayer *tileLayer = layer->asTileLayer())
+                        appliedPlace = tileLayer->region();
+                    else if (ObjectGroup *objectGroup = layer->asObjectGroup())
+                        appliedPlace = tileRegionOfObjectGroup(objectGroup);
+                    else
+                        continue;
+
+                    ruleRegionInLayer.append(appliedPlace.intersected(ruleOutputRegion));
+
+                    if (appliedRegions.at(i).intersects(ruleRegionInLayer.at(i).translated(x, y))) {
+                        overlap = true;
+                        break;
+                    }
+                }
+
+                if (overlap)
+                    continue;
+
+                for (int i = 0; i < translationTable.size(); ++i)
+                    appliedRegions[i] += ruleRegionInLayer.at(i).translated(x, y);
+            }
+
+            copyMapRegion(ruleOutputRegion, QPoint(x, y), translationTable);
+            ret = ret.united(rbr.translated(QPoint(x, y)));
+        }
+    }
+
+    return ret;
+}
+
 void AutoMapper::copyMapRegion(const QRegion &region, QPoint offset,
                                const RuleOutput &layerTranslation)
 {
@@ -771,7 +765,11 @@ void AutoMapper::copyMapRegion(const QRegion &region, QPoint offset,
         Layer *from = it.key();
         Layer *to = mMapWork->layerAt(it.value());
 
+#if QT_VERSION < 0x050800
         foreach (const QRect &rect, region.rects()) {
+#else
+        for (const QRect &rect : region) {
+#endif
             if (TileLayer *fromTileLayer = from->asTileLayer()) {
                 TileLayer *toTileLayer = to->asTileLayer();
                 Q_ASSERT(toTileLayer); //TODO check this before in prepareAutomap or such!
@@ -809,11 +807,18 @@ void AutoMapper::copyTileRegion(const TileLayer *srcLayer, int srcX, int srcY,
                                 int width, int height,
                                 TileLayer *dstLayer, int dstX, int dstY)
 {
-    const int startX = qMax(dstX, 0);
-    const int startY = qMax(dstY, 0);
+    int startX = dstX;
+    int startY = dstY;
 
-    const int endX = qMin(dstX + width, dstLayer->width());
-    const int endY = qMin(dstY + height, dstLayer->height());
+    int endX = dstX + width;
+    int endY = dstY + height;
+
+    if (!mMapWork->infinite()) {
+        startX = qMax(0, startX);
+        startY = qMax(0, startY);
+        endX = qMin(dstLayer->width(), endX);
+        endY = qMin(dstLayer->height(), endY);
+    }
 
     const int offsetX = srcX - dstX;
     const int offsetY = srcY - dstY;
@@ -860,8 +865,7 @@ void AutoMapper::cleanTilesets()
 {
     QUndoStack *undoStack = mMapDocument->undoStack();
 
-    const auto &addedTilesets = mAddedTilesets;
-    for (const SharedTileset &tileset : addedTilesets) {
+    for (const SharedTileset &tileset : qAsConst(mAddedTilesets)) {
         if (mMapWork->isTilesetUsed(tileset.data()))
             continue;
 
@@ -879,8 +883,7 @@ void AutoMapper::cleanTileLayers()
 {
     QUndoStack *undoStack = mMapDocument->undoStack();
 
-    const auto &addedLayers = mAddedLayers;
-    for (Layer *layer : addedLayers) {
+    for (Layer *layer : qAsConst(mAddedLayers)) {
         if (!layer->isEmpty())
             continue;
 
@@ -897,9 +900,6 @@ void AutoMapper::cleanUpRulesMap()
 {
     cleanTilesets();
 
-    TilesetManager *tilesetManager = TilesetManager::instance();
-    tilesetManager->removeReferences(mMapRules->tilesets());
-
     delete mMapRules;
     mMapRules = nullptr;
 
