@@ -30,7 +30,7 @@ namespace Internal {
 
 ChangeTileImageSource::ChangeTileImageSource(TilesetDocument *tilesetDocument,
                                              Tile *tile,
-                                             const QString &imageSource)
+                                             const QUrl &imageSource)
     : mTilesetDocument(tilesetDocument)
     , mTile(tile)
     , mOldImageSource(tile->imageSource())
@@ -40,10 +40,11 @@ ChangeTileImageSource::ChangeTileImageSource(TilesetDocument *tilesetDocument,
                                         "Change Tile Image"));
 }
 
-void ChangeTileImageSource::apply(const QString &imageSource)
+void ChangeTileImageSource::apply(const QUrl &imageSource)
 {
+    // todo: make sure remote source loading is triggered
     mTilesetDocument->setTileImage(mTile,
-                                   QPixmap(imageSource),
+                                   QPixmap(imageSource.toLocalFile()),
                                    imageSource);
 }
 
