@@ -43,17 +43,14 @@
 #include "utils.h"
 #include "zoomable.h"
 
-#include <QApplication>
-#include <QClipboard>
+#include <QCoreApplication>
 #include <QDialogButtonBox>
-#include <QDir>
 #include <QFileDialog>
 #include <QFileInfo>
 #include <QHBoxLayout>
 #include <QLabel>
 #include <QMenu>
 #include <QMessageBox>
-#include <QProcess>
 #include <QScrollBar>
 #include <QStackedLayout>
 #include <QTabBar>
@@ -65,39 +62,6 @@
 using namespace Tiled;
 using namespace Tiled::Internal;
 
-/*
- * Code based on FileUtils::showInGraphicalShell from Qt Creator
- * Copyright (C) 2016 The Qt Company Ltd.
- * Used under the terms of the GNU General Public License version 3
- */
-static void showInFileManager(const QString &fileName)
-{
-    // Mac, Windows support folder or file.
-#if defined(Q_OS_WIN)
-    QStringList param;
-    if (!QFileInfo(fileName).isDir())
-        param += QLatin1String("/select,");
-    param += QDir::toNativeSeparators(fileName);
-    QProcess::startDetached(QLatin1String("explorer.exe"), param);
-#elif defined(Q_OS_MAC)
-    QStringList scriptArgs;
-    scriptArgs << QLatin1String("-e")
-               << QString::fromLatin1("tell application \"Finder\" to reveal POSIX file \"%1\"")
-                                     .arg(fileName);
-    QProcess::execute(QLatin1String("/usr/bin/osascript"), scriptArgs);
-    scriptArgs.clear();
-    scriptArgs << QLatin1String("-e")
-               << QLatin1String("tell application \"Finder\" to activate");
-    QProcess::execute(QLatin1String("/usr/bin/osascript"), scriptArgs);
-#else
-    // We cannot select a file here, because xdg-open would open the file
-    // instead of the file browser...
-    QProcess::startDetached(QString(QLatin1String("xdg-open \"%1\""))
-                            .arg(QFileInfo(fileName).absolutePath()));
-#endif
-}
-
-
 
 DocumentManager *DocumentManager::mInstance;
 
@@ -126,6 +90,7 @@ DocumentManager::DocumentManager(QObject *parent)
     , mMapEditor(nullptr) // todo: look into removing this
     , mUndoGroup(new QUndoGroup(this))
     , mFileSystemWatcher(new FileSystemWatcher(this))
+    , mMultiDocumentClose(false)
 {
     mBrokenLinksWidget->setVisible(false);
 
@@ -157,8 +122,8 @@ DocumentManager::DocumentManager(QObject *parent)
             this, &DocumentManager::documentCloseRequested);
     connect(mTabBar, &QTabBar::tabMoved,
             this, &DocumentManager::documentTabMoved);
-    connect(mTabBar, SIGNAL(customContextMenuRequested(QPoint)),
-            SLOT(tabContextMenuRequested(QPoint)));
+    connect(mTabBar, &QWidget::customContextMenuRequested,
+            this, &DocumentManager::tabContextMenuRequested);
 
     connect(mFileSystemWatcher, &FileSystemWatcher::fileChanged,
             this, &DocumentManager::fileChanged);
@@ -168,10 +133,14 @@ DocumentManager::DocumentManager(QObject *parent)
 
     connect(TilesetManager::instance(), &TilesetManager::tilesetImagesChanged,
             this, &DocumentManager::tilesetImagesChanged);
+
+    mTabBar->installEventFilter(this);
 }
 
 DocumentManager::~DocumentManager()
 {
+    mTabBar->removeEventFilter(this);
+
     // All documents should be closed gracefully beforehand
     Q_ASSERT(mDocuments.isEmpty());
     Q_ASSERT(mTilesetDocumentsModel->rowCount() == 0);
@@ -457,30 +426,31 @@ bool DocumentManager::saveDocumentAs(Document *document)
             fileName += defaultFileName;
         }
 
-        fileName = QFileDialog::getSaveFileName(mWidget->window(), QString(),
-                                                fileName,
-                                                filter,
-                                                &selectedFilter);
+        while (true) {
+            fileName = QFileDialog::getSaveFileName(mWidget->window(), QString(),
+                                                    fileName,
+                                                    filter,
+                                                    &selectedFilter);
 
-        if (!fileName.isEmpty() &&
-            !Utils::fileNameMatchesNameFilter(QFileInfo(fileName).fileName(), selectedFilter))
-        {
-            QMessageBox messageBox(QMessageBox::Warning,
-                                   QCoreApplication::translate("Tiled::Internal::MainWindow", "Extension Mismatch"),
-                                   QCoreApplication::translate("Tiled::Internal::MainWindow", "The file extension does not match the chosen file type."),
-                                   QMessageBox::Yes | QMessageBox::No,
-                                   mWidget->window());
+            if (!fileName.isEmpty() &&
+                !Utils::fileNameMatchesNameFilter(QFileInfo(fileName).fileName(), selectedFilter))
+            {
+                QMessageBox messageBox(QMessageBox::Warning,
+                                       QCoreApplication::translate("Tiled::Internal::MainWindow", "Extension Mismatch"),
+                                       QCoreApplication::translate("Tiled::Internal::MainWindow", "The file extension does not match the chosen file type."),
+                                       QMessageBox::Yes | QMessageBox::No,
+                                       mWidget->window());
 
-            messageBox.setInformativeText(QCoreApplication::translate("Tiled::Internal::MainWindow",
-                                                                      "Tiled may not automatically recognize your file when loading. "
-                                                                      "Are you sure you want to save with this extension?"));
+                messageBox.setInformativeText(QCoreApplication::translate("Tiled::Internal::MainWindow",
+                                                                          "Tiled may not automatically recognize your file when loading. "
+                                                                          "Are you sure you want to save with this extension?"));
 
-            int answer = messageBox.exec();
-            if (answer != QMessageBox::Yes)
-                return QString();
+                int answer = messageBox.exec();
+                if (answer != QMessageBox::Yes)
+                    continue;
+            }
+            return fileName;
         }
-
-        return fileName;
     };
 
     if (auto mapDocument = qobject_cast<MapDocument*>(document)) {
@@ -490,12 +460,16 @@ bool DocumentManager::saveDocumentAs(Document *document)
         FormatHelper<MapFormat> helper(FileFormat::ReadWrite);
         filter = helper.filter();
 
-        fileName = getSaveFileName(QCoreApplication::translate("Tiled::Internal::MainWindow", "untitled.tmx"));
+        auto suggestedFileName = QCoreApplication::translate("Tiled::Internal::MainWindow", "untitled");
+        suggestedFileName.append(QLatin1String(".tmx"));
+
+        fileName = getSaveFileName(suggestedFileName);
         if (fileName.isEmpty())
             return false;
 
         MapFormat *format = helper.formatByNameFilter(selectedFilter);
         mapDocument->setWriterFormat(format);
+        mapDocument->setReaderFormat(format);
 
     } else if (auto tilesetDocument = qobject_cast<TilesetDocument*>(document)) {
         if (selectedFilter.isEmpty())
@@ -504,7 +478,12 @@ bool DocumentManager::saveDocumentAs(Document *document)
         FormatHelper<TilesetFormat> helper(FileFormat::ReadWrite);
         filter = helper.filter();
 
-        fileName = getSaveFileName(QCoreApplication::translate("Tiled::Internal::MainWindow", "untitled.tsx"));
+        auto suggestedFileName = tilesetDocument->tileset()->name().trimmed();
+        if (suggestedFileName.isEmpty())
+            suggestedFileName = QCoreApplication::translate("Tiled::Internal::MainWindow", "untitled");
+        suggestedFileName.append(QLatin1String(".tsx"));
+
+        fileName = getSaveFileName(suggestedFileName);
         if (fileName.isEmpty())
             return false;
 
@@ -524,6 +503,37 @@ void DocumentManager::closeCurrentDocument()
     closeDocumentAt(index);
 }
 
+void DocumentManager::closeOtherDocuments(int index)
+{
+    if (index == -1)
+        return;
+
+    mMultiDocumentClose = true;
+
+    for (int i = mTabBar->count() - 1; i >= 0; --i) {
+        if (i != index)
+            documentCloseRequested(i);
+
+        if (!mMultiDocumentClose)
+            return;
+    }
+}
+
+void DocumentManager::closeDocumentsToRight(int index)
+{
+    if (index == -1)
+        return;
+
+    mMultiDocumentClose = true;
+
+    for (int i = mTabBar->count() - 1; i > index; --i) {
+        documentCloseRequested(i);
+
+        if (!mMultiDocumentClose)
+            return;
+    }
+}
+
 void DocumentManager::closeDocumentAt(int index)
 {
     Document *document = mDocuments.at(index);
@@ -703,17 +713,25 @@ void DocumentManager::tabContextMenuRequested(const QPoint &pos)
 
     QMenu menu(mTabBar->window());
 
-    QString fileName = mDocuments.at(index)->fileName();
+    Utils::addFileManagerActions(menu, mDocuments.at(index)->fileName());
 
-    QAction *copyPath = menu.addAction(tr("Copy File Path"));
-    connect(copyPath, &QAction::triggered, [fileName] {
-        QClipboard *clipboard = QApplication::clipboard();
-        clipboard->setText(QDir::toNativeSeparators(fileName));
+    menu.addSeparator();
+
+    QAction *closeTab = menu.addAction(tr("Close"));
+    closeTab->setIcon(QIcon(QStringLiteral(":/images/16x16/window-close.png")));
+    Utils::setThemeIcon(closeTab, "window-close");
+    connect(closeTab, &QAction::triggered, [this, index] {
+        documentCloseRequested(index);
     });
 
-    QAction *openFolder = menu.addAction(tr("Open Containing Folder..."));
-    connect(openFolder, &QAction::triggered, [fileName] {
-        showInFileManager(fileName);
+    QAction *closeOtherTabs = menu.addAction(tr("Close Other Tabs"));
+    connect(closeOtherTabs, &QAction::triggered, [this, index] {
+        closeOtherDocuments(index);
+    });
+
+    QAction *closeTabsToRight = menu.addAction(tr("Close Tabs to the Right"));
+    connect(closeTabsToRight, &QAction::triggered, [this, index] {
+        closeDocumentsToRight(index);
     });
 
     menu.exec(mTabBar->mapToGlobal(pos));
@@ -816,7 +834,8 @@ TilesetDocument *DocumentManager::findTilesetDocument(const QString &fileName) c
 void DocumentManager::openTileset(const SharedTileset &tileset)
 {
     auto tilesetDocument = findTilesetDocument(tileset);
-    Q_ASSERT(tilesetDocument);
+    if (!tilesetDocument)
+        tilesetDocument = new TilesetDocument(tileset);
 
     if (!switchToDocument(tilesetDocument))
         addDocument(tilesetDocument);
@@ -867,7 +886,7 @@ static bool mayNeedColumnCountAdjustment(const Tileset &tileset)
 {
     if (tileset.isCollection())
         return false;
-    if (!tileset.imageLoaded())
+    if (tileset.imageStatus() != LoadingReady)
         return false;
     if (tileset.columnCount() == tileset.expectedColumnCount())
         return false;
@@ -967,3 +986,27 @@ bool DocumentManager::askForAdjustment(const Tileset &tileset)
 
     return r == QMessageBox::Yes;
 }
+
+bool DocumentManager::eventFilter(QObject *object, QEvent *event)
+{
+    if (object == mTabBar && event->type() == QEvent::MouseButtonRelease) {
+        // middle-click tab closing
+        QMouseEvent *mouseEvent = static_cast<QMouseEvent*>(event);
+
+        if (mouseEvent->button() == Qt::MidButton) {
+            int index = mTabBar->tabAt(mouseEvent->pos());
+
+            if (index != -1) {
+                documentCloseRequested(index);
+                return true;
+            }
+        }
+    }
+
+    return false;
+}
+
+void DocumentManager::abortMultiDocumentClose()
+{
+    mMultiDocumentClose = false;
+}
