@@ -38,6 +38,7 @@ class QTabBar;
 namespace Tiled {
 
 class FileSystemWatcher;
+class ObjectTemplate;
 
 namespace Internal {
 
@@ -135,6 +136,16 @@ public:
      */
     void closeCurrentDocument();
 
+    /**
+     * Closes all documents except the one pointed to by index.
+     */
+    void closeOtherDocuments(int index);
+
+    /**
+     * Closes all documents whose tabs are to the right of the index.
+     */
+    void closeDocumentsToRight(int index);
+
     /**
      * Closes the document at the given \a index. Will not ask the user whether
      * to save any changes!
@@ -187,10 +198,18 @@ public:
     void centerMapViewOn(const QPointF &pos)
     { centerMapViewOn(pos.x(), pos.y()); }
 
+    /**
+     * Unsets a flag to stop closeOtherDocuments() and closeDocumentsToRight()
+     * when Cancel is pressed
+     */
+    void abortMultiDocumentClose();
+
 signals:
     void fileOpenRequested();
     void fileOpenRequested(const QString &path);
     void fileSaveRequested();
+    void templateOpenRequested(const QString &path);
+    void templateTilesetReplaced();
 
     /**
      * Emitted when the current displayed map document changed.
@@ -253,6 +272,8 @@ private:
     void addToTilesetDocument(const SharedTileset &tileset, MapDocument *mapDocument);
     void removeFromTilesetDocument(const SharedTileset &tileset, MapDocument *mapDocument);
 
+    bool eventFilter(QObject *object, QEvent *event) override;
+
     QList<Document*> mDocuments;
     TilesetDocumentsModel *mTilesetDocumentsModel;
 
@@ -274,6 +295,8 @@ private:
     QMap<SharedTileset, TilesetDocument*> mTilesetToDocument;
 
     static DocumentManager *mInstance;
+
+    bool mMultiDocumentClose;
 };
 
 inline TilesetDocumentsModel *DocumentManager::tilesetDocumentsModel() const
