@@ -57,8 +57,8 @@
 #include "patreondialog.h"
 #include "pluginmanager.h"
 #include "resizedialog.h"
+#include "templatemanager.h"
 #include "terrain.h"
-#include "tileanimationeditor.h"
 #include "tile.h"
 #include "tilelayer.h"
 #include "tilesetdocument.h"
@@ -90,7 +90,7 @@
 #include <QUndoStack>
 #include <QUndoView>
 
-#if defined(Q_OS_WIN) && QT_VERSION >= 0x050600
+#ifdef Q_OS_WIN
 #include <QtPlatformHeaders\QWindowsWindowFunctions>
 #endif
 
@@ -99,6 +99,95 @@ using namespace Tiled::Internal;
 using namespace Tiled::Utils;
 
 
+namespace {
+
+template <typename Format>
+struct ExportDetails
+{
+    Format *mFormat = nullptr;
+    QString mFileName;
+
+    ExportDetails() {}
+    ExportDetails(Format *format, const QString& fileName)
+        : mFormat(format)
+        , mFileName(fileName)
+    {}
+
+    bool isValid() const { return mFormat != nullptr; }
+};
+
+template <typename Format>
+ExportDetails<Format> chooseExportDetails(const QString &fileName,
+                                          const QString &lastExportName,
+                                          const QString &lastExportFilter,
+                                          QWidget* window)
+{
+    FormatHelper<Format> helper(FileFormat::Write, MainWindow::tr("All Files (*)"));
+
+    Preferences *pref = Preferences::instance();
+
+    QString selectedFilter = lastExportFilter;
+    QString suggestedFilename = lastExportName;
+
+    if (suggestedFilename.isEmpty()) {
+        QFileInfo baseNameInfo = QFileInfo(fileName);
+        QString baseName = baseNameInfo.baseName();
+
+        QRegExp extensionFinder(QLatin1String("\\(\\*\\.([^\\)\\s]*)"));
+        extensionFinder.indexIn(selectedFilter);
+        const QString extension = extensionFinder.cap(1);
+
+        QString lastExportedFilePath = pref->lastPath(Preferences::ExportedFile);
+
+        suggestedFilename = lastExportedFilePath
+                            + QLatin1String("/") + baseName
+                            + QLatin1Char('.') + extension;
+    }
+
+    // No need to confirm overwrite here since it'll be prompted below
+    QString exportToFileName = QFileDialog::getSaveFileName(window, MainWindow::tr("Export As..."),
+                                                    suggestedFilename,
+                                                    helper.filter(),
+                                                    &selectedFilter,
+                                                    QFileDialog::DontConfirmOverwrite);
+    if (exportToFileName.isEmpty())
+        return ExportDetails<Format>();
+
+    // If a specific filter was selected, use that format
+    Format *chosenFormat = helper.formatByNameFilter(selectedFilter);
+
+    // If not, try to find the file extension among the name filters
+    QString suffix = QFileInfo(exportToFileName).completeSuffix();
+    if (!chosenFormat && !suffix.isEmpty()) {
+        suffix.prepend(QLatin1String("*."));
+
+        for (Format *format : helper.formats()) {
+            if (format->nameFilter().contains(suffix, Qt::CaseInsensitive)) {
+                if (chosenFormat) {
+                    QMessageBox::warning(window, MainWindow::tr("Non-unique file extension"),
+                                         MainWindow::tr("Non-unique file extension.\n"
+                                                    "Please select specific format."));
+                    return chooseExportDetails<Format>(exportToFileName, lastExportName, lastExportFilter, window);
+                } else {
+                    chosenFormat = format;
+                }
+            }
+        }
+    }
+
+    if (!chosenFormat) {
+        QMessageBox::critical(window, MainWindow::tr("Unknown File Format"),
+                              MainWindow::tr("The given filename does not have any known "
+                                         "file extension."));
+        return ExportDetails<Format>();
+    }
+
+    return ExportDetails<Format>(chosenFormat, exportToFileName);
+}
+
+} // namespace
+
+
 MainWindow::MainWindow(QWidget *parent, Qt::WindowFlags flags)
     : QMainWindow(parent, flags)
     , mActionManager(new ActionManager)
@@ -108,14 +197,9 @@ MainWindow::MainWindow(QWidget *parent, Qt::WindowFlags flags)
     , mObjectTypesEditor(new ObjectTypesEditor(this))
     , mAutomappingManager(new AutomappingManager(this))
     , mDocumentManager(DocumentManager::instance())
-    , mTmxMapFormat(new TmxMapFormat(this))
-    , mTsxTilesetFormat(new TsxTilesetFormat(this))
 {
     mUi->setupUi(this);
 
-    PluginManager::addObject(mTmxMapFormat);
-    PluginManager::addObject(mTsxTilesetFormat);
-
     ActionManager::registerAction(mUi->actionNewMap, "file.new_map");
     ActionManager::registerAction(mUi->actionNewTileset, "file.new_tileset");
 
@@ -253,6 +337,9 @@ MainWindow::MainWindow(QWidget *parent, Qt::WindowFlags flags)
     mUi->menuMap->insertAction(mUi->actionOffsetMap,
                                mActionHandler->actionCropToSelection());
 
+    mUi->menuMap->insertAction(mUi->actionOffsetMap,
+                               mActionHandler->actionAutocrop());
+
     mLayerMenu = new QMenu(tr("&Layer"), this);
     mNewLayerMenu = mActionHandler->createNewLayerMenu(mLayerMenu);
     mGroupLayerMenu = mActionHandler->createGroupLayerMenu(mLayerMenu);
@@ -376,15 +463,16 @@ MainWindow::MainWindow(QWidget *parent, Qt::WindowFlags flags)
     mViewsAndToolbarsMenu = new QMenu(this);
     mViewsAndToolbarsAction = new QAction(tr("Views and Toolbars"), this);
     mViewsAndToolbarsAction->setMenu(mViewsAndToolbarsMenu);
+
+    mResetToDefaultLayout = new QAction(tr("Reset to Default Layout"), this);
+
     mShowObjectTypesEditor = new QAction(tr("Object Types Editor"), this);
     mShowObjectTypesEditor->setCheckable(true);
     mUi->menuView->insertAction(mUi->actionShowGrid, mViewsAndToolbarsAction);
     mUi->menuView->insertAction(mUi->actionShowGrid, mShowObjectTypesEditor);
     mUi->menuView->insertSeparator(mUi->actionShowGrid);
 
-    mShowTileAnimationEditor = new QAction(tr("Tile Animation Editor"), this);
-    mShowTileAnimationEditor->setCheckable(true);
-    mUi->menuTileset->insertAction(mUi->actionTilesetProperties, mShowTileAnimationEditor);
+    mUi->menuTileset->insertAction(mUi->actionTilesetProperties, tilesetEditor->showAnimationEditor());
     mUi->menuTileset->insertAction(mUi->actionTilesetProperties, tilesetEditor->editCollisionAction());
     mUi->menuTileset->insertAction(mUi->actionTilesetProperties, tilesetEditor->editTerrainAction());
     mUi->menuTileset->insertSeparator(mUi->actionTilesetProperties);
@@ -399,11 +487,6 @@ MainWindow::MainWindow(QWidget *parent, Qt::WindowFlags flags)
             mObjectTypesEditor, SLOT(setVisible(bool)));
     connect(mObjectTypesEditor, SIGNAL(closed()), SLOT(onObjectTypesEditorClosed()));
 
-    connect(mShowTileAnimationEditor, &QAction::toggled,
-            tilesetEditor->tileAnimationEditor(), &TileAnimationEditor::setVisible);
-    connect(tilesetEditor->tileAnimationEditor(), &TileAnimationEditor::closed,
-            this, &MainWindow::onAnimationEditorClosed);
-
     connect(ClipboardManager::instance(), SIGNAL(hasMapChanged()), SLOT(updateActions()));
 
     connect(mDocumentManager, SIGNAL(fileOpenRequested(QString)),
@@ -419,6 +502,8 @@ MainWindow::MainWindow(QWidget *parent, Qt::WindowFlags flags)
     connect(mDocumentManager, SIGNAL(reloadError(QString)),
             this, SLOT(reloadError(QString)));
 
+    connect(mResetToDefaultLayout, &QAction::triggered, this, &MainWindow::resetToDefaultLayout);
+
     QShortcut *switchToLeftDocument = new QShortcut(tr("Alt+Left"), this);
     connect(switchToLeftDocument, SIGNAL(activated()),
             mDocumentManager, SLOT(switchToLeftDocument()));
@@ -448,7 +533,7 @@ MainWindow::MainWindow(QWidget *parent, Qt::WindowFlags flags)
     connect(mAutomappingManager, SIGNAL(errorsOccurred(bool)),
             this, SLOT(autoMappingError(bool)));
 
-#if defined(Q_OS_WIN) && QT_VERSION >= 0x050600
+#ifdef Q_OS_WIN
     connect(preferences, &Preferences::useOpenGLChanged, this, &MainWindow::ensureHasBorderInFullScreen);
 #endif
 
@@ -470,10 +555,8 @@ MainWindow::~MainWindow()
     mDocumentManager->deleteEditor(Document::MapDocumentType);
     mDocumentManager->deleteEditor(Document::TilesetDocumentType);
 
-    PluginManager::removeObject(mTmxMapFormat);
-    PluginManager::removeObject(mTsxTilesetFormat);
-
     DocumentManager::deleteInstance();
+    TemplateManager::deleteInstance();
     TilesetManager::deleteInstance();
     Preferences::deleteInstance();
     LanguageManager::deleteInstance();
@@ -676,7 +759,7 @@ void MainWindow::openFile()
                                      selectedFilter).toString();
 
     auto preferences = Preferences::instance();
-    const auto fileNames = QFileDialog::getOpenFileNames(this, tr("Open Map"),
+    const auto fileNames = QFileDialog::getOpenFileNames(this, tr("Open File"),
                                                          preferences->fileDialogStartLocation(),
                                                          helper.filter(),
                                                          &selectedFilter);
@@ -778,6 +861,7 @@ bool MainWindow::confirmSave(Document *document)
     case QMessageBox::Discard: return true;
     case QMessageBox::Cancel:
     default:
+        mDocumentManager->abortMultiDocumentClose();
         return false;
     }
 }
@@ -825,115 +909,10 @@ void MainWindow::export_()
 
 void MainWindow::exportAs()
 {
-    auto mapDocument = qobject_cast<MapDocument*>(mDocument);
-    if (!mapDocument)
-        return;
-
-    FormatHelper<MapFormat> helper(FileFormat::Write, tr("All Files (*)"));
-
-    Preferences *pref = Preferences::instance();
-
-    QString selectedFilter =
-            mSettings.value(QLatin1String("lastUsedExportFilter")).toString();
-    QString suggestedFilename = mapDocument->lastExportFileName();
-
-    if (suggestedFilename.isEmpty()) {
-        QFileInfo baseNameInfo = QFileInfo(mapDocument->fileName());
-        QString baseName = baseNameInfo.baseName();
-
-        QRegExp extensionFinder(QLatin1String("\\(\\*\\.([^\\)\\s]*)"));
-        extensionFinder.indexIn(selectedFilter);
-        const QString extension = extensionFinder.cap(1);
-
-        QString lastExportedFilePath = pref->lastPath(Preferences::ExportedFile);
-
-        suggestedFilename = lastExportedFilePath
-                + QLatin1String("/") + baseName
-                + QLatin1Char('.') + extension;
-    }
-
-    // No need to confirm overwrite here since it'll be prompted below
-    QString fileName = QFileDialog::getSaveFileName(this, tr("Export As..."),
-                                                    suggestedFilename,
-                                                    helper.filter(),
-                                                    &selectedFilter,
-                                                    QFileDialog::DontConfirmOverwrite);
-    if (fileName.isEmpty())
-        return;
-
-    // If a specific filter was selected, use that format
-    MapFormat *chosenFormat = helper.formatByNameFilter(selectedFilter);
-
-    // If not, try to find the file extension among the name filters
-    QString suffix = QFileInfo(fileName).completeSuffix();
-    if (!chosenFormat && !suffix.isEmpty()) {
-        suffix.prepend(QLatin1String("*."));
-
-        for (MapFormat *format : helper.formats()) {
-            if (format->nameFilter().contains(suffix, Qt::CaseInsensitive)) {
-                if (chosenFormat) {
-                    QMessageBox::warning(this, tr("Non-unique file extension"),
-                                         tr("Non-unique file extension.\n"
-                                            "Please select specific format."));
-                    return exportAs();
-                } else {
-                    chosenFormat = format;
-                }
-            }
-        }
-    }
-
-    if (!chosenFormat) {
-        QMessageBox::critical(this, tr("Unknown File Format"),
-                              tr("The given filename does not have any known "
-                                 "file extension."));
-        return;
-    }
-
-    // Check if writer will overwrite existing files here because some writers
-    // could save to multiple files at the same time. For example CSV saves
-    // each layer into a separate file.
-    QStringList outputFiles = chosenFormat->outputFiles(mapDocument->map(),
-                                                        fileName);
-    if (outputFiles.size() > 0) {
-        // Check if any output file already exists
-        QString message =
-                tr("Some export files already exist:") + QLatin1String("\n\n");
-
-        bool overwriteHappens = false;
-
-        for (const QString &outputFile : outputFiles) {
-            if (QFile::exists(outputFile)) {
-                overwriteHappens = true;
-                message += outputFile + QLatin1Char('\n');
-            }
-        }
-        message += QLatin1Char('\n') + tr("Do you want to replace them?");
-
-        // If overwrite happens, warn the user and get confirmation before exporting
-        if (overwriteHappens) {
-            const QMessageBox::StandardButton reply = QMessageBox::warning(
-                this,
-                tr("Overwrite Files"),
-                message,
-                QMessageBox::Yes | QMessageBox::No,
-                QMessageBox::No);
-
-            if (reply != QMessageBox::Yes)
-                return;
-        }
-    }
-
-    pref->setLastPath(Preferences::ExportedFile, QFileInfo(fileName).path());
-    mSettings.setValue(QLatin1String("lastUsedExportFilter"), selectedFilter);
-
-    if (!chosenFormat->write(mapDocument->map(), fileName)) {
-        QMessageBox::critical(this, tr("Error Exporting Map"),
-                              chosenFormat->errorString());
-    } else {
-        // Remember export parameters, so subsequent exports can be done faster
-        mapDocument->setLastExportFileName(fileName);
-        mapDocument->setExportFormat(chosenFormat);
+    if (auto mapDocument = qobject_cast<MapDocument*>(mDocument)) {
+        exportMapAs(mapDocument);
+    } else if (auto tilesetDocument = qobject_cast<TilesetDocument*>(mDocument)) {
+        exportTilesetAs(tilesetDocument);
     }
 }
 
@@ -1099,7 +1078,8 @@ bool MainWindow::newTileset(const QString &path)
     if (!tileset)
         return false;
 
-    prefs->setLastPath(Preferences::ImageFile, tileset->imageSource());
+    if (tileset->imageSource().isLocalFile())
+        prefs->setLastPath(Preferences::ImageFile, tileset->imageSource().toLocalFile());
 
     auto mapDocument = qobject_cast<MapDocument*>(mDocument);
 
@@ -1124,9 +1104,9 @@ void MainWindow::reloadTilesetImages()
         Map *map = mapDocument->map();
         const auto tilesets = map->tilesets();
         for (const SharedTileset &tileset : tilesets)
-            tilesetManager->reloadImages(tileset);
+            tilesetManager->reloadImages(tileset.data());
     } else if (auto tilesetDocument = qobject_cast<TilesetDocument*>(mDocument)) {
-        tilesetManager->reloadImages(tilesetDocument->tileset());
+        tilesetManager->reloadImages(tilesetDocument->tileset().data());
     }
 }
 
@@ -1174,26 +1154,45 @@ void MainWindow::resizeMap()
 
     Map *map = mapDocument->map();
 
+    QSize mapSize(map->size());
+    QPoint mapStart(0, 0);
+
+    if (map->infinite()) {
+        QRect mapBounds;
+
+        LayerIterator iterator(map);
+        while (Layer *layer = iterator.next()) {
+            if (TileLayer *tileLayer = dynamic_cast<TileLayer*>(layer))
+                mapBounds = mapBounds.united(tileLayer->bounds());
+        }
+
+        if (!mapBounds.isEmpty()) {
+            mapSize = mapBounds.size();
+            mapStart = mapBounds.topLeft();
+        }
+    }
+
     ResizeDialog resizeDialog(this);
-    resizeDialog.setOldSize(map->size());
+    resizeDialog.setOldSize(mapSize);
 
     // TODO: Look into fixing up the preview for maps that do not use square
     // tiles, and possibly also staggered maps.
     if (map->orientation() == Map::Orthogonal && map->tileWidth() == map->tileHeight()) {
         resizeDialog.setMiniMapRenderer([mapDocument](QSize size){
             QImage image(size, QImage::Format_ARGB32_Premultiplied);
-            MiniMapRenderer(mapDocument).renderToImage(image, MiniMapRenderer::DrawObjects
-                                                       | MiniMapRenderer::DrawImages
-                                                       | MiniMapRenderer::DrawTiles
-                                                       | MiniMapRenderer::IgnoreInvisibleLayer);
+            MiniMapRenderer(mapDocument->map()).renderToImage(image, MiniMapRenderer::DrawMapObjects
+                                                              | MiniMapRenderer::DrawImageLayers
+                                                              | MiniMapRenderer::DrawTileLayers
+                                                              | MiniMapRenderer::IgnoreInvisibleLayer
+                                                              | MiniMapRenderer::SmoothPixmapTransform);
             return image;
         });
     }
 
     if (resizeDialog.exec()) {
         const QSize &newSize = resizeDialog.newSize();
-        const QPoint &offset = resizeDialog.offset();
-        if (newSize != map->size() || !offset.isNull())
+        const QPoint &offset = resizeDialog.offset() - mapStart;
+        if (newSize != mapSize || !offset.isNull())
             mapDocument->resizeMap(newSize, offset, resizeDialog.removeObjects());
     }
 }
@@ -1269,14 +1268,9 @@ void MainWindow::onObjectTypesEditorClosed()
     mShowObjectTypesEditor->setChecked(false);
 }
 
-void MainWindow::onAnimationEditorClosed()
-{
-    mShowTileAnimationEditor->setChecked(false);
-}
-
 void MainWindow::ensureHasBorderInFullScreen()
 {
-#if defined(Q_OS_WIN) && QT_VERSION >= 0x050600
+#ifdef Q_OS_WIN
     // Workaround issue #1576
     static bool hasBorderInFullScreen = false;
 
@@ -1325,6 +1319,19 @@ void MainWindow::updateRecentFilesMenu()
     mUi->menuRecentFiles->setEnabled(numRecentFiles > 0);
 }
 
+void MainWindow::resetToDefaultLayout()
+{
+    // Make sure we're not in Clear View mode
+    mUi->actionClearView->setChecked(false);
+
+    // Reset the Console dock
+    addDockWidget(Qt::BottomDockWidgetArea, mConsoleDock);
+    mConsoleDock->setVisible(false);
+
+    // Reset the layout of the current editor
+    mDocumentManager->currentEditor()->resetLayout();
+}
+
 void MainWindow::updateViewsAndToolbarsMenu()
 {
     mViewsAndToolbarsMenu->clear();
@@ -1333,16 +1340,17 @@ void MainWindow::updateViewsAndToolbarsMenu()
 
     if (Editor *editor = mDocumentManager->currentEditor()) {
         mViewsAndToolbarsMenu->addSeparator();
-
         const auto dockWidgets = editor->dockWidgets();
         for (auto dockWidget : dockWidgets)
             mViewsAndToolbarsMenu->addAction(dockWidget->toggleViewAction());
 
         mViewsAndToolbarsMenu->addSeparator();
-
         const auto toolBars = editor->toolBars();
         for (auto toolBar : toolBars)
             mViewsAndToolbarsMenu->addAction(toolBar->toggleViewAction());
+
+        mViewsAndToolbarsMenu->addSeparator();
+        mViewsAndToolbarsMenu->addAction(mResetToDefaultLayout);
     }
 }
 
@@ -1363,7 +1371,7 @@ void MainWindow::updateActions()
 
     mUi->actionExportAsImage->setEnabled(mapDocument);
     mUi->actionExport->setEnabled(mapDocument);
-    mUi->actionExportAs->setEnabled(mapDocument);
+    mUi->actionExportAs->setEnabled(mapDocument || tilesetDocument);
     mUi->actionReload->setEnabled(mapDocument || (tilesetDocument && tilesetDocument->canReload()));
     mUi->actionClose->setEnabled(document);
     mUi->actionCloseAll->setEnabled(document);
@@ -1502,8 +1510,98 @@ void MainWindow::retranslateUi()
     mNewLayerMenu->setTitle(tr("&New"));
     mGroupLayerMenu->setTitle(tr("&Group"));
     mViewsAndToolbarsAction->setText(tr("Views and Toolbars"));
-    mShowTileAnimationEditor->setText(tr("Tile Animation Editor"));
     mActionHandler->retranslateUi();
+    CommandManager::instance()->retranslateUi();
+}
+
+void MainWindow::exportMapAs(MapDocument *mapDocument)
+{
+    QString fileName = mapDocument->fileName();
+    QString selectedFilter =
+            mSettings.value(QLatin1String("lastUsedExportFilter")).toString();
+    auto exportDetails = chooseExportDetails<MapFormat>(fileName,
+                                                        mapDocument->lastExportFileName(),
+                                                        selectedFilter,
+                                                        this);
+    if (!exportDetails.isValid()) {
+        return;
+    }
+
+    // Check if writer will overwrite existing files here because some writers
+    // could save to multiple files at the same time. For example CSV saves
+    // each layer into a separate file.
+    QStringList outputFiles = exportDetails.mFormat->outputFiles(mapDocument->map(), exportDetails.mFileName);
+    if (outputFiles.size() > 0) {
+        // Check if any output file already exists
+        QString message =
+                tr("Some export files already exist:") + QLatin1String("\n\n");
+
+        bool overwriteHappens = false;
+
+        for (const QString &outputFile : outputFiles) {
+            if (QFile::exists(outputFile)) {
+                overwriteHappens = true;
+                message += outputFile + QLatin1Char('\n');
+            }
+        }
+        message += QLatin1Char('\n') + tr("Do you want to replace them?");
+
+        // If overwrite happens, warn the user and get confirmation before exporting
+        if (overwriteHappens) {
+            const QMessageBox::StandardButton reply = QMessageBox::warning(
+                                                          this,
+                                                          tr("Overwrite Files"),
+                                                          message,
+                                                          QMessageBox::Yes | QMessageBox::No,
+                                                          QMessageBox::No);
+
+            if (reply != QMessageBox::Yes)
+                return;
+        }
+    }
+
+    Preferences *pref = Preferences::instance();
+
+    pref->setLastPath(Preferences::ExportedFile, QFileInfo(exportDetails.mFileName).path());
+    mSettings.setValue(QLatin1String("lastUsedExportFilter"), selectedFilter);
+
+    auto exportResult = exportDetails.mFormat->write(mapDocument->map(), exportDetails.mFileName);
+    if (!exportResult) {
+        QMessageBox::critical(this, tr("Error Exporting Map!"),
+                              exportDetails.mFormat->errorString());
+    } else {
+        // Remember export parameters, so subsequent exports can be done faster
+        mapDocument->setLastExportFileName(exportDetails.mFileName);
+        mapDocument->setExportFormat(exportDetails.mFormat);
+    }
+}
+
+void MainWindow::exportTilesetAs(TilesetDocument *tilesetDocument)
+{
+    QString fileName = tilesetDocument->fileName();
+    QString selectedFilter =
+            mSettings.value(QLatin1String("lastUsedExportFilter")).toString();
+    auto exportDetails = chooseExportDetails<TilesetFormat>(fileName,
+                                                    tilesetDocument->lastExportFileName(),
+                                                    selectedFilter,
+                                                    this);
+    if (!exportDetails.isValid())
+        return;
+
+    Preferences *pref = Preferences::instance();
+
+    pref->setLastPath(Preferences::ExportedFile, QFileInfo(exportDetails.mFileName).path());
+    mSettings.setValue(QLatin1String("lastUsedExportFilter"), selectedFilter);
+
+    auto exportResult = exportDetails.mFormat->write(*tilesetDocument->tileset(), exportDetails.mFileName);
+    if (!exportResult) {
+        QMessageBox::critical(this, tr("Error Exporting Map!"),
+                              exportDetails.mFormat->errorString());
+    } else {
+        // Remember export parameters, so subsequent exports can be done faster
+        tilesetDocument->setLastExportFileName(exportDetails.mFileName);
+        tilesetDocument->setExportFormat(exportDetails.mFormat);
+    }
 }
 
 void MainWindow::documentChanged(Document *document)
