@@ -25,6 +25,7 @@
 #include "addremovemapobject.h"
 #include "addremovetileset.h"
 #include "changelayer.h"
+#include "changemapobject.h"
 #include "changemapobjectsorder.h"
 #include "changeproperties.h"
 #include "changeselectedarea.h"
@@ -46,6 +47,7 @@
 #include "offsetlayer.h"
 #include "orthogonalrenderer.h"
 #include "painttilelayer.h"
+#include "preferences.h"
 #include "rangeset.h"
 #include "reparentlayers.h"
 #include "resizemap.h"
@@ -56,13 +58,14 @@
 #include "tile.h"
 #include "tilelayer.h"
 #include "tilesetdocument.h"
-#include "tilesetmanager.h"
 #include "tmxmapformat.h"
 
 #include <QFileInfo>
 #include <QRect>
 #include <QUndoStack>
 
+#include "qtcompat_p.h"
+
 using namespace Tiled;
 using namespace Tiled::Internal;
 
@@ -107,18 +110,10 @@ MapDocument::MapDocument(Map *map, const QString &fileName)
             SLOT(onMapObjectModelRowsInsertedOrRemoved(QModelIndex,int,int)));
     connect(mMapObjectModel, SIGNAL(rowsMoved(QModelIndex,int,int,QModelIndex,int)),
             SLOT(onObjectsMoved(QModelIndex,int,int,QModelIndex,int)));
-
-    // Register tileset references
-    TilesetManager *tilesetManager = TilesetManager::instance();
-    tilesetManager->addReferences(mMap->tilesets());
 }
 
 MapDocument::~MapDocument()
 {
-    // Unregister tileset references
-    TilesetManager *tilesetManager = TilesetManager::instance();
-    tilesetManager->removeReferences(mMap->tilesets());
-
     delete mRenderer;
     delete mMap;
 }
@@ -198,9 +193,10 @@ MapFormat *MapDocument::exportFormat() const
     return mExportFormat;
 }
 
-void MapDocument::setExportFormat(MapFormat *format)
+void MapDocument::setExportFormat(FileFormat *format)
 {
-    mExportFormat = format;
+    mExportFormat = qobject_cast<MapFormat*>(format);
+    Q_ASSERT(mExportFormat);
 }
 
 /**
@@ -333,6 +329,20 @@ void MapDocument::resizeMap(const QSize &size, const QPoint &offset, bool remove
     // TODO: Handle layers that don't match the map size correctly
 }
 
+void MapDocument::autocropMap()
+{
+    if (!mCurrentLayer || !mCurrentLayer->isTileLayer())
+        return;
+
+    TileLayer *tileLayer = static_cast<TileLayer*>(mCurrentLayer);
+
+    const QRect bounds = tileLayer->region().boundingRect();
+    if (bounds.isNull())
+        return;
+
+    resizeMap(bounds.size(), -bounds.topLeft(), true);
+}
+
 void MapDocument::offsetMap(const QList<Layer*> &layers,
                             const QPoint &offset,
                             const QRect &bounds,
@@ -341,17 +351,12 @@ void MapDocument::offsetMap(const QList<Layer*> &layers,
     if (layers.empty())
         return;
 
-    if (layers.size() == 1) {
-        mUndoStack->push(new OffsetLayer(this, layers.first(), offset,
+    mUndoStack->beginMacro(tr("Offset Map"));
+    for (auto layer : layers) {
+        mUndoStack->push(new OffsetLayer(this, layer, offset,
                                          bounds, wrapX, wrapY));
-    } else {
-        mUndoStack->beginMacro(tr("Offset Map"));
-        for (auto layer : layers) {
-            mUndoStack->push(new OffsetLayer(this, layer, offset,
-                                             bounds, wrapX, wrapY));
-        }
-        mUndoStack->endMacro();
     }
+    mUndoStack->endMacro();
 }
 
 /**
@@ -593,8 +598,6 @@ void MapDocument::insertTileset(int index, const SharedTileset &tileset)
 {
     emit tilesetAboutToBeAdded(index);
     mMap->insertTileset(index, tileset);
-    TilesetManager *tilesetManager = TilesetManager::instance();
-    tilesetManager->addReference(tileset);
     emit tilesetAdded(index, tileset.data());
 }
 
@@ -612,9 +615,6 @@ void MapDocument::removeTilesetAt(int index)
     SharedTileset tileset = mMap->tilesets().at(index);
     mMap->removeTilesetAt(index);
     emit tilesetRemoved(tileset.data());
-
-    TilesetManager *tilesetManager = TilesetManager::instance();
-    tilesetManager->removeReference(tileset);
 }
 
 /**
@@ -629,11 +629,6 @@ SharedTileset MapDocument::replaceTileset(int index, const SharedTileset &tilese
 
     bool added = mMap->replaceTileset(oldTileset, tileset);
 
-    TilesetManager *tilesetManager = TilesetManager::instance();
-    if (added)
-        tilesetManager->addReference(tileset);
-    tilesetManager->removeReference(oldTileset);
-
     if (added)
         emit tilesetReplaced(index, tileset.data(), oldTileset.data());
     else
@@ -642,6 +637,16 @@ SharedTileset MapDocument::replaceTileset(int index, const SharedTileset &tilese
     return oldTileset;
 }
 
+void MapDocument::replaceObjectTemplate(const ObjectTemplate *oldObjectTemplate,
+                                        const ObjectTemplate *newObjectTemplate)
+{
+    auto changedObjects = mMap->replaceObjectTemplate(oldObjectTemplate, newObjectTemplate);
+
+    // Update the objects in the map scene
+    emit objectsChanged(changedObjects);
+    emit objectTemplateReplaced(newObjectTemplate, oldObjectTemplate);
+}
+
 void MapDocument::setSelectedArea(const QRegion &selection)
 {
     if (mSelectedArea != selection) {
@@ -651,6 +656,34 @@ void MapDocument::setSelectedArea(const QRegion &selection)
     }
 }
 
+static QList<MapObject *> sortObjects(const Map *map, const QList<MapObject *> &objects)
+{
+    QList<MapObject *> sorted;
+    sorted.reserve(objects.size());
+
+    LayerIterator iterator(map);
+    while (Layer *layer = iterator.next()) {
+        if (layer->layerType() != Layer::ObjectGroupType)
+            continue;
+
+        for (MapObject *mapObject : static_cast<ObjectGroup*>(layer)->objects()) {
+            if (objects.contains(mapObject))
+                sorted.append(mapObject);
+        }
+    }
+
+    return sorted;
+}
+
+/**
+ * Returns the list of selected objects, in their display order (when
+ * ObjectGroup::IndexOrder is used).
+ */
+QList<MapObject *> MapDocument::selectedObjectsOrdered() const
+{
+    return sortObjects(mMap, mSelectedObjects);
+}
+
 void MapDocument::setSelectedObjects(const QList<MapObject *> &selectedObjects)
 {
     mSelectedObjects = selectedObjects;
@@ -696,27 +729,22 @@ QList<Object*> MapDocument::currentObjects() const
  * To reach the aim, all similar tilesets will be replaced by the version
  * in the current map document and all missing tilesets will be added to
  * the current map document.
- *
- * \warning This method assumes that the tilesets in \a map are managed by
- *          the TilesetManager!
  */
 void MapDocument::unifyTilesets(Map *map)
 {
     QList<QUndoCommand*> undoCommands;
-    const QVector<SharedTileset> &existingTilesets = mMap->tilesets();
-    QVector<SharedTileset> addedTilesets;
-    TilesetManager *tilesetManager = TilesetManager::instance();
+    QVector<SharedTileset> availableTilesets = mMap->tilesets();
 
     // Iterate over a copy because map->replaceTileset may invalidate iterator
     const QVector<SharedTileset> tilesets = map->tilesets();
     for (const SharedTileset &tileset : tilesets) {
-        if (existingTilesets.contains(tileset))
+        if (availableTilesets.contains(tileset))
             continue;
 
-        SharedTileset replacement = tileset->findSimilarTileset(existingTilesets);
-        if (!replacement && !addedTilesets.contains(replacement)) {
+        SharedTileset replacement = tileset->findSimilarTileset(availableTilesets);
+        if (!replacement) {
             undoCommands.append(new AddTileset(this, tileset));
-            addedTilesets.append(replacement);
+            availableTilesets.append(tileset);
             continue;
         }
 
@@ -732,9 +760,7 @@ void MapDocument::unifyTilesets(Map *map)
             }
         }
 
-        if (map->replaceTileset(tileset, replacement))
-            tilesetManager->addReference(replacement);
-        tilesetManager->removeReference(tileset);
+        map->replaceTileset(tileset, replacement);
     }
 
     if (!undoCommands.isEmpty()) {
@@ -750,35 +776,32 @@ void MapDocument::unifyTilesets(Map *map)
  * Replaces tilesets in \a map by similar tilesets in this map when possible,
  * and adds tilesets to \a missingTilesets whenever there is a tileset without
  * replacement in this map.
- *
- * \warning This method assumes that the tilesets in \a map are managed by
- *          the TilesetManager!
  */
 void MapDocument::unifyTilesets(Map *map, QVector<SharedTileset> &missingTilesets)
 {
-    const QVector<SharedTileset> &existingTilesets = mMap->tilesets();
-    TilesetManager *tilesetManager = TilesetManager::instance();
+    QVector<SharedTileset> availableTilesets = mMap->tilesets();
+    for (const SharedTileset &tileset : qAsConst(missingTilesets))
+        if (!availableTilesets.contains(tileset))
+            availableTilesets.append(tileset);
 
     // Iterate over a copy because map->replaceTileset may invalidate iterator
     const QVector<SharedTileset> tilesets = map->tilesets();
     for (const SharedTileset &tileset : tilesets) {
         // tileset already added
-        if (existingTilesets.contains(tileset))
+        if (availableTilesets.contains(tileset))
             continue;
 
-        SharedTileset replacement = tileset->findSimilarTileset(existingTilesets);
+        SharedTileset replacement = tileset->findSimilarTileset(availableTilesets);
 
         // tileset not present and no replacement tileset found
         if (!replacement) {
-            if (!missingTilesets.contains(tileset))
-                missingTilesets.append(tileset);
+            missingTilesets.append(tileset);
+            availableTilesets.append(tileset);
             continue;
         }
 
         // replacement tileset found, change given map
-        if (map->replaceTileset(tileset, replacement))
-            tilesetManager->addReference(replacement);
-        tilesetManager->removeReference(tileset);
+        map->replaceTileset(tileset, replacement);
     }
 }
 
@@ -886,6 +909,30 @@ void MapDocument::onLayerRemoved(Layer *layer)
     emit layerRemoved(layer);
 }
 
+void MapDocument::updateTemplateInstances(const ObjectTemplate *objectTemplate)
+{
+    QList<MapObject*> objectList;
+    for (ObjectGroup *group : mMap->objectGroups()) {
+        for (auto object : group->objects()) {
+            if (object->objectTemplate() == objectTemplate) {
+                object->syncWithTemplate();
+                objectList.append(object);
+            }
+        }
+    }
+    emit objectsChanged(objectList);
+}
+
+void MapDocument::selectAllInstances(const ObjectTemplate *objectTemplate)
+{
+    QList<MapObject*> objectList;
+    for (ObjectGroup *group : mMap->objectGroups())
+        for (auto object : group->objects())
+            if (object->objectTemplate() == objectTemplate)
+                objectList.append(object);
+    setSelectedObjects(objectList);
+}
+
 void MapDocument::deselectObjects(const QList<MapObject *> &objects)
 {
     // Unset the current object when it was part of this list of objects
@@ -943,8 +990,8 @@ void MapDocument::moveObjectsToGroup(const QList<MapObject *> &objects,
     mUndoStack->beginMacro(tr("Move %n Object(s) to Layer", "",
                               objects.size()));
 
-    const auto objectsCopy = objects;   // original list may get modified
-    for (MapObject *mapObject : objectsCopy) {
+    const auto objectsToMove = sortObjects(mMap, objects);
+    for (MapObject *mapObject : objectsToMove) {
         if (mapObject->objectGroup() == objectGroup)
             continue;
 
@@ -1043,6 +1090,14 @@ void MapDocument::moveObjectsDown(const QList<MapObject *> &objects)
         mUndoStack->push(command.take());
 }
 
+void MapDocument::detachObjects(const QList<MapObject *> &objects)
+{
+    if (objects.isEmpty())
+        return;
+
+    mUndoStack->push(new DetachObjects(this, objects));
+}
+
 void MapDocument::createRenderer()
 {
     if (mRenderer)
