@@ -24,6 +24,7 @@
 
 #include "document.h"
 #include "layer.h"
+#include "mapformat.h"
 #include "tiled.h"
 #include "tileset.h"
 
@@ -42,9 +43,10 @@ namespace Tiled {
 class Map;
 class MapObject;
 class MapRenderer;
-class MapFormat;
+class ObjectTemplate;
 class Terrain;
 class Tile;
+class WangSet;
 
 namespace Internal {
 
@@ -66,13 +68,20 @@ class MapDocument : public Document
     Q_OBJECT
 
 public:
+    enum TileLayerChangeFlag {
+        LayerDrawMarginsChanged,
+        LayerBoundsChanged
+    };
+    Q_DECLARE_FLAGS(TileLayerChangeFlags, TileLayerChangeFlag)
+    Q_FLAG(TileLayerChangeFlags)
+
     /**
      * Constructs a map document around the given map. The map document takes
      * ownership of the map.
      */
     MapDocument(Map *map, const QString &fileName = QString());
 
-    ~MapDocument();
+    ~MapDocument() override;
 
     bool save(const QString &fileName, QString *error = nullptr) override;
 
@@ -84,17 +93,14 @@ public:
                              MapFormat *format,
                              QString *error = nullptr);
 
-    QString lastExportFileName() const;
-    void setLastExportFileName(const QString &fileName);
-
     MapFormat *readerFormat() const;
     void setReaderFormat(MapFormat *format);
 
     FileFormat *writerFormat() const override;
     void setWriterFormat(MapFormat *format);
 
-    MapFormat *exportFormat() const;
-    void setExportFormat(MapFormat *format);
+    MapFormat *exportFormat() const override;
+    void setExportFormat(FileFormat *format) override;
 
     QString displayName() const override;
 
@@ -120,6 +126,8 @@ public:
      */
     void resizeMap(const QSize &size, const QPoint &offset, bool removeObjects);
 
+    void autocropMap();
+
     /**
      * Offsets the \a layers by \a offset, within \a bounds, and optionally
      * wraps on the X or Y axis.
@@ -146,12 +154,16 @@ public:
     void removeTilesetAt(int index);
     SharedTileset replaceTileset(int index, const SharedTileset &tileset);
 
+    void replaceObjectTemplate(const ObjectTemplate *oldObjectTemplate,
+                               const ObjectTemplate *newObjectTemplate);
+
     void duplicateObjects(const QList<MapObject*> &objects);
     void removeObjects(const QList<MapObject*> &objects);
     void moveObjectsToGroup(const QList<MapObject*> &objects,
                             ObjectGroup *objectGroup);
     void moveObjectsUp(const QList<MapObject*> &objects);
     void moveObjectsDown(const QList<MapObject*> &objects);
+    void detachObjects(const QList<MapObject*> &objects);
 
     /**
      * Returns the layer model. Can be used to modify the layer stack of the
@@ -188,6 +200,8 @@ public:
     const QList<MapObject*> &selectedObjects() const
     { return mSelectedObjects; }
 
+    QList<MapObject*> selectedObjectsOrdered() const;
+
     /**
      * Sets the list of selected objects, emitting the selectedObjectsChanged
      * signal.
@@ -236,10 +250,10 @@ signals:
     void currentLayerChanged(Layer *layer);
 
     /**
-     * Emitted when a certain region of the map changes. The region is given in
-     * tile coordinates.
+     * Emitted when a certain \a region of a \a tileLayer changes. The region
+     * is given in tile coordinates.
      */
-    void regionChanged(const QRegion &region, Layer *layer);
+    void regionChanged(const QRegion &region, TileLayer *tileLayer);
 
     /**
      * Emitted when a certain region of the map was edited by user input.
@@ -248,7 +262,7 @@ signals:
      */
     void regionEdited(const QRegion &region, Layer *layer);
 
-    void tileLayerDrawMarginsChanged(TileLayer *layer);
+    void tileLayerChanged(TileLayer *layer, TileLayerChangeFlags flags);
 
     /**
      * Should be emitted when changing the color or drawing order of an object
@@ -268,6 +282,9 @@ signals:
     void tilesetRemoved(Tileset *tileset);
     void tilesetReplaced(int index, Tileset *tileset, Tileset *oldTileset);
 
+    void objectTemplateReplaced(const ObjectTemplate *newObjectTemplate,
+                                const ObjectTemplate *oldObjectTemplate);
+
     void objectsAdded(const QList<MapObject*> &objects);
     void objectsInserted(ObjectGroup *objectGroup, int first, int last);
     void objectsRemoved(const QList<MapObject*> &objects);
@@ -293,12 +310,14 @@ private slots:
     void onLayerAboutToBeRemoved(GroupLayer *groupLayer, int index);
     void onLayerRemoved(Layer *layer);
 
+public slots:
+    void updateTemplateInstances(const ObjectTemplate *objectTemplate);
+    void selectAllInstances(const ObjectTemplate *objectTemplate);
+
 private:
     void deselectObjects(const QList<MapObject*> &objects);
     void moveObjectIndex(const MapObject *object, int count);
 
-    QString mLastExportFileName;
-
     /*
      * QPointer is used since the formats referenced here may be dynamically
      * added by a plugin, and can also be removed again.
@@ -315,16 +334,7 @@ private:
     MapObjectModel *mMapObjectModel;
 };
 
-
-inline QString MapDocument::lastExportFileName() const
-{
-    return mLastExportFileName;
-}
-
-inline void MapDocument::setLastExportFileName(const QString &fileName)
-{
-    mLastExportFileName = fileName;
-}
-
 } // namespace Internal
 } // namespace Tiled
+
+Q_DECLARE_OPERATORS_FOR_FLAGS(Tiled::Internal::MapDocument::TileLayerChangeFlags)
