@@ -22,6 +22,7 @@
 #include "ui_noeditorwidget.h"
 
 #include "actionmanager.h"
+#include "documentmanager.h"
 
 namespace Tiled {
 namespace Internal {
@@ -34,6 +35,7 @@ NoEditorWidget::NoEditorWidget(QWidget *parent) :
 
     connect(ui->newMapButton, &QPushButton::clicked, this, &NoEditorWidget::newMap);
     connect(ui->newTilesetButton, &QPushButton::clicked, this, &NoEditorWidget::newTileset);
+    connect(ui->openFileButton, &QPushButton::clicked, this, &NoEditorWidget::openFile);
 }
 
 NoEditorWidget::~NoEditorWidget()
@@ -63,5 +65,10 @@ void NoEditorWidget::newTileset()
     ActionManager::action("file.new_tileset")->trigger();
 }
 
+void NoEditorWidget::openFile()
+{
+    DocumentManager::instance()->openFile();
+}
+
 } // namespace Internal
 } // namespace Tiled
