@@ -30,6 +30,8 @@
 #include "changetile.h"
 #include "changetileimagesource.h"
 #include "changetileprobability.h"
+#include "changewangsetdata.h"
+#include "changewangcolordata.h"
 #include "flipmapobjects.h"
 #include "imagelayer.h"
 #include "map.h"
@@ -37,11 +39,13 @@
 #include "mapobject.h"
 #include "movemapobject.h"
 #include "objectgroup.h"
+#include "objecttemplate.h"
 #include "preferences.h"
 #include "replacetileset.h"
 #include "resizemapobject.h"
 #include "renamelayer.h"
 #include "renameterrain.h"
+#include "renamewangset.h"
 #include "rotatemapobject.h"
 #include "terrain.h"
 #include "tile.h"
@@ -51,9 +55,12 @@
 #include "tilesetformat.h"
 #include "tilesetmanager.h"
 #include "tilesetterrainmodel.h"
+#include "tilesetwangsetmodel.h"
 #include "utils.h"
 #include "varianteditorfactory.h"
 #include "variantpropertymanager.h"
+#include "wangset.h"
+#include "wangcolormodel.h"
 
 #include <QtGroupPropertyManager>
 
@@ -68,6 +75,7 @@ namespace Internal {
 PropertyBrowser::PropertyBrowser(QWidget *parent)
     : QtTreePropertyBrowser(parent)
     , mUpdating(false)
+    , mMapObjectFlags(0)
     , mObject(nullptr)
     , mDocument(nullptr)
     , mMapDocument(nullptr)
@@ -83,36 +91,7 @@ PropertyBrowser::PropertyBrowser(QWidget *parent)
     setPropertiesWithoutValueMarked(true);
     setAllowMultiSelection(true);
 
-    mStaggerAxisNames.append(tr("X"));
-    mStaggerAxisNames.append(tr("Y"));
-
-    mStaggerIndexNames.append(tr("Odd"));
-    mStaggerIndexNames.append(tr("Even"));
-
-    mOrientationNames.append(QCoreApplication::translate("Tiled::Internal::NewMapDialog", "Orthogonal"));
-    mOrientationNames.append(QCoreApplication::translate("Tiled::Internal::NewMapDialog", "Isometric"));
-    mOrientationNames.append(QCoreApplication::translate("Tiled::Internal::NewMapDialog", "Isometric (Staggered)"));
-    mOrientationNames.append(QCoreApplication::translate("Tiled::Internal::NewMapDialog", "Hexagonal (Staggered)"));
-
-    mTilesetOrientationNames.append(mOrientationNames.at(0));
-    mTilesetOrientationNames.append(mOrientationNames.at(1));
-
-    mLayerFormatNames.append(QCoreApplication::translate("PreferencesDialog", "XML"));
-    mLayerFormatNames.append(QCoreApplication::translate("PreferencesDialog", "Base64 (uncompressed)"));
-    mLayerFormatNames.append(QCoreApplication::translate("PreferencesDialog", "Base64 (gzip compressed)"));
-    mLayerFormatNames.append(QCoreApplication::translate("PreferencesDialog", "Base64 (zlib compressed)"));
-    mLayerFormatNames.append(QCoreApplication::translate("PreferencesDialog", "CSV"));
-
-    mRenderOrderNames.append(QCoreApplication::translate("PreferencesDialog", "Right Down"));
-    mRenderOrderNames.append(QCoreApplication::translate("PreferencesDialog", "Right Up"));
-    mRenderOrderNames.append(QCoreApplication::translate("PreferencesDialog", "Left Down"));
-    mRenderOrderNames.append(QCoreApplication::translate("PreferencesDialog", "Left Up"));
-
-    mFlippingFlagNames.append(tr("Horizontal"));
-    mFlippingFlagNames.append(tr("Vertical"));
-
-    mDrawOrderNames.append(tr("Top Down"));
-    mDrawOrderNames.append(tr("Manual"));
+    retranslateUi();
 
     connect(mVariantManager, SIGNAL(valueChanged(QtProperty*,QVariant)),
             SLOT(valueChanged(QtProperty*,QVariant)));
@@ -145,8 +124,10 @@ void PropertyBrowser::setDocument(Document *document)
 
     if (mDocument) {
         mDocument->disconnect(this);
-        if (mTilesetDocument)
+        if (mTilesetDocument) {
             mTilesetDocument->terrainModel()->disconnect(this);
+            mTilesetDocument->wangSetModel()->disconnect(this);
+        }
     }
 
     mDocument = document;
@@ -154,18 +135,18 @@ void PropertyBrowser::setDocument(Document *document)
     mTilesetDocument = tilesetDocument;
 
     if (mapDocument) {
-        connect(mapDocument, SIGNAL(mapChanged()),
-                SLOT(mapChanged()));
-        connect(mapDocument, SIGNAL(objectsChanged(QList<MapObject*>)),
-                SLOT(objectsChanged(QList<MapObject*>)));
-        connect(mapDocument, SIGNAL(objectsTypeChanged(QList<MapObject*>)),
-                SLOT(objectsTypeChanged(QList<MapObject*>)));
+        connect(mapDocument, &MapDocument::mapChanged,
+                this, &PropertyBrowser::mapChanged);
+        connect(mapDocument, &MapDocument::objectsChanged,
+                this, &PropertyBrowser::objectsChanged);
+        connect(mapDocument, &MapDocument::objectsTypeChanged,
+                this, &PropertyBrowser::objectsTypeChanged);
         connect(mapDocument, &MapDocument::layerChanged,
                 this, &PropertyBrowser::layerChanged);
-        connect(mapDocument, SIGNAL(objectGroupChanged(ObjectGroup*)),
-                SLOT(objectGroupChanged(ObjectGroup*)));
-        connect(mapDocument, SIGNAL(imageLayerChanged(ImageLayer*)),
-                SLOT(imageLayerChanged(ImageLayer*)));
+        connect(mapDocument, &MapDocument::objectGroupChanged,
+                this, &PropertyBrowser::objectGroupChanged);
+        connect(mapDocument, &MapDocument::imageLayerChanged,
+                this, &PropertyBrowser::imageLayerChanged);
         connect(mapDocument, &MapDocument::tileTypeChanged,
                 this, &PropertyBrowser::tileTypeChanged);
 
@@ -194,6 +175,10 @@ void PropertyBrowser::setDocument(Document *document)
         TilesetTerrainModel *terrainModel = tilesetDocument->terrainModel();
         connect(terrainModel, &TilesetTerrainModel::terrainChanged,
                 this, &PropertyBrowser::terrainChanged);
+
+        TilesetWangSetModel *wangSetModel = tilesetDocument->wangSetModel();
+        connect(wangSetModel, &TilesetWangSetModel::wangSetChanged,
+                this, &PropertyBrowser::wangSetChanged);
     }
 
     if (document) {
@@ -237,6 +222,11 @@ void PropertyBrowser::editCustomProperty(const QString &name)
         editItem(propertyItems.first());
 }
 
+QSize PropertyBrowser::sizeHint() const
+{
+    return Utils::dpiScaled(QSize(260, 100));
+}
+
 bool PropertyBrowser::event(QEvent *event)
 {
     if (event->type() == QEvent::LanguageChange)
@@ -324,6 +314,12 @@ void PropertyBrowser::terrainChanged(Tileset *tileset, int index)
         updateProperties();
 }
 
+void PropertyBrowser::wangSetChanged(Tileset *tileset, int index)
+{
+    if (mObject == tileset->wangSet(index))
+        updateProperties();
+}
+
 static QVariant predefinedPropertyValue(Object *object, const QString &name)
 {
     QString objectType;
@@ -349,14 +345,16 @@ static QVariant predefinedPropertyValue(Object *object, const QString &name)
     case Object::MapType:
     case Object::TerrainType:
     case Object::TilesetType:
+    case Object::WangSetType:
+    case Object::WangColorType:
+    case Object::ObjectTemplateType:
         break;
     }
 
     if (objectType.isEmpty())
         return QVariant();
 
-    const ObjectTypes objectTypes = Preferences::instance()->objectTypes();
-    for (const ObjectType &type : objectTypes) {
+    for (const ObjectType &type : Object::objectTypes()) {
         if (type.name == objectType)
             if (type.defaultProperties.contains(name))
                 return type.defaultProperties.value(name);
@@ -526,12 +524,15 @@ void PropertyBrowser::valueChanged(QtProperty *property, const QVariant &val)
     }
 
     switch (mObject->typeId()) {
-    case Object::MapType:       applyMapValue(id, val); break;
-    case Object::MapObjectType: applyMapObjectValue(id, val); break;
-    case Object::LayerType:     applyLayerValue(id, val); break;
-    case Object::TilesetType:   applyTilesetValue(id, val); break;
-    case Object::TileType:      applyTileValue(id, val); break;
-    case Object::TerrainType:   applyTerrainValue(id, val); break;
+    case Object::MapType:               applyMapValue(id, val); break;
+    case Object::MapObjectType:         applyMapObjectValue(id, val); break;
+    case Object::LayerType:             applyLayerValue(id, val); break;
+    case Object::TilesetType:           applyTilesetValue(id, val); break;
+    case Object::TileType:              applyTileValue(id, val); break;
+    case Object::TerrainType:           applyTerrainValue(id, val); break;
+    case Object::WangSetType:           applyWangSetValue(id, val); break;
+    case Object::WangColorType:         applyWangColorValue(id, val); break;
+    case Object::ObjectTemplateType:    break;
     }
 }
 
@@ -564,6 +565,7 @@ void PropertyBrowser::addMapProperties()
     addProperty(HeightProperty, QVariant::Int, tr("Height"), groupProperty)->setEnabled(false);
     addProperty(TileWidthProperty, QVariant::Int, tr("Tile Width"), groupProperty);
     addProperty(TileHeightProperty, QVariant::Int, tr("Tile Height"), groupProperty);
+    addProperty(InfiniteProperty, QVariant::Bool, tr("Infinite"), groupProperty);
 
     addProperty(HexSideLengthProperty, QVariant::Int, tr("Tile Side Length (Hex)"), groupProperty);
 
@@ -606,17 +608,36 @@ void PropertyBrowser::addMapProperties()
 static QStringList objectTypeNames()
 {
     QStringList names;
-    for (const ObjectType &type : Preferences::instance()->objectTypes())
+    for (const ObjectType &type : Object::objectTypes())
         names.append(type.name);
     return names;
 }
 
+enum MapObjectFlags {
+    ObjectHasDimensions = 0x1,
+    ObjectHasTile = 0x2,
+    ObjectIsText = 0x4
+};
+
+static int mapObjectFlags(const MapObject *mapObject)
+{
+    int flags = 0;
+    if (mapObject->hasDimensions())
+        flags |= ObjectHasDimensions;
+    if (!mapObject->cell().isEmpty())
+        flags |= ObjectHasTile;
+    if (mapObject->shape() == MapObject::Text)
+        flags |= ObjectIsText;
+    return flags;
+}
+
 void PropertyBrowser::addMapObjectProperties()
 {
     // DEFAULT MAP OBJECT PROPERTIES
     QtProperty *groupProperty = mGroupManager->addProperty(tr("Object"));
 
     addProperty(IdProperty, QVariant::Int, tr("ID"), groupProperty)->setEnabled(false);
+    addProperty(TemplateProperty, filePathTypeId(), tr("Template"), groupProperty)->setEnabled(false);
     addProperty(NameProperty, QVariant::String, tr("Name"), groupProperty);
 
     QtVariantProperty *typeProperty =
@@ -628,16 +649,17 @@ void PropertyBrowser::addMapObjectProperties()
     addProperty(YProperty, QVariant::Double, tr("Y"), groupProperty);
 
     auto mapObject = static_cast<const MapObject*>(mObject);
+    mMapObjectFlags = mapObjectFlags(mapObject);
 
-    if (!mapObject->isPolyShape()) {
+    if (mMapObjectFlags & ObjectHasDimensions) {
         addProperty(WidthProperty, QVariant::Double, tr("Width"), groupProperty);
         addProperty(HeightProperty, QVariant::Double, tr("Height"), groupProperty);
     }
 
-    addProperty(RotationProperty, QVariant::Double, tr("Rotation"), groupProperty);
+    bool isPoint = mapObject->shape() == MapObject::Point;
+    addProperty(RotationProperty, QVariant::Double, tr("Rotation"), groupProperty)->setEnabled(!isPoint);
 
-
-    if (!mapObject->cell().isEmpty()) {
+    if (mMapObjectFlags & ObjectHasTile) {
         QtVariantProperty *flippingProperty =
                 addProperty(FlippingProperty, VariantPropertyManager::flagTypeId(),
                                tr("Flipping"), groupProperty);
@@ -645,7 +667,7 @@ void PropertyBrowser::addMapObjectProperties()
         flippingProperty->setAttribute(QLatin1String("flagNames"), mFlippingFlagNames);
     }
 
-    if (mapObject->shape() == MapObject::Text) {
+    if (mMapObjectFlags & ObjectIsText) {
         addProperty(TextProperty, QVariant::String, tr("Text"), groupProperty)->setAttribute(QLatin1String("multiline"), true);
         addProperty(TextAlignmentProperty, VariantPropertyManager::alignmentTypeId(), tr("Alignment"), groupProperty);
         addProperty(FontProperty, QVariant::Font, tr("Font"), groupProperty);
@@ -660,6 +682,7 @@ void PropertyBrowser::addLayerProperties(QtProperty *parent)
 {
     addProperty(NameProperty, QVariant::String, tr("Name"), parent);
     addProperty(VisibleProperty, QVariant::Bool, tr("Visible"), parent);
+    addProperty(LockedProperty, QVariant::Bool, tr("Locked"), parent);
 
     QtVariantProperty *opacityProperty =
             addProperty(OpacityProperty, QVariant::Double, tr("Opacity"), parent);
@@ -831,6 +854,50 @@ void PropertyBrowser::addTerrainProperties()
     addProperty(groupProperty);
 }
 
+void PropertyBrowser::addWangSetProperties()
+{
+    QtProperty *groupProperty = mGroupManager->addProperty(tr("WangSet"));
+    QtVariantProperty *nameProperty = addProperty(NameProperty, QVariant::String, tr("Name"), groupProperty);
+    QtVariantProperty *edgeProperty = addProperty(EdgeCountProperty, QVariant::Int, tr("Edge Count"), groupProperty);
+    QtVariantProperty *cornerProperty = addProperty(CornerCountProperty, QVariant::Int, tr("Corner Count"), groupProperty);
+
+    edgeProperty->setAttribute(QLatin1String("minimum"), 1);
+    edgeProperty->setAttribute(QLatin1String("maximum"), 15);
+    cornerProperty->setAttribute(QLatin1String("minimum"), 1);
+    cornerProperty->setAttribute(QLatin1String("maximum"), 15);
+
+    nameProperty->setEnabled(mTilesetDocument);
+    edgeProperty->setEnabled(mTilesetDocument);
+    cornerProperty->setEnabled(mTilesetDocument);
+
+    addProperty(groupProperty);
+}
+
+void PropertyBrowser::addWangColorProperties()
+{
+    QtProperty *groupProperty = mGroupManager->addProperty(tr("Wang Color"));
+    QtVariantProperty *nameProperty = addProperty(NameProperty,
+                                                  QVariant::String,
+                                                  tr("Name"),
+                                                  groupProperty);
+    QtVariantProperty *colorProperty = addProperty(ColorProperty,
+                                                   QVariant::Color,
+                                                   tr("Color"),
+                                                   groupProperty);
+    QtVariantProperty *probabilityProperty = addProperty(WangColorProbabilityProperty,
+                                                         QVariant::Double,
+                                                         tr("Probability"),
+                                                         groupProperty);
+
+    probabilityProperty->setAttribute(QLatin1String("minimum"), 0.01);
+
+    nameProperty->setEnabled(mTilesetDocument);
+    colorProperty->setEnabled(mTilesetDocument);
+    probabilityProperty->setEnabled(mTilesetDocument);
+
+    addProperty(groupProperty);
+}
+
 void PropertyBrowser::applyMapValue(PropertyId id, const QVariant &val)
 {
     QUndoCommand *command = nullptr;
@@ -844,6 +911,32 @@ void PropertyBrowser::applyMapValue(PropertyId id, const QVariant &val)
         command = new ChangeMapProperty(mMapDocument, ChangeMapProperty::TileHeight,
                                         val.toInt());
         break;
+    case InfiniteProperty: {
+        bool infinite = val.toInt();
+
+        QUndoStack *undoStack = mDocument->undoStack();
+        undoStack->beginMacro(tr("Change Infinite Property"));
+
+        if (!infinite) {
+            QRect mapBounds(QPoint(0, 0), mMapDocument->map()->size());
+
+            LayerIterator iterator(mMapDocument->map());
+            while (Layer *layer = iterator.next()) {
+                if (TileLayer *tileLayer = dynamic_cast<TileLayer*>(layer))
+                    mapBounds = mapBounds.united(tileLayer->bounds());
+            }
+
+            if (mapBounds.size() == QSize(0, 0))
+                mapBounds.setSize(QSize(1, 1));
+
+            mMapDocument->resizeMap(mapBounds.size(), -mapBounds.topLeft(), false);
+        }
+
+        undoStack->push(new ChangeMapProperty(mMapDocument, ChangeMapProperty::Infinite,
+                                              val.toInt()));
+        undoStack->endMacro();
+        break;
+    }
     case OrientationProperty: {
         Map::Orientation orientation = static_cast<Map::Orientation>(val.toInt() + 1);
         command = new ChangeMapProperty(mMapDocument, orientation);
@@ -933,28 +1026,29 @@ QUndoCommand *PropertyBrowser::applyMapObjectValueTo(PropertyId id, const QVaria
         command = new ResizeMapObject(mMapDocument, mapObject, newSize, oldSize);
         break;
     }
-    case RotationProperty: {
-        const qreal newRotation = val.toDouble();
-        const qreal oldRotation = mapObject->rotation();
-        command = new RotateMapObject(mMapDocument, mapObject, newRotation, oldRotation);
+    case RotationProperty:
+        if (mapObject->canRotate()) {
+            const qreal newRotation = val.toDouble();
+            const qreal oldRotation = mapObject->rotation();
+            command = new RotateMapObject(mMapDocument, mapObject, newRotation, oldRotation);
+        }
         break;
-    }
     case FlippingProperty: {
         const int flippingFlags = val.toInt();
-        const bool flippedHorizontally = flippingFlags & 1;
-        const bool flippedVertically = flippingFlags & 2;
-
-        // You can only change one checkbox at a time
-        Cell newCell = mapObject->cell();
-        newCell.setFlippedHorizontally(flippedHorizontally);
-        newCell.setFlippedVertically(flippedVertically);
 
         MapObjectCell mapObjectCell;
         mapObjectCell.object = mapObject;
-        mapObjectCell.cell = newCell;
+        mapObjectCell.cell = mapObject->cell();
+        mapObjectCell.cell.setFlippedHorizontally(flippingFlags & 1);
+        mapObjectCell.cell.setFlippedVertically(flippingFlags & 2);
 
         command = new ChangeMapObjectCells(mMapDocument,
-                                           QVector<MapObjectCell> () << mapObjectCell);
+                                           QVector<MapObjectCell>() << mapObjectCell);
+
+        command->setText(QCoreApplication::translate("Undo Commands",
+                                                     "Flip %n Object(s)",
+                                                     nullptr,
+                                                     mMapDocument->selectedObjects().size()));
         break;
     }
     }
@@ -995,6 +1089,9 @@ void PropertyBrowser::applyLayerValue(PropertyId id, const QVariant &val)
     case VisibleProperty:
         command = new SetLayerVisible(mMapDocument, layer, val.toBool());
         break;
+    case LockedProperty:
+        command = new SetLayerLocked(mMapDocument, layer, val.toBool());
+        break;
     case OpacityProperty:
         command = new SetLayerOpacity(mMapDocument, layer, val.toDouble());
         break;
@@ -1072,12 +1169,12 @@ void PropertyBrowser::applyImageLayerValue(PropertyId id, const QVariant &val)
         undoStack->push(new ChangeImageLayerProperties(mMapDocument,
                                                        imageLayer,
                                                        color,
-                                                       imageSource.absolutePath));
+                                                       imageSource.url));
         break;
     }
     case ColorProperty: {
         const QColor color = val.value<QColor>();
-        const QString &imageSource = imageLayer->imageSource();
+        const QUrl &imageSource = imageLayer->imageSource();
         undoStack->push(new ChangeImageLayerProperties(mMapDocument,
                                                        imageLayer,
                                                        color,
@@ -1104,7 +1201,7 @@ void PropertyBrowser::applyTilesetValue(PropertyId id, const QVariant &val)
     case FileNameProperty: {
         FilePath filePath = val.value<FilePath>();
         QString error;
-        SharedTileset newTileset = TilesetManager::instance()->loadTileset(filePath.absolutePath, &error);
+        SharedTileset newTileset = TilesetManager::instance()->loadTileset(filePath.url.toLocalFile(), &error);
         if (!newTileset) {
             QMessageBox::critical(window(), tr("Error Reading Tileset"), error);
             return;
@@ -1184,7 +1281,7 @@ void PropertyBrowser::applyTileValue(PropertyId id, const QVariant &val)
     case ImageSourceProperty: {
         const FilePath filePath = val.value<FilePath>();
         undoStack->push(new ChangeTileImageSource(mTilesetDocument,
-                                                  tile, filePath.absolutePath));
+                                                  tile, filePath.url));
         break;
     }
     default:
@@ -1206,6 +1303,65 @@ void PropertyBrowser::applyTerrainValue(PropertyId id, const QVariant &val)
     }
 }
 
+void PropertyBrowser::applyWangSetValue(PropertyId id, const QVariant &val)
+{
+    Q_ASSERT(mTilesetDocument);
+
+    WangSet *wangSet = static_cast<WangSet*>(mObject);
+
+    switch (id) {
+    case NameProperty:
+        mDocument->undoStack()->push(new RenameWangSet(mTilesetDocument,
+                                                       mTilesetDocument->tileset()->wangSets().indexOf(wangSet),
+                                                       val.toString()));
+        break;
+    case EdgeCountProperty:
+        mDocument->undoStack()->push(new ChangeWangSetEdges(mTilesetDocument,
+                                                            mTilesetDocument->tileset()->wangSets().indexOf(wangSet),
+                                                            val.toInt()));
+        break;
+    case CornerCountProperty:
+        mDocument->undoStack()->push(new ChangeWangSetCorners(mTilesetDocument,
+                                                              mTilesetDocument->tileset()->wangSets().indexOf(wangSet),
+                                                              val.toInt()));
+        break;
+    default:
+        break;
+    }
+}
+
+void PropertyBrowser::applyWangColorValue(PropertyId id, const QVariant &val)
+{
+    Q_ASSERT(mTilesetDocument);
+
+    WangColor *wangColor = static_cast<WangColor*>(mObject);
+    WangColorModel *wangColorModel = mTilesetDocument->wangColorModel();
+    Q_ASSERT(wangColorModel);
+
+    switch (id) {
+    case NameProperty:
+        mDocument->undoStack()->push(new ChangeWangColorName(val.toString(),
+                                                             wangColor->colorIndex(),
+                                                             wangColor->isEdge(),
+                                                             wangColorModel));
+        break;
+    case ColorProperty:
+        mDocument->undoStack()->push(new ChangeWangColorColor(val.value<QColor>(),
+                                                              wangColor->colorIndex(),
+                                                              wangColor->isEdge(),
+                                                              wangColorModel));
+        break;
+    case WangColorProbabilityProperty:
+        mDocument->undoStack()->push(new ChangeWangColorProbability(val.toDouble(),
+                                                                    wangColor->colorIndex(),
+                                                                    wangColor->isEdge(),
+                                                                    wangColorModel));
+        break;
+    default:
+        break;
+    }
+}
+
 /**
  * @warning This function does not add the property to the view.
  */
@@ -1333,6 +1489,9 @@ void PropertyBrowser::addProperties()
     case Object::TilesetType:           addTilesetProperties(); break;
     case Object::TileType:              addTileProperties(); break;
     case Object::TerrainType:           addTerrainProperties(); break;
+    case Object::WangSetType:           addWangSetProperties(); break;
+    case Object::WangColorType:         addWangColorProperties(); break;
+    case Object::ObjectTemplateType:    break;
     }
 
     // Make sure the color and font properties are collapsed, to save space
@@ -1376,6 +1535,7 @@ void PropertyBrowser::updateProperties()
         mIdToProperty[HeightProperty]->setValue(map->height());
         mIdToProperty[TileWidthProperty]->setValue(map->tileWidth());
         mIdToProperty[TileHeightProperty]->setValue(map->tileHeight());
+        mIdToProperty[InfiniteProperty]->setValue(map->infinite());
         mIdToProperty[OrientationProperty]->setValue(map->orientation() - 1);
         mIdToProperty[HexSideLengthProperty]->setValue(map->hexSideLength());
         mIdToProperty[StaggerAxisProperty]->setValue(map->staggerAxis());
@@ -1387,12 +1547,24 @@ void PropertyBrowser::updateProperties()
     }
     case Object::MapObjectType: {
         const MapObject *mapObject = static_cast<const MapObject*>(mObject);
+        const int flags = mapObjectFlags(mapObject);
+
+        if (mMapObjectFlags != flags) {
+            removeProperties();
+            addProperties();
+            return;
+        }
 
         const QString &type = mapObject->effectiveType();
         const auto typeColorGroup = mapObject->type().isEmpty() ? QPalette::Disabled
                                                                 : QPalette::Active;
 
+        FilePath templateFilePath;
+        if (auto objectTemplate = mapObject->objectTemplate())
+            templateFilePath.url = QUrl::fromLocalFile(objectTemplate->fileName());
+
         mIdToProperty[IdProperty]->setValue(mapObject->id());
+        mIdToProperty[TemplateProperty]->setValue(QVariant::fromValue(templateFilePath));
         mIdToProperty[NameProperty]->setValue(mapObject->name());
         mIdToProperty[TypeProperty]->setValue(type);
         mIdToProperty[TypeProperty]->setValueColor(palette().color(typeColorGroup, QPalette::WindowText));
@@ -1400,23 +1572,23 @@ void PropertyBrowser::updateProperties()
         mIdToProperty[XProperty]->setValue(mapObject->x());
         mIdToProperty[YProperty]->setValue(mapObject->y());
 
-        if (!mapObject->isPolyShape()) {
+        if (flags & ObjectHasDimensions) {
             mIdToProperty[WidthProperty]->setValue(mapObject->width());
             mIdToProperty[HeightProperty]->setValue(mapObject->height());
         }
 
         mIdToProperty[RotationProperty]->setValue(mapObject->rotation());
 
-        if (QtVariantProperty *property = mIdToProperty[FlippingProperty]) {
+        if (flags & ObjectHasTile) {
             int flippingFlags = 0;
             if (mapObject->cell().flippedHorizontally())
                 flippingFlags |= 1;
             if (mapObject->cell().flippedVertically())
                 flippingFlags |= 2;
-            property->setValue(flippingFlags);
+            mIdToProperty[FlippingProperty]->setValue(flippingFlags);
         }
 
-        if (mapObject->shape() == MapObject::Text) {
+        if (flags & ObjectIsText) {
             const auto& textData = mapObject->textData();
             mIdToProperty[TextProperty]->setValue(textData.text);
             mIdToProperty[FontProperty]->setValue(textData.font);
@@ -1431,6 +1603,7 @@ void PropertyBrowser::updateProperties()
 
         mIdToProperty[NameProperty]->setValue(layer->name());
         mIdToProperty[VisibleProperty]->setValue(layer->isVisible());
+        mIdToProperty[LockedProperty]->setValue(layer->isLocked());
         mIdToProperty[OpacityProperty]->setValue(layer->opacity());
         mIdToProperty[OffsetXProperty]->setValue(layer->offset().x());
         mIdToProperty[OffsetYProperty]->setValue(layer->offset().y());
@@ -1460,7 +1633,7 @@ void PropertyBrowser::updateProperties()
         Tileset *tileset = static_cast<Tileset*>(mObject);
 
         if (QtVariantProperty *fileNameProperty = mIdToProperty.value(FileNameProperty))
-            fileNameProperty->setValue(QVariant::fromValue(FilePath { tileset->fileName() }));
+            fileNameProperty->setValue(QVariant::fromValue(FilePath { QUrl::fromLocalFile(tileset->fileName()) }));
 
         mIdToProperty[BackgroundColorProperty]->setValue(tileset->backgroundColor());
 
@@ -1474,7 +1647,7 @@ void PropertyBrowser::updateProperties()
 
         if (!tileset->isCollection()) {
             mIdToProperty[TilesetImageParametersProperty]->setValue(QVariant::fromValue(mTilesetDocument));
-            mIdToProperty[ImageSourceProperty]->setValue(tileset->imageSource());
+            mIdToProperty[ImageSourceProperty]->setValue(tileset->imageSource().toString(QUrl::PreferLocalFile));
             mIdToProperty[TileWidthProperty]->setValue(tileset->tileWidth());
             mIdToProperty[TileHeightProperty]->setValue(tileset->tileHeight());
             mIdToProperty[MarginProperty]->setValue(tileset->margin());
@@ -1500,6 +1673,22 @@ void PropertyBrowser::updateProperties()
         mIdToProperty[NameProperty]->setValue(terrain->name());
         break;
     }
+    case Object::WangSetType: {
+        const WangSet *wangSet = static_cast<const WangSet*>(mObject);
+        mIdToProperty[NameProperty]->setValue(wangSet->name());
+        mIdToProperty[EdgeCountProperty]->setValue(wangSet->edgeColorCount());
+        mIdToProperty[CornerCountProperty]->setValue(wangSet->cornerColorCount());
+        break;
+    }
+    case Object::WangColorType: {
+        const WangColor *wangColor = static_cast<const WangColor*>(mObject);
+        mIdToProperty[NameProperty]->setValue(wangColor->name());
+        mIdToProperty[ColorProperty]->setValue(wangColor->color());
+        mIdToProperty[WangColorProbabilityProperty]->setValue(wangColor->probability());
+        break;
+    }
+    case Object::ObjectTemplateType:
+        break;
     }
 
     mUpdating = false;
@@ -1541,6 +1730,16 @@ void PropertyBrowser::updateCustomProperties()
         auto mapObject = static_cast<MapObject*>(mObject);
         objectType = mapObject->type();
 
+        // Inherit properties from the template
+        if (const MapObject *templateObject = mapObject->templateObject()) {
+            QMapIterator<QString,QVariant> it(templateObject->properties());
+            while (it.hasNext()) {
+                it.next();
+                if (!mCombinedProperties.contains(it.key()))
+                    mCombinedProperties.insert(it.key(), it.value());
+            }
+        }
+
         if (Tile *tile = mapObject->cell().tile()) {
             if (objectType.isEmpty())
                 objectType = tile->type();
@@ -1559,13 +1758,15 @@ void PropertyBrowser::updateCustomProperties()
     case Object::MapType:
     case Object::TerrainType:
     case Object::TilesetType:
+    case Object::WangSetType:
+    case Object::WangColorType:
+    case Object::ObjectTemplateType:
         break;
     }
 
     if (!objectType.isEmpty()) {
         // Inherit properties from the object type
-        const ObjectTypes objectTypes = Preferences::instance()->objectTypes();
-        for (const ObjectType &type : objectTypes) {
+        for (const ObjectType &type : Object::objectTypes()) {
             if (type.name == objectType) {
                 QMapIterator<QString,QVariant> it(type.defaultProperties);
                 while (it.hasNext()) {
@@ -1634,5 +1835,51 @@ void PropertyBrowser::updateCustomPropertyColor(const QString &name)
     property->setValueColor(textColor);
 }
 
+void PropertyBrowser::retranslateUi()
+{
+    mStaggerAxisNames.clear();
+    mStaggerIndexNames.clear();
+    mOrientationNames.clear();
+    mTilesetOrientationNames.clear();
+    mLayerFormatNames.clear();
+    mRenderOrderNames.clear();
+    mFlippingFlagNames.clear();
+    mDrawOrderNames.clear();
+
+    mStaggerAxisNames.append(tr("X"));
+    mStaggerAxisNames.append(tr("Y"));
+
+    mStaggerIndexNames.append(tr("Odd"));
+    mStaggerIndexNames.append(tr("Even"));
+
+    mOrientationNames.append(QCoreApplication::translate("Tiled::Internal::NewMapDialog", "Orthogonal"));
+    mOrientationNames.append(QCoreApplication::translate("Tiled::Internal::NewMapDialog", "Isometric"));
+    mOrientationNames.append(QCoreApplication::translate("Tiled::Internal::NewMapDialog", "Isometric (Staggered)"));
+    mOrientationNames.append(QCoreApplication::translate("Tiled::Internal::NewMapDialog", "Hexagonal (Staggered)"));
+
+    mTilesetOrientationNames.append(mOrientationNames.at(0));
+    mTilesetOrientationNames.append(mOrientationNames.at(1));
+
+    mLayerFormatNames.append(QCoreApplication::translate("PreferencesDialog", "XML"));
+    mLayerFormatNames.append(QCoreApplication::translate("PreferencesDialog", "Base64 (uncompressed)"));
+    mLayerFormatNames.append(QCoreApplication::translate("PreferencesDialog", "Base64 (gzip compressed)"));
+    mLayerFormatNames.append(QCoreApplication::translate("PreferencesDialog", "Base64 (zlib compressed)"));
+    mLayerFormatNames.append(QCoreApplication::translate("PreferencesDialog", "CSV"));
+
+    mRenderOrderNames.append(QCoreApplication::translate("PreferencesDialog", "Right Down"));
+    mRenderOrderNames.append(QCoreApplication::translate("PreferencesDialog", "Right Up"));
+    mRenderOrderNames.append(QCoreApplication::translate("PreferencesDialog", "Left Down"));
+    mRenderOrderNames.append(QCoreApplication::translate("PreferencesDialog", "Left Up"));
+
+    mFlippingFlagNames.append(tr("Horizontal"));
+    mFlippingFlagNames.append(tr("Vertical"));
+
+    mDrawOrderNames.append(tr("Top Down"));
+    mDrawOrderNames.append(tr("Manual"));
+
+    removeProperties();
+    addProperties();
+}
+
 } // namespace Internal
 } // namespace Tiled
