@@ -35,7 +35,6 @@
 #include "preferences.h"
 #include "replacetileset.h"
 #include "swaptiles.h"
-#include "changemapobject.h"
 #include "terrain.h"
 #include "tile.h"
 #include "tilelayer.h"
@@ -58,6 +57,7 @@
 #include <QMenu>
 #include <QMessageBox>
 #include <QMimeData>
+#include <QPushButton>
 #include <QSettings>
 #include <QSignalMapper>
 #include <QStackedWidget>
@@ -74,10 +74,30 @@ using namespace Tiled::Internal;
 
 namespace {
 
+class NoTilesetWidget : public QWidget
+{
+    Q_OBJECT
+
+public:
+    explicit NoTilesetWidget(QWidget *parent = nullptr)
+        : QWidget(parent)
+    {
+        QPushButton *newTilesetButton = new QPushButton(this);
+        newTilesetButton->setText(tr("New Tileset..."));
+
+        QGridLayout *gridLayout = new QGridLayout(this);
+        gridLayout->addWidget(newTilesetButton, 0, 0, Qt::AlignCenter);
+
+        connect(newTilesetButton, &QPushButton::clicked, [] {
+            ActionManager::action("file.new_tileset")->trigger();
+        });
+    }
+};
+
 class TilesetMenuButton : public QToolButton
 {
 public:
-    TilesetMenuButton(QWidget *parent = nullptr)
+    explicit TilesetMenuButton(QWidget *parent = nullptr)
         : QToolButton(parent)
     {
         setArrowType(Qt::DownArrow);
@@ -112,7 +132,7 @@ protected:
 class WheelEnabledTabBar : public QTabBar
 {
 public:
-    WheelEnabledTabBar(QWidget *parent = nullptr)
+    explicit WheelEnabledTabBar(QWidget *parent = nullptr)
        : QTabBar(parent)
     {}
 
@@ -131,21 +151,36 @@ public:
 static void removeTileReferences(MapDocument *mapDocument,
                                  std::function<bool(const Cell &)> condition)
 {
-    QUndoStack *undoStack = mapDocument->undoStack();
+    // Commands delayed to avoid invalidating object list iterator
+    QList<QUndoCommand*> commands;
 
-    for (Layer *layer : mapDocument->map()->layers()) {
-        if (TileLayer *tileLayer = layer->asTileLayer()) {
+    LayerIterator it(mapDocument->map());
+    while (Layer *layer = it.next()) {
+        switch (layer->layerType()) {
+        case Layer::TileLayerType: {
+            auto tileLayer = static_cast<TileLayer*>(layer);
             const QRegion refs = tileLayer->region(condition);
             if (!refs.isEmpty())
-                undoStack->push(new EraseTiles(mapDocument, tileLayer, refs));
-
-        } else if (ObjectGroup *objectGroup = layer->asObjectGroup()) {
+                commands.append(new EraseTiles(mapDocument, tileLayer, refs));
+            break;
+        }
+        case Layer::ObjectGroupType: {
+            auto objectGroup = static_cast<ObjectGroup*>(layer);
             for (MapObject *object : *objectGroup) {
                 if (condition(object->cell()))
-                    undoStack->push(new RemoveMapObject(mapDocument, object));
+                    commands.append(new RemoveMapObject(mapDocument, object));
             }
+            break;
+        }
+        case Layer::ImageLayerType:
+        case Layer::GroupLayerType:
+            break;
         }
     }
+
+    QUndoStack *undoStack = mapDocument->undoStack();
+    for (QUndoCommand *command : commands)
+        undoStack->push(command);
 }
 
 } // anonymous namespace
@@ -155,6 +190,7 @@ TilesetDock::TilesetDock(QWidget *parent)
     , mMapDocument(nullptr)
     , mTilesetDocumentsFilterModel(new TilesetDocumentsFilterModel(this))
     , mTabBar(new WheelEnabledTabBar)
+    , mSuperViewStack(new QStackedWidget)
     , mViewStack(new QStackedWidget)
     , mToolBar(new QToolBar)
     , mCurrentTile(nullptr)
@@ -175,9 +211,12 @@ TilesetDock::TilesetDock(QWidget *parent)
 
     mTabBar->setUsesScrollButtons(true);
     mTabBar->setExpanding(false);
+    mTabBar->setContextMenuPolicy(Qt::CustomContextMenu);
 
     connect(mTabBar, &QTabBar::currentChanged, this, &TilesetDock::updateActions);
     connect(mTabBar, &QTabBar::tabMoved, this, &TilesetDock::onTabMoved);
+    connect(mTabBar, &QWidget::customContextMenuRequested,
+            this, &TilesetDock::tabContextMenuRequested);
 
     QWidget *w = new QWidget(this);
 
@@ -190,7 +229,10 @@ TilesetDock::TilesetDock(QWidget *parent)
     vertical->setSpacing(0);
     vertical->setMargin(0);
     vertical->addLayout(horizontal);
-    vertical->addWidget(mViewStack);
+    vertical->addWidget(mSuperViewStack);
+
+    mSuperViewStack->insertWidget(0, new NoTilesetWidget(this));
+    mSuperViewStack->insertWidget(1, mViewStack);
 
     horizontal = new QHBoxLayout;
     horizontal->setSpacing(0);
@@ -509,6 +551,9 @@ void TilesetDock::createTilesetView(int index, TilesetDocument *tilesetDocument)
 
     TilesetView *view = new TilesetView;
 
+    // Hides the "New Tileset..." special view if it is shown.
+    mSuperViewStack->setCurrentIndex(1);
+
     // Insert view before the tab to make sure it is there when the tab index
     // changes (happens when first tab is inserted).
     mViewStack->insertWidget(index, view);
@@ -532,8 +577,6 @@ void TilesetDock::createTilesetView(int index, TilesetDocument *tilesetDocument)
             this, &TilesetDock::updateCurrentTiles);
     connect(view, &TilesetView::swapTilesRequested,
             this, &TilesetDock::swapTiles);
-    connect(view, &TilesetView::changeSelectedMapObjectsTileRequested,
-            this, &TilesetDock::changeSelectedMapObjectsTile);
 }
 
 void TilesetDock::deleteTilesetView(int index)
@@ -556,6 +599,10 @@ void TilesetDock::deleteTilesetView(int index)
     delete view;                    // view needs to go before the tab
     mTabBar->removeTab(index);
 
+    // Make the "New Tileset..." special tab reappear if there is no tileset open
+    if (mTilesets.isEmpty())
+        mSuperViewStack->setCurrentIndex(0);
+
     // Make sure we don't reference this tileset anymore
     if (mCurrentTiles && mCurrentTiles->referencesTileset(tileset)) {
         TileLayer *cleaned = mCurrentTiles->clone();
@@ -777,6 +824,21 @@ void TilesetDock::onTabMoved(int from, int to)
     mViewStack->insertWidget(to, widget);
 }
 
+void TilesetDock::tabContextMenuRequested(const QPoint &pos)
+{
+    int index = mTabBar->tabAt(pos);
+    if (index == -1)
+        return;
+
+    QMenu menu;
+
+    QString fileName = mTilesetDocuments.at(index)->fileName();
+    if (!fileName.isEmpty())
+        Utils::addFileManagerActions(menu, fileName);
+
+    menu.exec(mTabBar->mapToGlobal(pos));
+}
+
 Tileset *TilesetDock::currentTileset() const
 {
     const int index = mTabBar->currentIndex();
@@ -984,34 +1046,4 @@ void TilesetDock::swapTiles(Tile *tileA, Tile *tileB)
     undoStack->push(new SwapTiles(mMapDocument, tileA, tileB));
 }
 
-static QList<MapObject*> chooseTileObjects(QList<MapObject*> objects)
-{
-    QList<MapObject*> tileObjects;
-
-    for (auto object : objects)
-        if (!object->cell().isEmpty())
-            tileObjects.append(object);
-
-    return tileObjects;
-}
-
-void TilesetDock::changeSelectedMapObjectsTile(Tile *tile)
-{
-    if (!mMapDocument)
-        return;
-
-    // Only change tiles of tile objects
-    QList<MapObject*> tileObjects = chooseTileObjects(mMapDocument->selectedObjects());
-
-    if (tileObjects.isEmpty())
-        return;
-
-    auto changeMapObjectCommand = new ChangeMapObjectsTile(mMapDocument, tileObjects, tile);
-
-    // Make sure the tileset is part of the map
-    SharedTileset sharedTileset = tile->tileset()->sharedPointer();
-    if (!mMapDocument->map()->tilesets().contains(sharedTileset))
-        new AddTileset(mMapDocument, sharedTileset, changeMapObjectCommand);
-
-    mMapDocument->undoStack()->push(changeMapObjectCommand);
-}
+#include "tilesetdock.moc"
