@@ -23,10 +23,11 @@
 #include "mapdocument.h"
 #include "map.h"
 #include "terrain.h"
+#include "wangset.h"
 #include "tile.h"
 #include "tilesetformat.h"
-#include "tilesetmanager.h"
 #include "tilesetterrainmodel.h"
+#include "tilesetwangsetmodel.h"
 
 #include <QCoreApplication>
 #include <QFileInfo>
@@ -58,6 +59,8 @@ TilesetDocument::TilesetDocument(const SharedTileset &tileset, const QString &fi
     : Document(TilesetDocumentType, fileName)
     , mTileset(tileset)
     , mTerrainModel(new TilesetTerrainModel(this, this))
+    , mWangSetModel(new TilesetWangSetModel(this, this))
+    , mWangColorModel(nullptr)
 {
     mCurrentObject = tileset.data();
 
@@ -76,26 +79,21 @@ TilesetDocument::TilesetDocument(const SharedTileset &tileset, const QString &fi
     connect(mTerrainModel, &TilesetTerrainModel::terrainRemoved,
             this, &TilesetDocument::onTerrainRemoved);
 
-    TilesetManager *tilesetManager = TilesetManager::instance();
-    tilesetManager->addReference(tileset);
+    connect(mWangSetModel, &TilesetWangSetModel::wangSetRemoved,
+            this, &TilesetDocument::onWangSetRemoved);
 }
 
 TilesetDocument::~TilesetDocument()
 {
-    TilesetManager *tilesetManager = TilesetManager::instance();
-    tilesetManager->removeReference(mTileset);
 }
 
 bool TilesetDocument::save(const QString &fileName, QString *error)
 {
     TilesetFormat *tilesetFormat = mTileset->format();
 
-    if (!tilesetFormat || !(tilesetFormat->capabilities() & MapFormat::Write))
+    if (!tilesetFormat || !(tilesetFormat->capabilities() & FileFormat::Write))
         return false;
 
-    // todo: workaround to avoid writing the tileset like an external tileset reference
-    mTileset->setFileName(QString());
-
     if (!tilesetFormat->write(*tileset(), fileName)) {
         if (error)
             *error = tilesetFormat->errorString();
@@ -169,6 +167,17 @@ void TilesetDocument::setWriterFormat(TilesetFormat *format)
     mTileset->setFormat(format);
 }
 
+TilesetFormat* TilesetDocument::exportFormat() const
+{
+    return mExportFormat;
+}
+
+void TilesetDocument::setExportFormat(FileFormat *format)
+{
+    mExportFormat = qobject_cast<TilesetFormat*>(format);
+    Q_ASSERT(mExportFormat);
+}
+
 QString TilesetDocument::displayName() const
 {
     QString displayName;
@@ -189,7 +198,7 @@ QString TilesetDocument::displayName() const
 
 /**
  * Exchanges the tileset data of the tileset wrapped by this document with the
- * data in the given \a tileset, and vica-versa.
+ * data in the given \a tileset, and vice-versa.
  */
 void TilesetDocument::swapTileset(SharedTileset &tileset)
 {
@@ -293,7 +302,7 @@ void TilesetDocument::setTileType(Tile *tile, const QString &type)
         emit mapDocument->tileTypeChanged(tile);
 }
 
-void TilesetDocument::setTileImage(Tile *tile, const QPixmap &image, const QString &source)
+void TilesetDocument::setTileImage(Tile *tile, const QPixmap &image, const QUrl &source)
 {
     Q_ASSERT(tile->tileset() == mTileset.data());
 
@@ -335,5 +344,11 @@ void TilesetDocument::onTerrainRemoved(Terrain *terrain)
         setCurrentObject(nullptr);
 }
 
+void TilesetDocument::onWangSetRemoved(WangSet *wangSet)
+{
+    if (wangSet == mCurrentObject)
+        setCurrentObject(nullptr);
+}
+
 } // namespace Internal
 } // namespace Tiled
