@@ -21,6 +21,7 @@
 #include "tilesetview.h"
 
 #include "changetileterrain.h"
+#include "changetilewangid.h"
 #include "map.h"
 #include "preferences.h"
 #include "stylehelper.h"
@@ -176,10 +177,12 @@ static void paintCorners(QPainter *painter,
 static void setCosmeticPen(QPainter *painter, const QBrush &brush, qreal width)
 {
 #if QT_VERSION >= 0x050600
-    QPen pen(brush, width * painter->device()->devicePixelRatioF());
+    const qreal devicePixelRatio = painter->device()->devicePixelRatioF();
 #else
-    QPen pen(brush, width * painter->device()->devicePixelRatio());
+    const int devicePixelRatio = painter->device()->devicePixelRatio();
 #endif
+
+    QPen pen(brush, width * devicePixelRatio);
     pen.setCosmetic(true);
     painter->setPen(pen);
 }
@@ -227,7 +230,7 @@ static QTransform tilesetGridTransform(const Tileset &tileset, QPoint tileCenter
 
         transform.translate(tileCenter.x(), tileCenter.y());
 
-        const auto ratio = (double) gridSize.height() / gridSize.width();
+        const auto ratio = (qreal) gridSize.height() / gridSize.width();
         const auto scaleX = 1.0 / sqrt(2.0);
         const auto scaleY = scaleX * ratio;
         transform.scale(scaleX, scaleY);
@@ -240,6 +243,287 @@ static QTransform tilesetGridTransform(const Tileset &tileset, QPoint tileCenter
     return transform;
 }
 
+static void setWangStyle(QPainter *painter, WangSet *wangSet, int index, bool edge)
+{
+    QColor c;
+    if (edge)
+        c = wangSet->edgeColorAt(index)->color();
+    else
+        c = wangSet->cornerColorAt(index)->color();
+
+    painter->setBrush(QColor(c.red(), c.green(), c.blue(), 200));
+    setCosmeticPen(painter, c, 2);
+}
+
+static void paintWangOverlay(QPainter *painter,
+                             WangId wangId,
+                             WangSet *wangSet,
+                             const QRect &rect)
+{
+    painter->save();
+    painter->setClipRect(rect);
+    painter->setRenderHint(QPainter::Antialiasing);
+
+    //arbitrary fraction, could be made constant.
+    int thicknessW = rect.width()/6;
+    int thicknessH = rect.height()/6;
+
+    if (wangSet->edgeColorCount() > 1) {
+        if (wangSet->cornerColorCount() > 1) {
+            QRect wRect;
+            int edge;
+
+            //top
+            edge = wangId.edgeColor(0);
+            if (edge > 0) {
+                setWangStyle(painter, wangSet, edge, true);
+
+                wRect = QRect(QPoint(rect.left() + rect.width()/3, rect.top()),
+                              QPoint(rect.right() - rect.width()/3, rect.top() + thicknessH));
+                painter->drawRect(wRect);
+            }
+
+            //right
+            edge = wangId.edgeColor(1);
+            if (edge > 0) {
+                setWangStyle(painter, wangSet, edge, true);
+
+                wRect = QRect(QPoint(rect.right() - thicknessW, rect.top() + rect.height()/3),
+                              QPoint(rect.right(), rect.bottom() - rect.height()/3));
+                painter->drawRect(wRect);
+            }
+
+            //bottom
+            edge = wangId.edgeColor(2);
+            if (edge > 0) {
+                setWangStyle(painter, wangSet, edge, true);
+
+                wRect = QRect(QPoint(rect.left() + rect.width()/3, rect.bottom() - thicknessH),
+                              QPoint(rect.right() - rect.width()/3, rect.bottom()));
+                painter->drawRect(wRect);
+            }
+
+            //left
+            edge = wangId.edgeColor(3);
+            if (edge > 0) {
+                setWangStyle(painter, wangSet, edge, true);
+
+                wRect = QRect(QPoint(rect.left(), rect.top() + rect.height()/3),
+                              QPoint(rect.left() + thicknessW, rect.bottom() - rect.height()/3));
+                painter->drawRect(wRect);
+            }
+        } else {
+            int edge;
+
+            //top
+            edge = wangId.edgeColor(0);
+            if (edge > 0) {
+                setWangStyle(painter, wangSet, edge, true);
+
+                const QPoint points[] = {
+                    rect.topLeft(),
+                    rect.topRight(),
+                    rect.topRight() + QPoint(-thicknessW, thicknessH),
+                    rect.topLeft() + QPoint(thicknessW, thicknessH)
+                };
+
+                painter->drawPolygon(points, 4);
+            }
+
+            //right
+            edge = wangId.edgeColor(1);
+            if (edge > 0) {
+                setWangStyle(painter, wangSet, edge, true);
+
+                const QPoint points[] = {
+                    rect.topRight(),
+                    rect.bottomRight(),
+                    rect.bottomRight() + QPoint(-thicknessW, -thicknessH),
+                    rect.topRight() + QPoint(-thicknessW, thicknessH)
+                };
+
+                painter->drawPolygon(points, 4);
+            }
+
+            //bottom
+            edge = wangId.edgeColor(2);
+            if (edge > 0) {
+                setWangStyle(painter, wangSet, edge, true);
+
+                const QPoint points[] = {
+                    rect.bottomRight(),
+                    rect.bottomLeft(),
+                    rect.bottomLeft() + QPoint(thicknessW, -thicknessH),
+                    rect.bottomRight() + QPoint(-thicknessW, -thicknessH)
+                };
+
+                painter->drawPolygon(points, 4);
+            }
+
+            //left
+            edge = wangId.edgeColor(3);
+            if (edge > 0) {
+                setWangStyle(painter, wangSet, edge, true);
+
+                const QPoint points[] = {
+                    rect.topLeft(),
+                    rect.bottomLeft(),
+                    rect.bottomLeft() + QPoint(thicknessW, -thicknessH),
+                    rect.topLeft() + QPoint(thicknessW, thicknessH)
+                };
+
+                painter->drawPolygon(points, 4);
+            }
+        }
+    }
+
+    if (wangSet->cornerColorCount() > 1) {
+        if (wangSet->edgeColorCount() > 1) {
+            int corner;
+
+            //top right
+            corner = wangId.cornerColor(0);
+            if (corner > 0) {
+                setWangStyle(painter, wangSet, corner, false);
+
+                const QPoint points[] = {
+                    rect.topRight(),
+                    QPoint(rect.right(), rect.top() + rect.height()/3),
+                    QPoint(rect.right() - thicknessW, rect.top() + rect.height()/3),
+                    rect.topRight() + QPoint(-thicknessW, thicknessH),
+                    QPoint(rect.right() - rect.width()/3, rect.top() + thicknessH),
+                    QPoint(rect.right() - rect.width()/3, rect.top())
+                };
+
+                painter->drawPolygon(points, 6);
+            }
+
+            //bottom right
+            corner = wangId.cornerColor(1);
+            if (corner > 0) {
+                setWangStyle(painter, wangSet, corner, false);
+
+                const QPoint points[] = {
+                    rect.bottomRight(),
+                    QPoint(rect.right(), rect.bottom() - rect.height()/3),
+                    QPoint(rect.right() - thicknessW, rect.bottom() - rect.height()/3),
+                    rect.bottomRight() + QPoint(-thicknessW, -thicknessH),
+                    QPoint(rect.right() - rect.width()/3, rect.bottom() - thicknessH),
+                    QPoint(rect.right() - rect.width()/3, rect.bottom())
+                };
+
+                painter->drawPolygon(points, 6);
+            }
+
+            //bottom left
+            corner = wangId.cornerColor(2);
+            if (corner > 0) {
+                setWangStyle(painter, wangSet, corner, false);
+
+                const QPoint points[] = {
+                    rect.bottomLeft(),
+                    QPoint(rect.left(), rect.bottom() - rect.height()/3),
+                    QPoint(rect.left() + thicknessW, rect.bottom() - rect.height()/3),
+                    rect.bottomLeft() + QPoint(thicknessW, -thicknessH),
+                    QPoint(rect.left() + rect.width()/3, rect.bottom() - thicknessH),
+                    QPoint(rect.left() + rect.width()/3, rect.bottom()),
+                };
+
+                painter->drawPolygon(points, 6);
+            }
+
+            //top left
+            corner = wangId.cornerColor(3);
+            if (corner > 0) {
+                setWangStyle(painter, wangSet, corner, false);
+
+                const QPoint points[] = {
+                    rect.topLeft(),
+                    QPoint(rect.left(), rect.top() + rect.height()/3),
+                    QPoint(rect.left() + thicknessW, rect.top() + rect.height()/3),
+                    rect.topLeft() + QPoint(thicknessW, thicknessH),
+                    QPoint(rect.left() + rect.width()/3, rect.top() + thicknessH),
+                    QPoint(rect.left() + rect.width()/3, rect.top())
+                };
+
+                painter->drawPolygon(points, 6);
+            }
+        } else {
+            int corner;
+
+            //top right
+            corner = wangId.cornerColor(0);
+            if (corner > 0) {
+                setWangStyle(painter, wangSet, corner, false);
+
+                const QPoint points[] = {
+                    rect.topRight(),
+                    QPoint(rect.right(), rect.center().y()),
+                    QPoint(rect.right() - thicknessW, rect.center().y()),
+                    rect.topRight() + QPoint(-thicknessW, thicknessH),
+                    QPoint(rect.center().x(), rect.top() + thicknessH),
+                    QPoint(rect.center().x(), rect.top())
+                };
+
+                painter->drawPolygon(points, 6);
+            }
+
+            //bottom right
+            corner = wangId.cornerColor(1);
+            if (corner > 0) {
+                setWangStyle(painter, wangSet, corner, false);
+
+                const QPoint points[] = {
+                    rect.bottomRight(),
+                    QPoint(rect.right(), rect.center().y()),
+                    QPoint(rect.right() - thicknessW, rect.center().y()),
+                    rect.bottomRight() + QPoint(-thicknessW, -thicknessH),
+                    QPoint(rect.center().x(), rect.bottom() - thicknessH),
+                    QPoint(rect.center().x(), rect.bottom()),
+                };
+
+                painter->drawPolygon(points, 6);
+            }
+
+            //top left
+            corner = wangId.cornerColor(3);
+            if (corner > 0) {
+                setWangStyle(painter, wangSet, corner, false);
+
+                const QPoint points[] = {
+                    rect.topLeft(),
+                    QPoint(rect.left(), rect.center().y()),
+                    QPoint(rect.left() + thicknessW, rect.center().y()),
+                    rect.topLeft() + QPoint(thicknessW, thicknessH),
+                    QPoint(rect.center().x(), rect.top() + thicknessH),
+                    QPoint(rect.center().x(), rect.top())
+                };
+
+                painter->drawPolygon(points, 6);
+            }
+
+            //bottom left
+            corner = wangId.cornerColor(2);
+            if (corner > 0) {
+                setWangStyle(painter, wangSet, corner, false);
+
+                const QPoint points[] = {
+                    rect.bottomLeft(),
+                    QPoint(rect.left(), rect.center().y()),
+                    QPoint(rect.left() + thicknessW, rect.center().y()),
+                    rect.bottomLeft() + QPoint(thicknessW, -thicknessH),
+                    QPoint(rect.center().x(), rect.bottom() - thicknessH),
+                    QPoint(rect.center().x(), rect.bottom())
+                };
+
+                painter->drawPolygon(points, 6);
+            }
+        }
+    }
+
+    painter->restore();
+}
+
 void TileDelegate::paint(QPainter *painter,
                          const QStyleOptionViewItem &option,
                          const QModelIndex &index) const
@@ -359,6 +643,29 @@ void TileDelegate::paint(QPainter *painter,
 
         painter->restore();
     }
+
+    if (mTilesetView->isEditWangSet()) {
+        painter->save();
+        painter->setTransform(tilesetGridTransform(*tile->tileset(), targetRect.center()), true);
+
+        if (WangSet *wangSet = mTilesetView->wangSet()) {
+
+            paintWangOverlay(painter, wangSet->wangIdOfTile(tile),
+                             wangSet,
+                             targetRect);
+
+            if (mTilesetView->hoveredIndex() == index) {
+                qreal opacity = painter->opacity();
+                painter->setOpacity(0.9);
+                paintWangOverlay(painter, mTilesetView->wangId(),
+                                 wangSet,
+                                 targetRect);
+                painter->setOpacity(opacity);
+            }
+        }
+
+        painter->restore();
+    }
 }
 
 QSize TileDelegate::sizeHint(const QStyleOptionViewItem & /* option */,
@@ -398,10 +705,16 @@ TilesetView::TilesetView(QWidget *parent)
     , mTilesetDocument(nullptr)
     , mMarkAnimatedTiles(true)
     , mEditTerrain(false)
+    , mEditWangSet(false)
+    , mWangBehavior(WholeId)
     , mEraseTerrain(false)
     , mTerrain(nullptr)
+    , mWangSet(nullptr)
+    , mWangId(0)
+    , mWangColor(0)
     , mHoveredCorner(0)
     , mTerrainChanged(false)
+    , mWangIdChanged(false)
     , mHandScrolling(false)
     , mImageMissingIcon(QStringLiteral("://images/32x32/image-missing.png"))
 {
@@ -445,7 +758,7 @@ void TilesetView::setTilesetDocument(TilesetDocument *tilesetDocument)
 
 QSize TilesetView::sizeHint() const
 {
-    return Utils::dpiScaled(QSize(130, 100));
+    return Utils::dpiScaled(QSize(260, 100));
 }
 
 int TilesetView::sizeHintForColumn(int column) const
@@ -527,6 +840,44 @@ void TilesetView::keyPressEvent(QKeyEvent *event)
         mZoomable->resetZoom();
         return;
     }
+
+    if (mEditWangSet && !(event->modifiers() & Qt::ControlModifier)) {
+
+        if (event->key() == Qt::Key_Z) {
+            if (event->modifiers() & Qt::ShiftModifier)
+                mWangId.rotate(-1);
+            else
+                mWangId.rotate(1);
+
+            if (mHoveredIndex.isValid())
+                update(mHoveredIndex);
+
+            emit currentWangIdChanged(mWangId);
+
+            return;
+        }
+        if (event->key() == Qt::Key_X) {
+            mWangId.flipHorizontally();
+
+            if (mHoveredIndex.isValid())
+                update(mHoveredIndex);
+
+            emit currentWangIdChanged(mWangId);
+
+            return;
+        }
+        if (event->key() == Qt::Key_Y) {
+            mWangId.flipVertically();
+
+            if (mHoveredIndex.isValid())
+                update(mHoveredIndex);
+
+            emit currentWangIdChanged(mWangId);
+
+            return;
+        }
+    }
+
     return QTableView::keyPressEvent(event);
 }
 
@@ -540,6 +891,16 @@ void TilesetView::setEditTerrain(bool enabled)
     viewport()->update();
 }
 
+void TilesetView::setEditWangSet(bool enabled)
+{
+    if (mEditWangSet == enabled)
+        return;
+
+    mEditWangSet = enabled;
+    setMouseTracking(true);
+    viewport()->update();
+}
+
 /**
  * The id of the terrain currently being specified. Returns -1 when no terrain
  * is set (used for erasing terrain info).
@@ -559,6 +920,57 @@ void TilesetView::setTerrain(const Terrain *terrain)
         viewport()->update();
 }
 
+void TilesetView::setWangSet(WangSet *wangSet)
+{
+    if (mWangSet == wangSet)
+        return;
+
+    mWangSet = wangSet;
+
+    if (mEditWangSet)
+        viewport()->update();
+}
+
+void TilesetView::setWangId(WangId wangId)
+{
+    mWangBehavior = WholeId;
+    mWangColor = 0;
+
+    if (!mWangSet || wangId == mWangId)
+        return;
+
+    Q_ASSERT(mWangSet->wangIdIsValid(wangId));
+
+    mWangId = wangId;
+
+    if (mEditWangSet && hoveredIndex().isValid())
+        update(hoveredIndex());
+}
+
+void TilesetView::setWangEdgeColor(int color)
+{
+    if (!color)
+        setWangId(0);
+
+    mWangBehavior = Edge;
+
+    Q_ASSERT(color <= mWangSet->edgeColorCount());
+
+    mWangColor = color;
+}
+
+void TilesetView::setWangCornerColor(int color)
+{
+    if (!color)
+        setWangId(0);
+
+    mWangBehavior = Corner;
+
+    Q_ASSERT(color <= mWangSet->cornerColorCount());
+
+    mWangColor = color;
+}
+
 QIcon TilesetView::imageMissingIcon() const
 {
     return QIcon::fromTheme(QLatin1String("image-missing"), mImageMissingIcon);
@@ -572,13 +984,21 @@ void TilesetView::mousePressEvent(QMouseEvent *event)
         return;
     }
 
-    if (!mEditTerrain) {
-        QTableView::mousePressEvent(event);
+    if (mEditTerrain) {
+        if (event->button() == Qt::LeftButton)
+            applyTerrain();
+
         return;
     }
 
-    if (event->button() == Qt::LeftButton)
-        applyTerrain();
+    if (mEditWangSet) {
+        if (event->button() == Qt::LeftButton)
+            applyWangId();
+
+        return;
+    }
+
+    QTableView::mousePressEvent(event);
 }
 
 void TilesetView::mouseMoveEvent(QMouseEvent *event)
@@ -598,40 +1018,110 @@ void TilesetView::mouseMoveEvent(QMouseEvent *event)
         return;
     }
 
-    if (!mEditTerrain) {
-        QTableView::mouseMoveEvent(event);
+    if (mEditWangSet) {
+        if (!mWangSet)
+            return;
+
+        const QPoint pos = event->pos();
+        const QModelIndex hoveredIndex = indexAt(pos);
+        const QModelIndex previousHoveredIndex = mHoveredIndex;
+        mHoveredIndex = hoveredIndex;
+
+        WangId wangId = mWangId;
+
+        if (mWangBehavior != WholeId) {
+            QRect tileRect = visualRect(mHoveredIndex);
+            const auto t = tilesetGridTransform(*tilesetDocument()->tileset(), tileRect.center());
+            const auto mappedPos = t.inverted().map(pos);
+            QPoint tileLocalPos = mappedPos - tileRect.topLeft();
+            QPointF tileLocalPosF((qreal) tileLocalPos.x() / tileRect.width(),
+                                  (qreal) tileLocalPos.y() / tileRect.height());
+            tileLocalPosF -= QPointF(0.5, 0.5);
+
+            wangId = 0;
+            if (mWangBehavior == Edge) {
+                if (tileLocalPosF.x() < tileLocalPosF.y()) {
+                    if (tileLocalPosF.x() > -tileLocalPosF.y())
+                        wangId.setEdgeColor(2, mWangColor);
+                    else
+                        wangId.setEdgeColor(3, mWangColor);
+                } else {
+                    if (tileLocalPosF.x() > -tileLocalPosF.y())
+                        wangId.setEdgeColor(1, mWangColor);
+                    else
+                        wangId.setEdgeColor(0, mWangColor);
+                }
+            } else {
+                if (tileLocalPosF.x() > 0) {
+                    if (tileLocalPosF.y() > 0)
+                        wangId.setCornerColor(1, mWangColor);
+                    else
+                        wangId.setCornerColor(0, mWangColor);
+                } else {
+                    if (tileLocalPosF.y() > 0)
+                        wangId.setCornerColor(2, mWangColor);
+                    else
+                        wangId.setCornerColor(3, mWangColor);
+                }
+            }
+        }
+
+        Q_ASSERT(mWangSet->wangIdIsValid(wangId));
+
+        if (previousHoveredIndex != mHoveredIndex || wangId != mWangId) {
+            mWangId = wangId;
+
+            if (previousHoveredIndex.isValid())
+                update(previousHoveredIndex);
+            if (mHoveredIndex.isValid())
+                update(mHoveredIndex);
+        }
+
+        if (event->buttons() & Qt::LeftButton)
+            applyWangId();
+
         return;
     }
 
-    const QPoint pos = event->pos();
-    const QModelIndex hoveredIndex = indexAt(pos);
-    int hoveredCorner = 0;
-
-    if (hoveredIndex.isValid()) {
-        const QPoint center = visualRect(hoveredIndex).center();
-
-        const auto t = tilesetGridTransform(*tilesetDocument()->tileset(), center);
-        const auto mappedPos = t.inverted().map(pos);
-
-        if (mappedPos.x() > center.x())
-            hoveredCorner += 1;
-        if (mappedPos.y() > center.y())
-            hoveredCorner += 2;
-    }
-
-    if (mHoveredIndex != hoveredIndex || mHoveredCorner != hoveredCorner) {
+    if (mEditTerrain) {
+        const QPoint pos = event->pos();
+        const QModelIndex hoveredIndex = indexAt(pos);
         const QModelIndex previousHoveredIndex = mHoveredIndex;
         mHoveredIndex = hoveredIndex;
-        mHoveredCorner = hoveredCorner;
+        int previousHoverCorner = mHoveredCorner;
+        int hoveredCorner = 0;
 
-        if (previousHoveredIndex.isValid())
-            update(previousHoveredIndex);
-        if (previousHoveredIndex != mHoveredIndex && mHoveredIndex.isValid())
-            update(mHoveredIndex);
+        if (mHoveredIndex.isValid()) {
+            const QPoint center = visualRect(hoveredIndex).center();
+
+            const auto t = tilesetGridTransform(*tilesetDocument()->tileset(), center);
+            const auto mappedPos = t.inverted().map(pos);
+
+            if (mappedPos.x() > center.x())
+                hoveredCorner += 1;
+            if (mappedPos.y() > center.y())
+                hoveredCorner += 2;
+
+            mHoveredCorner = hoveredCorner;
+        }
+
+        if (previousHoveredIndex != mHoveredIndex) {
+            if (previousHoveredIndex.isValid())
+                update(previousHoveredIndex);
+            if (mHoveredIndex.isValid())
+                update(mHoveredIndex);
+        } else if (previousHoverCorner != mHoveredCorner) {
+            if (mHoveredIndex.isValid())
+                update(mHoveredIndex);
+        }
+
+        if (event->buttons() & Qt::LeftButton)
+            applyTerrain();
+
+        return;
     }
 
-    if (event->buttons() & Qt::LeftButton)
-        applyTerrain();
+    QTableView::mouseMoveEvent(event);
 }
 
 void TilesetView::mouseReleaseEvent(QMouseEvent *event)
@@ -641,13 +1131,30 @@ void TilesetView::mouseReleaseEvent(QMouseEvent *event)
         return;
     }
 
-    if (!mEditTerrain) {
-        QTableView::mouseReleaseEvent(event);
+    if (mEditTerrain) {
+        if (event->button() == Qt::LeftButton)
+            finishTerrainChange();
+
         return;
     }
 
-    if (event->button() == Qt::LeftButton)
-        finishTerrainChange();
+    if (mEditWangSet) {
+        if (event->button() == Qt::LeftButton)
+            finishWangIdChange();
+
+        return;
+    }
+
+    QTableView::mouseReleaseEvent(event);
+    return;
+}
+
+void TilesetView::enterEvent(QEvent *event)
+{
+    if (mEditWangSet)
+        setFocus();
+
+    QTableView::enterEvent(event);
 }
 
 void TilesetView::leaveEvent(QEvent *event)
@@ -707,6 +1214,19 @@ void TilesetView::contextMenuEvent(QContextMenuEvent *event)
                 QAction *setImage = menu.addAction(tr("Set Terrain Image"));
                 connect(setImage, SIGNAL(triggered()), SLOT(selectTerrainImage()));
             }
+        } else if (mEditWangSet) {
+            selectionModel()->setCurrentIndex(index,
+                                              QItemSelectionModel::SelectCurrent |
+                                              QItemSelectionModel::Clear);
+
+            if (mWangSet) {
+                QAction *setImage = menu.addAction(tr("Set Wang Set Image"));
+                connect(setImage, SIGNAL(triggered()), SLOT(selectWangSetImage()));
+            }
+            if (mWangBehavior != WholeId && mWangColor) {
+                QAction *setImage = menu.addAction(tr("Set Wang Color Image"));
+                connect(setImage, SIGNAL(triggered()), SLOT(selectWangColorImage()));
+            }
         } else if (mTilesetDocument) {
             QAction *tileProperties = menu.addAction(propIcon,
                                                      tr("Tile &Properties..."));
@@ -724,17 +1244,6 @@ void TilesetView::contextMenuEvent(QContextMenuEvent *event)
             swapTilesAction->setEnabled(exactlyTwoTilesSelected);
             connect(swapTilesAction, SIGNAL(triggered()),
                     SLOT(swapTiles()));
-
-            bool onlyOneTileSelected =
-                    (selectionModel()->selectedIndexes().size() == 1);
-
-            // TODO: check that at least one object is selected
-
-            QAction *changeSelectedMapObjectsTileAction =
-                    menu.addAction(tr("&Replace Tile of Selected Objects"));
-            changeSelectedMapObjectsTileAction->setEnabled(onlyOneTileSelected);
-            connect(changeSelectedMapObjectsTileAction, &QAction::triggered,
-                    this, &TilesetView::changeSelectedMapObjectsTile);
         }
 
         menu.addSeparator();
@@ -763,6 +1272,18 @@ void TilesetView::selectTerrainImage()
         emit terrainImageSelected(tile);
 }
 
+void TilesetView::selectWangSetImage()
+{
+    if (Tile *tile = currentTile())
+        emit wangSetImageSelected(tile);
+}
+
+void TilesetView::selectWangColorImage()
+{
+    if (Tile *tile = currentTile())
+        emit wangColorImageSelected(tile, mWangBehavior == Edge, mWangColor);
+}
+
 void TilesetView::editTileProperties()
 {
     Q_ASSERT(mTilesetDocument);
@@ -791,14 +1312,6 @@ void TilesetView::swapTiles()
     emit swapTilesRequested(tile1, tile2);
 }
 
-void TilesetView::changeSelectedMapObjectsTile()
-{
-    const QModelIndexList selectedIndexes = selectionModel()->selectedIndexes();
-    const TilesetModel *model = tilesetModel();
-    Tile *tile = model->tileAt(selectedIndexes[0]);
-    emit changeSelectedMapObjectsTileRequested(tile);
-}
-
 void TilesetView::setDrawGrid(bool drawGrid)
 {
     mDrawGrid = drawGrid;
@@ -843,6 +1356,50 @@ void TilesetView::finishTerrainChange()
     mTerrainChanged = false;
 }
 
+void TilesetView::applyWangId()
+{
+    if (!mHoveredIndex.isValid() || !mWangSet)
+        return;
+
+    Tile *tile = tilesetModel()->tileAt(mHoveredIndex);
+    if (!tile)
+        return;
+
+    WangId previousWangId = mWangSet->wangIdOfTile(tile);
+    WangId newWangId = mWangId;
+
+    if (mWangBehavior != WholeId) {
+        for (int i = 0; i < 8; ++i) {
+            if (!newWangId.indexColor(i))
+                newWangId.setIndexColor(i, previousWangId.indexColor(i));
+        }
+    }
+
+    if (newWangId == previousWangId)
+        return;
+
+    bool wasUnused = !mWangSet->wangIdIsUsed(newWangId);
+
+    QUndoCommand *command = new ChangeTileWangId(mTilesetDocument, mWangSet, tile, newWangId);
+    mTilesetDocument->undoStack()->push(command);
+    mWangIdChanged = true;
+
+    if (!mWangSet->wangIdIsUsed(previousWangId))
+        emit wangIdUsedChanged(previousWangId);
+
+    if (wasUnused)
+        emit wangIdUsedChanged(newWangId);
+}
+
+void TilesetView::finishWangIdChange()
+{
+    if (!mWangIdChanged)
+        return;
+
+    mTilesetDocument->undoStack()->push(new ChangeTileWangId);
+    mWangIdChanged = false;
+}
+
 Tile *TilesetView::currentTile() const
 {
     const TilesetModel *model = tilesetModel();
