@@ -21,6 +21,7 @@
 #pragma once
 
 #include "tilesetmodel.h"
+#include "wangset.h"
 
 #include <QTableView>
 
@@ -91,6 +92,9 @@ public:
      */
     void setEditTerrain(bool enabled);
 
+    void setEditWangSet(bool enabled);
+    bool isEditWangSet() const { return mEditWangSet; }
+
     /**
      * Sets whether terrain editing is in "erase" mode.
      * \sa setEditTerrain
@@ -105,6 +109,17 @@ public:
      */
     void setTerrain(const Terrain *terrain);
 
+    WangSet *wangSet() const { return mWangSet; }
+    void setWangSet(WangSet *wangSet);
+
+    WangId wangId() const { return mWangId; }
+    //sets the WangId and changes WangBehavior to WholeId
+    void setWangId(WangId  wangId);
+
+    //Sets the wangColor, and changes WangBehavior to edges/corners
+    void setWangEdgeColor(int color);
+    void setWangCornerColor(int color);
+
     QModelIndex hoveredIndex() const { return mHoveredIndex; }
     int hoveredCorner() const { return mHoveredCorner; }
 
@@ -115,8 +130,11 @@ public:
 signals:
     void createNewTerrain(Tile *tile);
     void terrainImageSelected(Tile *tile);
+    void wangSetImageSelected(Tile *tile);
+    void wangColorImageSelected(Tile *tile, bool isEdge, int index);
+    void wangIdUsedChanged(WangId wangId);
+    void currentWangIdChanged(WangId wangId);
     void swapTilesRequested(Tile *tileA, Tile *tileB);
-    void changeSelectedMapObjectsTileRequested(Tile *tile);
 
 protected:
     bool event(QEvent *event) override;
@@ -124,6 +142,7 @@ protected:
     void mousePressEvent(QMouseEvent *event) override;
     void mouseMoveEvent(QMouseEvent *event) override;
     void mouseReleaseEvent(QMouseEvent *event) override;
+    void enterEvent(QEvent *) override;
     void leaveEvent(QEvent *) override;
     void wheelEvent(QWheelEvent *event) override;
     void contextMenuEvent(QContextMenuEvent *event) override;
@@ -131,9 +150,10 @@ protected:
 private slots:
     void addTerrainType();
     void selectTerrainImage();
+    void selectWangSetImage();
+    void selectWangColorImage();
     void editTileProperties();
     void swapTiles();
-    void changeSelectedMapObjectsTile();
     void setDrawGrid(bool drawGrid);
 
     void adjustScale();
@@ -141,20 +161,34 @@ private slots:
 private:
     void applyTerrain();
     void finishTerrainChange();
+    void applyWangId();
+    void finishWangIdChange();
     Tile *currentTile() const;
     void setHandScrolling(bool handScrolling);
 
+    enum WangBehavior {
+        WholeId, //Assigning templates
+        Corner,  //Assigning color to corners
+        Edge     //Assigning color to edges
+    };
+
     Zoomable *mZoomable;
     TilesetDocument *mTilesetDocument;
     bool mDrawGrid;
 
     bool mMarkAnimatedTiles;
     bool mEditTerrain;
+    bool mEditWangSet;
+    WangBehavior mWangBehavior;
     bool mEraseTerrain;
     const Terrain *mTerrain;
+    WangSet *mWangSet;
+    WangId mWangId;
+    int mWangColor;
     QModelIndex mHoveredIndex;
     int mHoveredCorner;
     bool mTerrainChanged;
+    bool mWangIdChanged;
 
     bool mHandScrolling;
     QPoint mLastMousePos;
